/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 17, 2010

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "arg_parser.h"
#include "platform.h"
#include "solvers.h"


int			VERBOSE = 1;
bool		RUN_BOUNDED = false;
bool		RANDOM_BOUNDS = true;
bool		ALLOW_NEG_ROW_UB = true;
bool		SAT_ONLY = false;
bool		RUN_CALCS = true;
bool		SKIP_WARMUP = false;
bool		ABORT_ON_DEVIATE = false;
int			HEURISTIC = 1;
bool		USE_METHOD [4] = { false, false, false, false };
int			PROBLEM_SIZE [3] = { 3, 3, 1 };
int			PROBLEM_SEED [2] = { 0, 0 };
char const*	PROBLEM_FILE = 0;
char		PROCESS_FILE [1024];
char		TIMINGS_FILE [1024];
char		STATISTICS_FILE [1024];

void print_help ()
{
	printf ("usage: gpu_simplex [options] [file]\n"
			"\n"
			"options:\n"
			"    --help          print a list of options\n"
			"    --heuristic=m   select heuristic to use:\n"
			"                      0: Dantzig\n"
			"                      1: steepest-edge\n"
			"    --gpu           use GPU implementation \n"
			"    --glpk          use GLPK\n"
			"    --uploadonly    only benchmark the upload performance\n"
			"    --nowarmup      skip solving 20 small LP problems for warm-up\n"
			"    --errabort      abort process on precision deviations\n"
			"    --bounded       assume bounds for variables, i.e., [0,1]\n"
			"    --unbounded     assume all variables are >0\n"
			"    --sat           test for satisfiability only\n"
			"    --optimal       find the optimal solution\n"
			"    --negative_b    for a random problem, allow negative values for b\n"
			"    --positive_b    for a random problem, allow only positive values for b\n"
			"    --random_bounds assume random bounds for all variables\n"
			"    --fixed_bounds  assume fixed bounds of [0,1] for all variables\n"
			"    --size_start=x  for benchmarking, start creating random problems of size x\n"
			"    --size_end=y      through y, where the size step indicates the interval\n"
			"    --size_step=s     between two steps\n"
			"    --size=x        creating random problems of size x only\n"
			"    --seed_start=x  for benchmarking, the start and end seeds to use for each\n"
			"    --seed_end=y      problem size, ie. for each size (y-x) problems are solved\n"
			"    --seed=x        creating random problems with seed x only\n"
			"    --log=L file    base name for the log files; one will be 'L -- process.log'\n"
			"                      and the other will be 'L -- timings.log'\n"
			"    --verbose=x     set verbosity level\n"
			"\n"
			"\n"
			"If options are conflicting, the last one will take effect. If a file name is\n"
			"given, the options for benchmarks and random problems will be ignored.\n"
			"\n");
}

int read_args (int argc, char const* argv [])
{
	if (argc == 0)
		return 1;

	for (int i = 0; i < argc; ++ i)
	{
		if (stricmp (argv [i], "--help") == 0)
		{
			return 1;
		}
		else if (strnicmp (argv [i], "--heuristic=", 12) == 0)
		{
			HEURISTIC = strtol (argv [i] + 12, 0, 10);
		}
		else if (stricmp (argv [i], "--gpu") == 0)
		{
			USE_METHOD [imp_gpu] = true;
		}
		else if (stricmp (argv [i], "--glpk") == 0)
		{
			USE_METHOD [imp_glpk] = true;
		}
		else if (stricmp (argv [i], "--uploadonly") == 0)
		{
			RUN_CALCS = false;
		}
		else if (stricmp (argv [i], "--nowarmup") == 0)
		{
			SKIP_WARMUP = true;
		}
		else if (stricmp (argv [i], "--bounded") == 0)
		{
			RUN_BOUNDED = true;
		}
		else if (stricmp (argv [i], "--unbounded") == 0)
		{
			RUN_BOUNDED = false;
		}
		else if (stricmp (argv [i], "--sat") == 0)
		{
			SAT_ONLY = true;
		}
		else if (stricmp (argv [i], "--optimal") == 0)
		{
			SAT_ONLY = false;
		}
		else if (stricmp (argv [i], "--negative_b") == 0)
		{
			ALLOW_NEG_ROW_UB = true;
		}
		else if (stricmp (argv [i], "--positive_b") == 0)
		{
			ALLOW_NEG_ROW_UB = false;
		}
		else if (stricmp (argv [i], "--random_bounds") == 0)
		{
			RANDOM_BOUNDS = true;
		}
		else if (stricmp (argv [i], "--fixed_bounds") == 0)
		{
			RANDOM_BOUNDS = false;
		}
		else if (stricmp (argv [i], "--errabort") == 0)
		{
			ABORT_ON_DEVIATE = true;
		}
		else if (strnicmp (argv [i], "--verbose=", 10) == 0)
		{
			VERBOSE = strtol (argv [i] + 10, 0, 10);
		}
		else if (strnicmp (argv [i], "--size_start=", 13) == 0)
		{
			PROBLEM_SIZE [0] = strtol (argv [i] + 13, 0, 10);
		}
		else if (strnicmp (argv [i], "--size_end=", 11) == 0)
		{
			PROBLEM_SIZE [1] = strtol (argv [i] + 11, 0, 10);
		}
		else if (strnicmp (argv [i], "--size_step=", 12) == 0)
		{
			PROBLEM_SIZE [2] = strtol (argv [i] + 12, 0, 10);
		}
		else if (strnicmp (argv [i], "--size=", 7) == 0)
		{
			PROBLEM_SIZE [0] = PROBLEM_SIZE [1] = strtol (argv [i] + 7, 0, 10);
			PROBLEM_SIZE [2] = 1;
		}
		else if (strnicmp (argv [i], "--seed_start=", 13) == 0)
		{
			PROBLEM_SEED [0] = strtol (argv [i] + 13, 0, 10);
		}
		else if (strnicmp (argv [i], "--seed_end=", 11) == 0)
		{
			PROBLEM_SEED [1] = strtol (argv [i] + 11, 0, 10);
		}
		else if (strnicmp (argv [i], "--seed=", 7) == 0)
		{
			PROBLEM_SEED [0] = PROBLEM_SEED [1] = strtol (argv [i] + 7, 0, 10);
		}
		else if (strnicmp (argv [i], "--log=", 6) == 0)
		{
			strncpy (PROCESS_FILE, argv [i] + 6, sizeof (PROCESS_FILE) - 16);
			strncpy (TIMINGS_FILE, argv [i] + 6, sizeof (TIMINGS_FILE) - 16);
			strncpy (STATISTICS_FILE, argv [i] + 6, sizeof (STATISTICS_FILE) - 20);

			strcat (PROCESS_FILE, " -- process.txt");
			strcat (TIMINGS_FILE, " -- timings.txt");
			strcat (STATISTICS_FILE, " -- statistics.txt");

			PROCESS_FILE [sizeof (PROCESS_FILE) - 1] = 0;
			TIMINGS_FILE [sizeof (TIMINGS_FILE) - 1] = 0;
			STATISTICS_FILE [sizeof (STATISTICS_FILE) - 1] = 0;
		}
		else if (i + 1 == argc)
		{
			PROBLEM_FILE = argv [i];
		}
		else
		{
			return -1;
		}
	}

	if (PROCESS_FILE [0] == 0)
		sprintf (PROCESS_FILE, "%s,%s,%s -- process.txt", RUN_BOUNDED ? "bounded" : "unbounded",
														  ALLOW_NEG_ROW_UB ? "negative" : "positive",
														  SAT_ONLY ? "feasible" : "optimal");

	if (TIMINGS_FILE [0] == 0)
		sprintf (TIMINGS_FILE, "%s,%s,%s -- timings.txt", RUN_BOUNDED ? "bounded" : "unbounded",
														  ALLOW_NEG_ROW_UB ? "negative" : "positive",
														  SAT_ONLY ? "feasible" : "optimal");
	if (STATISTICS_FILE [0] == 0)
		sprintf (STATISTICS_FILE, "%s,%s,%s -- statistics.txt", RUN_BOUNDED ? "bounded" : "unbounded",
																ALLOW_NEG_ROW_UB ? "negative" : "positive",
																SAT_ONLY ? "feasible" : "optimal");

	return 0;
}

