/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011


#pragma once

#include "bounds.h"
#include "matrix_storage.h"

template <typename environment_t, typename storage>
struct matrix;


/**
 *
 * @author Jakob Bieling
 */
template <typename environment_t>
class engine_base
{
	engine_base (engine_base const&);
	engine_base& operator= (engine_base const&);

protected:

	typedef typename environment_t::value_type	value_type;
	typedef typename environment_t::shader		shader;
	typedef typename environment_t::texture		texture;
	typedef typename environment_t::texture4f	texture4f;
	typedef typename environment_t::matrix		matrix;
	typedef typename environment_t::matrix4f	matrix4f;

	value_type	m_rd_buff [4];
	float 		texCoords0 [4 * 2];
	float		texCoords1 [4 * 2];

	engine_base ()
	{
	}
	
	typename environment_t::shader::manager* shader_manager;

public:

	virtual ~engine_base () {}

	float get_last_result (int chn)
	{
		return m_rd_buff [chn];
	}

	virtual void precache_matrix (matrix const& res) = 0;
	virtual void precache_run () = 0;

	virtual void zero (texture& tex) = 0;
	virtual void zero (texture& tex, int x, int y, int w, int h) = 0;

	/**
	 * Copy 'src' into 'dst'.
	 * 
	 * @param dst			Destination matrix.
	 * @param dstx			x-offset in the destination matrix.
	 * @param dsty			y-offset in the destination matrix.
	 * @param src			Source matrix.
	 * @param srcx			x-offset in the source matrix.
	 * @param srcy			y-offset in the source matrix.
	 * @param srcw			Width of the rectangle in the source matrix to copy.
	 * @param srch			Height of the rectangle in the source matrix to copy.
	 */
	void copy (matrix& dst, matrix const& src)
	{
		dst.bounds = src.bounds;

		shader& s = shader_manager->get_shader (shader::COPY);
		s.bind ();
		s.set_parameter ("src", src.texture->data);

		set_tex_coords (texCoords0, src.bounds);
		render_arrays (dst);
	}

	void copy (matrix& dst, int dstx, int dsty, matrix const& src, int srcx, int srcy, int srcw, int srch)
	{
		dst.bounds.width = std::max (dst.bounds.width, dstx + srcw);
		dst.bounds.height = std::max (dst.bounds.height, dsty + srch);
		
		shader& s = shader_manager->get_shader (shader::COPY);
		s.bind ();
		s.set_parameter ("src", src.texture->data);

		set_tex_coords (texCoords0, srcx, srcy, srcw, srch);
		render_arrays (dst.texture->data, dstx, dsty, srcw, srch);
		reset_tex_coord_offsets (texCoords0);
	}

	/**
	 * Set a texture's contents to the identity.
	 * 
	 * @param res	Texture to set to the identity.
	 * @param xoff	x-offset, before which no data will be modified.
	 * @param yoff	y-offset, above which all values will be set to 0.
	 */
	void identity (matrix& res, int xoff, int yoff)
	{
		shader& s = shader_manager->get_shader (shader::IDENTITY);
		s.bind ();
	//	if (u == texture.Usage.COMPACT)
	//	{
	//		s.set_parameter ("RELXOFFSET", xoff % 4);
	//		s.set_parameter ("tex", tex);
	//		xoff /= 4;
	//	}

		yoff += res.bounds.yoffset;

		set_tex_coords (texCoords0, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
		render_arrays (res.texture->data, xoff, yoff, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
	}

	/**
	 * Set a texture's contents to the identity but takes the sign of the
	 * respective element from the values of a column-vector 'signs'.
	 * 
	 * @param res	Texture to set to the identity.
	 * @param xoff	x-offset, before which no data will be modified.
	 * @param yoff	y-offset, above which all values will be set to 0.
	 * @param signs	Texture to get the signs from. The actual values do
	 *				do not matter.
	 */
	void identity (matrix& res, int xoff, int yoff, matrix const& signs)
	{
		shader& s = shader_manager->get_shader (shader::SIGNED_IDENTITY);
		s.bind ();
		s.set_parameter ("signs", signs.texture->data);

		yoff += res.bounds.yoffset;

		set_tex_coords (texCoords0, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
		render_arrays (res.texture->data, xoff, yoff, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
	}

	void artificials (matrix& res, int xoff, int yoff, matrix const& signs)
	{
		shader& s = shader_manager->get_shader (shader::ARTIFICIALS);
		s.bind ();
		s.set_parameter ("signs", signs.texture->data);

		yoff += res.bounds.yoffset;

		set_tex_coords (texCoords0, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
		render_arrays (res.texture->data, xoff, yoff, res.bounds.tex_width () - xoff, res.bounds.tex_height () - yoff);
	}

	/**
	 * Subtract 'rhs' from 'lhs' and store the result in 'res'.
	 * 
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 * @param res	Matrix with enough storage to hold the result.
	 */
	void subtract (matrix const& lhs, matrix const& rhs, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::SUB);
		s.bind ();
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);

		res.bounds = lhs.bounds;

		set_tex_coords (texCoords0, lhs.bounds.width, lhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Multiply 'lhs' and 'rhs' and store the result in 'res'. Note that the
	 * boundary y-offset for 'lhs' must be 0! The caller is responsible for
	 * passing a matrix as 'res' that can hold the result. The used width and
	 * height values of 'res' will be updated accordingly.
	 * 
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 * @param res	Matrix with enough storage to hold the result.
	 */
	void multiply (matrix const& lhs, matrix const& rhs, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::MUL);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("rhsYoffset", rhs.bounds.yoffset);

		res.bounds.width = rhs.bounds.width;
		res.bounds.height = lhs.bounds.height;

		set_tex_coords (texCoords0, lhs.bounds.width, lhs.bounds.height);
		set_tex_coords (texCoords1, rhs.bounds.width, rhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Multiply 'lhs' and 'rhs' and store the result in 'res'. Note that the
	 * boundary y-offset for 'lhs' must be 0! The caller is responsible for
	 * passing a matrix as 'res' that can hold the result. The used width and
	 * height values of 'res' will be updated accordingly.
	 *
	 * Note: 'ncols' overrides any boundary information of the matrices!
	 * 
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 * @param res	Matrix with enough storage to hold the result.
	 * @param ncols	Number of columns of 'rhs' to take part in the multiplication.
	 *				If this value is less than the actual amount of columns, the
	 *				remaining columns will be discarded. If this value is greater
	 *				than the actual amount of columns, the results are undefined.
	 */
	void multiply (matrix const& lhs, matrix const& rhs, matrix& res, int ncols)
	{
		shader& s = shader_manager->get_shader (shader::MUL);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("rhsYoffset", rhs.bounds.yoffset);

		res.bounds.width = ncols;
		res.bounds.height = lhs.bounds.height;

		set_tex_coords (texCoords0, lhs.bounds.width, lhs.bounds.height);
		set_tex_coords (texCoords1, ncols, rhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Multiply 'lhs' and 'rhs' and store the result in 'res'. Note that the
	 * boundary y-offset for 'lhs' must be 0! The caller is responsible for
	 * passing a matrix as 'res' that can hold the result. The used width and
	 * height values of 'res' will be updated accordingly.
	 *
	 * Note: 'offx' abd 'ncols' overrides any boundary information of the matrices!
	 * 
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 * @param res	Matrix with enough storage to hold the result.
	 * @param offx	First column of 'rhs' to take part in the multiplication.
	 * @param ncols	Number of columns of 'rhs' to take part in the multiplication.
	 *				If this value is less than the actual amount of columns, the
	 *				remaining columns will be discarded. If this value is greater
	 *				than the actual amount of columns, the results are undefined.
	 */
	void multiply (matrix const& lhs, matrix const& rhs, matrix& res, int offx, int ncols)
	{
		shader& s = shader_manager->get_shader (shader::MUL);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("rhsYoffset", rhs.bounds.yoffset);

		res.bounds.width = ncols;
		res.bounds.height = lhs.bounds.height;

		set_tex_coords (texCoords0, lhs.bounds.width, lhs.bounds.height);
		set_tex_coords (texCoords1, offx, 0, ncols, rhs.bounds.height);
		render_arrays (res);
		reset_tex_coord_offsets (texCoords1);
	}

	/**
	 * Multiply 'lhs' and with the specified column vector of 'rhs' and store
	 * the result in 'res'. Note that the boundary y-offset for 'lhs' must be
	 * 0! The caller is responsible for passing a matrix as 'res' that can hold
	 * the result. The used width and height values of 'res' will be updated
	 * accordingly.
	 * 
	 * @param lhs		Left-hand-side matrix.
	 * @param rhs		Right-hand-side matrix.
	 * @param rhsCol	Specifies the index of the column vector of 'rhs'
	 *					which is to be used for the multiplication. Must
	 *					always be of the form x+0.5, ie. if the third column
	 *					vector is to be multiplied, pass 2.5.
	 * @param res		Matrix with enough storage to hold the result.
	 */
	void multiply (matrix const& lhs, matrix const& rhs, float rhsCol, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::MUL_COL);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("rhsYoffset", rhs.bounds.yoffset);
		s.set_parameter ("rhsCol", rhsCol);

		res.bounds.width = 1;
		res.bounds.height = lhs.bounds.height;

		set_tex_coords (texCoords0, lhs.bounds.width, lhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Multiply a specified row of 'lhs' with 'rhs' and store the result in 'res'.
	 * Note that the boundary y-offsets for 'lhs' and 'rhs' must be 0! The caller
	 * is responsible for passing a matrix as 'res' that can hold the result. The
	 * used width and height values of 'res' will be updated accordingly.
	 * 
	 * @param lhs		Left-hand-side matrix (used transposed).
	 * @param lhsRow	Specifies the index of the row vector of 'lhs' which is be
	 *					used for the multiplication. Must always be of the form
	 *					x+0.5, ie. if the third row vector is to be multiplied,
	 *					pass 2.5.
	 * @param rhs		Right-hand-side matrix.
	 * @param res		Matrix with enough storage to hold the result.
	 * @param ncols		Number of columns of 'rhs' to take part in the multiplication.
	 *					If this value is less than the actual amount of columns, the
	 *					remaining columns will be discarded. If this value is greater
	 *					than the actual amount of columns, the results are undefined.
	 */
	void multiply (matrix const& lhs, float lhsRow, matrix const& rhs, matrix& res, int ncols)
	{
		shader& s = shader_manager->get_shader (shader::MUL_ROW);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("lhs_row", lhsRow);

		res.bounds.width = ncols;
		res.bounds.height = 1;

		set_tex_coords (texCoords0, ncols, rhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Similar to 'multiply', but 'lhs' is used in transposed order when multiplying.
	 * Note that both boundary y-offsets must be zero!
	 * 
	 * @param lhs		Left-hand-side matrix (used transposed).
	 * @param rhs		Right-hand-side matrix.
	 * @param lhsCol	Specifies the index of the column vector of 'lhs'
	 *					which is to be used for the multiplication. Must
	 *					always be of the form x+0.5, ie. if the third column
	 *					vector is to be multiplied, pass 2.5.
	 * @param res		Matrix with enough storage to hold the result.

	 */
	void multiply_transposed1 (matrix const& lhsT, matrix const& rhs, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::MUL_T1);
		s.set_special_parameter ("DIM", lhsT.bounds.tex_height ());
		s.set_parameter ("lhs", lhsT.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);

		res.bounds.width = rhs.bounds.width;
		res.bounds.height = lhsT.bounds.width;

		set_tex_coords (texCoords0, lhsT.bounds.height, lhsT.bounds.width);
		set_tex_coords (texCoords1, rhs.bounds.width, rhs.bounds.height);
		render_arrays (res);
	}

	/**
	 * Similar to 'multiply', but 'rhs' is used in transposed order when multiplying.
	 * Note that both boundary y-offsets may be non-zero.
	 * 
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix (used transposed).
	 * @param res	Matrix with enough storage to hold the result.

	 */
	void multiply_transposed2 (matrix const& lhs, matrix const& rhsT, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::MUL_T2);
		s.set_special_parameter ("DIM", lhs.bounds.tex_width ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhsT.texture->data);

		res.bounds.width = rhsT.bounds.height;	//	usually 'width', but we use 'rhsT' in transposed order
		res.bounds.height = lhs.bounds.height;

		set_tex_coords (texCoords0, lhs.bounds);
		set_tex_coords_transposed (texCoords1, rhsT.bounds);

		render_arrays (res);

		reset_tex_coord_offsets (texCoords0);
		reset_tex_coord_offsets (texCoords1);
	}

	/**
	 *	Update g, see shader for details.
	 */
	void update_g (matrix const& gamma,
				   matrix const& alpha,
				   matrix const& v,
				   matrix const& A,
				   float gamma_e,
				   matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::UPDATE_G);
		s.set_special_parameter ("DIM", v.bounds.tex_height ());
		s.set_parameter ("gamma", gamma.texture->data);
		s.set_parameter ("alpha", alpha.texture->data);
		s.set_parameter ("v", v.texture->data);
		s.set_parameter ("A", A.texture->data);
		s.set_parameter ("gamma_e", gamma_e);

		res.bounds.width = gamma.bounds.width;
		res.bounds.height = 1;

		set_tex_coords (texCoords0, gamma.bounds.width, A.bounds.height);
		render_arrays (res);
	}
	
	/**
	 *	Update g, see shader for details.
	 */
	void update_g2 (matrix const& gamma,
					matrix const& alpha,
					matrix const& v,
					int O,
					float gamma_e,
					matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::UPDATE_G2);
		s.set_special_parameter ("DIM", v.bounds.tex_height ());
		s.set_parameter ("gamma", gamma.texture->data);
		s.set_parameter ("alpha", alpha.texture->data);
		s.set_parameter ("v", v.texture->data);
		s.set_parameter ("gamma_e", gamma_e);
		s.set_parameter ("O", O);

		res.bounds.width = gamma.bounds.width;
		res.bounds.height = 1;

		set_tex_coords (texCoords0, O, 0, gamma.bounds.width - O, 1);
		render_arrays (res.texture->data, O, 0, gamma.bounds.width - O, 1);
		reset_tex_coord_offsets (texCoords0);
	}
	
	/**
	 * Subtract the first row of 'lhs' from 'rhs' and find the smallest
	 * (possibly negative) difference of the resulting elements. The index
	 * of the smallest element is stored internally and can be retrieved
	 * by calling 'get_last_result'. Element 0 of the result will always
	 * contain the index and element 1 will contain the result. In case the
	 * bounded version is used, element 2 will contain the element of 'rhs'
	 * at the "winning" column in the third row. So in our special case,
	 * 'rhs' is usually the cost-vector 'c' holding the x-values in the
	 * third row.
	 *
	 * To determine the extent of the row to use for subtraction, only the
	 * boundary information of 'lhs' is used!
	 *
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 */
	void find_entering (matrix const& lhs, matrix const& rhs) { lhs_rhs_op <shader::ENTER> (lhs, rhs, lhs.bounds.tex_width ()); }
	void find_entering_bounded (matrix const& lhs, matrix const& rhs) { lhs_rhs_op <shader::ENTER_BND> (lhs, rhs, lhs.bounds.tex_width ()); }

	/**
	 *	Identical to find_entering_*, except that GPU parallelism is
	 *	taken advantage of through a reduction scheme.
	 */
	void find_entering_reduce (matrix const& lhs, matrix const& rhs, matrix4f pingpong [2])
	{
		lhs_rhs_op_reduce_init (shader_manager->get_shader (shader::REDUCE_INIT_ENTER), lhs, rhs, lhs.bounds.tex_width (), pingpong);
		int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MINY), pingpong, 999999.0);
		lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MINYNEG_REST), pingpong [read]);
	}
	void find_entering_bounded_reduce (matrix const& lhs, matrix const& rhs, matrix4f pingpong [2])
	{
		lhs_rhs_op_reduce_init (shader_manager->get_shader (shader::REDUCE_INIT_ENTERBND), lhs, rhs, lhs.bounds.tex_width (), pingpong);
		int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MINY), pingpong, 999999.0);
		lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MINYNEG_REST), pingpong [read]);
	}
	void find_entering_bounded_reduce (matrix const& lhs, matrix const& rhs, matrix const& gamma, matrix4f pingpong [2])
	{
		lhs_rhs_op_reduce_init (shader_manager->get_shader (shader::REDUCE_INIT_ENTERBND_SE), lhs, rhs, gamma, lhs.bounds.tex_width (), pingpong);
		int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MINY), pingpong, 999999.0);
		lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MINYNEG_REST), pingpong [read]);
	}

	/**
	 * Compute the element-wise quotient of the first columns of 'lhs' and
	 * 'rhs'. The index of the smallest quotient is stored internally and
	 * can be retrieved by calling 'get_last_result'. Element 0 will contain
	 * the index and element 1 will contain the quotient.
	 *
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 */
	void find_leaving (matrix const& lhs, matrix const& rhs) { lhs_rhs_op <shader::LEAVE> (lhs, rhs, lhs.bounds.tex_height ()); }

	/**
	 * Compute the element-wise quotient of the first columns of 'lhs' and
	 * 'rhs'. In contrast to the unbounded version, this version will take
	 * variable bounds into account, as stated by the bounded simplex algorithm.
	 * 'ux' contains 1.0 if the upper bound of the variable is 1.0, and 0.0
	 * if the upper bound is +inf. The result is stored internally and can be
	 * retrieved by calling 'get_last_result'. The result will be stored as
	 * follows. Element 0: index, element 1: corresponding 't', element 2:
	 * value of the x-element to leave, element 3: 0, if the basis needs to be
	 * updated.
	 *
	 * @param lhs	Left-hand-side matrix.
	 * @param rhs	Right-hand-side matrix.
	 * @param chi	cB-index matrix.
	 * @param ux	Matrix with the upper-bound information (in the second row).
	 * @param enter	Previously chosing entering index.
	 */
	void find_leaving_bounded (matrix const& lhs, matrix const& rhs, matrix const& chi,
							   matrix const& ux, float enter)
	{
		shader& s = shader_manager->get_shader (shader::LEAVE_BND);
		s.set_special_parameter ("DIM", lhs.bounds.tex_height ());
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("chi", chi.texture->data);
		s.set_parameter ("ux", ux.texture->data);
		s.set_parameter ("enter", enter);
	//	if (res.channelUsage == texture.Usage.QUAD)
	//		s.set_parameter ("chnDim", lhs_bounds.height, lhs.channelUsage);

		render_arrays_fetch ();
	}

	/**
	 *	Identical to find_entering_*, except that GPU parallelism is
	 *	taken advantage of through a reduction scheme.
	 */
	void find_leaving_reduce (matrix const& lhs, matrix const& rhs, matrix4f pingpong [2])
	{
		lhs_rhs_op_reduce_init (shader_manager->get_shader (shader::REDUCE_INIT_LEAVE), lhs, rhs, lhs.bounds.tex_height (), pingpong);
		int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MINY), pingpong, 999999.0);
		lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MINY_REST), pingpong [read]);
	}
	virtual void find_leaving_bounded_reduce (matrix const& lhs, matrix const& rhs, matrix const& chi,
											  matrix const& ux, float enter, matrix4f pingpong [2], float maxDiff) = 0;

	void get_solution (matrix const& cB, matrix const& xB, int num_vars_with_surplus)
	{
		shader& s = shader_manager->get_shader (shader::RES);
		s.set_special_parameter ("DIM", cB.bounds.tex_width ());
		s.set_parameter ("cB", cB.texture->data);
		s.set_parameter ("xB", xB.texture->data);
		s.set_parameter ("num_vars_with_surplus", num_vars_with_surplus);

		render_arrays_fetch ();
	}

	void get_solution_bounded (matrix const& chi, matrix const& xB, matrix const& x, int num_vars_with_surplus)
	{
		shader& s = shader_manager->get_shader (shader::RES_BND);
		s.set_special_parameters ("DIM", x.bounds.tex_width (), chi.bounds.tex_width ());
		s.set_parameter ("cB", chi.texture->data);
		s.set_parameter ("xB", xB.texture->data);
		s.set_parameter ("x", x.texture->data);
		s.set_parameter ("num_vars_with_surplus", num_vars_with_surplus);

		render_arrays_fetch ();
	}

	/**
	 * Implicitely create a pivot-texture, ie. an identity texture with the
	 * 'leave'-th column set to a column-vector calculated from 'd' and
	 * multiply the result with B. Semantically, it is identical to first
	 * creating a pivot-matrix and multiplying it with 'B'.
	 * 
	 * @param d				Column-vector (as matrix) to be used as a replacement.
	 * @param leave			Index of the column of the identity matrix to be
	 *						replaced.
	 * @param B				Current inverse of the B matrix.
	 * @param res			Large enough matrix to hold the result.
	 */
	void pivot_multiply (matrix const& d, float leave, matrix const& B, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::PIVOT_MUL);
		s.bind ();
		s.set_parameter ("B", B.texture->data);
		s.set_parameter ("d", d.texture->data);
		s.set_parameter ("leave", leave);

		res.bounds.width =
		res.bounds.height = d.bounds.height;

		set_tex_coords (texCoords0, res.bounds.width, res.bounds.height);
		render_arrays (res);
	}

	/**
	 * Calculates 'xB - t * d' and stores the result in 'res'.
	 */
	void weighted_subtract (matrix const& xB, float t, matrix const& d, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::WSUB);
		s.bind ();
		s.set_parameter ("xB", xB.texture->data);
		s.set_parameter ("t", t);
		s.set_parameter ("d", d.texture->data);

		res.bounds = xB.bounds;

		set_tex_coords (texCoords0, res.bounds.width, res.bounds.height);
		render_arrays (res);
	}


	/**
	 * Use the values of 'idx' as indices into the first row of 'src' and copy
	 * the respective values from 'src' to 'res'.
	 * 
	 * @param src			Matrix holding the source values.
	 * @param idx			Matrix holding the indices.
	 * @param res			Large enough matrix (at least as wide as
	 *						'idx') to hold the result.
	 */
	void lookup (matrix const& src, matrix const& idx, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::LOOKUP);
		s.bind ();
		s.set_parameter ("src", src.texture->data);
		s.set_parameter ("idx", idx.texture->data);
	//	if (res.channelUsage != texture.Usage.REGULAR)
	//		s.set_parameter ("chnDim", idx_bounds.width, idx.channelUsage);

		res.bounds.width = idx.bounds.width;
		res.bounds.height = idx.bounds.height;

		set_tex_coords (texCoords0, res.bounds.width, res.bounds.height);
		render_arrays (res);
	}

	/**
	 * Calculate the Euclidian norms of the column vectors of 'mat'.
	 * 
	 * @param mat	The matrix to calculate the column-wise Euklidian norms of.
	 * @param res	Large enough matrix to hold the result.
	 */
	void norm2 (matrix const& mat, matrix& res)
	{
		shader& s = shader_manager->get_shader (shader::NORM2);
		s.set_special_parameter ("DIM", mat.bounds.height);
		s.set_parameter ("mat", mat.texture->data);

		res.bounds.width = mat.bounds.width;
		res.bounds.height = 1;

		set_tex_coords (texCoords0, res.bounds.width, res.bounds.height);
		render_arrays (res);
	}

	/**
	 * Calculate the Euclidian norms of the column vectors of 'mat'.
	 * 
	 * @param mat	The matrix to calculate the column-wise Euklidian norms of.
	 * @param res	Large enough matrix to hold the result.
	 * @param offx	Offset where to start storing the result.
	 */
	void norm2 (matrix const& mat, matrix& res, int res_offx)
	{
		shader& s = shader_manager->get_shader (shader::NORM2);
		s.set_special_parameter ("DIM", mat.bounds.height);
		s.set_parameter ("mat", mat.texture->data);

		res.bounds.width = res_offx + mat.bounds.width;
		res.bounds.height = 1;

		render_arrays (res.texture->data, res_offx, 0, mat.bounds.width, 1);
	}

	/**
	 * Calculate the Euclidian norms of the first column vector of 'mat'.
	 * 
	 * @param mat	The column-vector to calculate the Euklidian norm of.
	 */
	void norm2 (matrix const& mat, matrix4f pingpong [2])
	{
		lhs_rhs_op_reduce_init (shader_manager->get_shader (shader::SQUARE), mat, mat.bounds.height, pingpong);
		int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_ADD), pingpong, 0);
		lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_ADD_REST), pingpong [read]);
/*
		shader& s = shader_manager->get_shader (shader::NORM2);
		s.set_special_parameter ("DIM", mat.bounds.height);
		s.set_parameter ("mat", mat.texture->data);
		render_arrays_fetch ();*/
	}

	/**
	 * Retrieve an element (the red-channel of the texel) from a texture.
	 * 
	 * @param tex	Texture to retrieve the texel from.
	 * @param x		x-coordinate of the texel.
	 * @param y		y-coordinate of the texel.
	 * @param buff	Buffer to hold the texel, at least four floats.
	 */
	virtual void get_element (texture const& tex, int x, int y, float* buff) = 0;

	/**
	 * Retrieve a rectangle of elements (the red-channel of the texels) from a texture.
	 * 
	 * @param tex	Texture to retrieve the texel from.
	 * @param x		x-coordinate of the texel.
	 * @param y		y-coordinate of the texel.
	 * @param w		Width of the rectangle to retrieve texels from.
	 * @param h		Height of the rectangle to retrieve texels from.
	 * @param buff	Buffer to hold the texel, at least four floats.
	 */
	virtual void get_elements (texture const& tex, int x, int y, int w, int h, float* buff) = 0;

	/**
	 * Identical to get_elements, except that you may specify the channel ID (GL_RED etc) as well.
	 */
	virtual void get_elements (texture4f const& tex, int chn, int x, int y, int w, int h, float* buff) = 0;
#ifdef USE_1CHN_TEXTURE
	/**
	 * Identical to get_elements, except that all channels are retrieved. If the texture features
	 * only one channel, the remaining ones are filled with zeros.
	 */
	virtual void get_elements (texture4f const& tex,  int x, int y, int w, int h, float* buff) = 0;
#endif
	/**
	 * Sets a specified element of the texture to a value. Note that for
	 * compact textures, only one element, ie. only one channel of the texel
	 * will be modified.
	 * 
	 * @param res	Texture to set an element of.
	 * @param value	Value to set the element to.
	 * @param x		x-coordinate of the element to set. In case of compact
	 *				textures, this will specify the element index as it would
	 *				appear in the matrix, ie. an x-coordinate of 4 will change
	 *				the red channel of the texel at x-coordinate 1.
	 * @param y		y-coordinate of the element to set.
	 */
	virtual void set_element (texture& res, float value, int x, int y) = 0;

	/**
	 * Very similar to 'setElement', except that a whole range of elements
	 * is set to the same value.
	 * 
	 * @param res	Texture to set an element of.
	 * @param value	Value to set the elements to.
	 * @param x		Starting x-coordinate of the elements to set. In case
	 *				of compact textures, this will specify the element index
	 *				as it would appear in the matrix, ie. an x-coordinate of 4
	 *				will change the red channel of the texel at x-coordinate 1.
	 * @param y		y-coordinate of the elements to set.
	 * @param w		Width of the rectangle of elements to set.
	 * @param w		Height of the rectangle of elements to set.
	 */
	virtual void set_elements (texture& res, float value, int x, int y, int w, int h) = 0;

//	/**
//	 * Create a compact representation of a texture.
//	 * 
//	 * @param tex			Input texure of 'REGULAR' usage type.
//	 * @param tex_bounds	Boundary information for 'tex'. The y-offset is
//	 *						assumed to be 0!
//	 * @param res			Will contain the compacted texture.
//	 * @param res_bounds	Boundary information for 'res'. Upon return, it will
//	 *						contain the apropriate boundary information.
//	 */
//	public void compact (texture tex, bounds_t const& tex_bounds,
//						 texture res, bounds_t const& res_bounds)

private:
	
	template <typename shader::operation OP>
	void lhs_rhs_op (matrix const& lhs, matrix const& rhs, int DIM)
	{
		shader& s = shader_manager->get_shader (OP);
		s.set_special_parameter ("DIM", DIM);
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
	//	if (res.channelUsage != texture.Usage.REGULAR)
	//		s.set_parameter ("chnDim", lhs_bounds.width, lhs.channelUsage);

		render_arrays_fetch ();
	}

	virtual void lhs_rhs_op_reduce_init (shader& s, matrix const& mat, int DIM, matrix4f pingpong [2]) = 0;
	virtual void lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, int DIM, matrix4f pingpong [2]) = 0;
	virtual void lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, matrix const& gamma, int DIM, matrix4f pingpong [2]) = 0;
	virtual int  lhs_rhs_op_reduce_step (shader& s, matrix4f pingpong [2], float ydummy) { return -1; }
	virtual void lhs_rhs_op_reduce_final (shader& s, matrix4f const& pingpong) {}
	virtual void lhs_rhs_op_reduce_final (shader& s, float xN [2], matrix4f const& pingpong) {}

	/**
	 * Render a quad to the texture associated with 'res'. The boundary
	 * values specify the portion of the texture to render to.
	 * 
	 * @param res			Matrix to render to.
	 */
	virtual void render_arrays (matrix& res) = 0;

	/**
	 * Render a quad to the texture specified by 'resID'. The boundary values
	 * are given by the last four arguments and specify the portion of the
	 * texture to render to.
	 * 
	 * @param resID	ID of the texture to render to.
	 * @param x		x-offset to render to.
	 * @param y		y-offset to render to.
	 * @param w		Width to render to.
	 * @param h		Height to render to.
	 */
	virtual void render_arrays (texture& res, int x, int y, int w, int h) = 0;

	/**
	 * Identical to render_arrays with a result of 1x1 size, but fetches the
	 * result of the calculation at (0,0) after rendering. The result can be
	 * inspected through the get_last_result member.
	 */
	virtual void render_arrays_fetch () = 0;

	/**
	 * Modify the texture-coordinate buffer so it contains the passed dimensions.
	 * The buffer will be used internally by 'render_arrays'.
	 * 
	 * @param buff	Buffer to modify. Null-elements are assumed to be
	 *				already 0.
	 * @param w		Value to use for the texture width.
	 * @param h		Value to use for the texture height.
	 */
	virtual void set_tex_coords (float* buff, int w, int h) = 0;

	/**
	 * Modify the texture-coordinate buffer so it contains the passed dimensions.
	 * The buffer will be used internally by 'render_arrays'.
	 * 
	 * @param buff	Buffer to modify. Null-elements are assumed to be
	 *				already 0.
	 * @param x		Value to use for the texture x-offset.
	 * @param y		Value to use for the texture y-offset.
	 * @param w		Value to use for the texture width.
	 * @param h		Value to use for the texture height.
	 */
	virtual void set_tex_coords (float* buff, int x, int y, int w, int h) = 0;

	/**
	 * Modify the texture-coordinate buffer so it contains the passed boundary
	 * values. The buffer will be used internally by 'render_arrays'. Note that
	 * this function will also take into account the yoffset of the bounds_t
	 * structure. Call 'reset_tex_coord_offset' to reset those offsets back to the
	 * defaults.
	 * 
	 * @param buff			Buffer to modify. Null-elements are assumed to be
	 *						already 0 and will not be set again (ie. (0,0)).
	 * @param tex_bounds	Boundary information to extract the coordinates from.
	 */
	virtual void set_tex_coords (float* buff, bounds_t const& tex_bounds) = 0;

	/**
	 * Identical to 'set_tex_coords', except that, as the name suggests, the texture
	 * coordinates are applied in a transposed fashion.
	 * 
	 * @param buff			Buffer to modify. Null-elements are assumed to be
	 *						already 0 and will not be set again (ie. (0,0)).
	 * @param tex_bounds	Boundary information to extract the coordinates from.
	 */
	virtual void set_tex_coords_transposed (float* buff, bounds_t const& tex_bounds) = 0;

	/**
	 * Resets the offsets previously modified by 'set_tex_coords' back to
	 * the defaults.
	 *
	 * @param buff			Buffer to modify. Null-elements are assumed to be
	 *						already 0 and will not be set again (ie. (0,0)).
	 */
	virtual void reset_tex_coord_offsets (float* buff) = 0;
};
