/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011


#include "extensions.h"

#include <stdexcept>
#include <cmath>
#include <cstdio>
#include <errno.h>

#include "../matrix.h"
#include "../engine_base.h"
#include "engine_gpu.h"



namespace gpu
{

	engine* environment::pengine = 0;

	float const engine::vertices [] = { 0, 0,   1, 0,   1, 1,   0, 1 };

	template <typename T>
	void check_output_rect (T const& t, int ox, int oy)
	{
#ifdef _DEBUG
		assert (t.width () >= ox &&
				t.height () >= oy);
#endif
	}

#ifdef WIN32
	void check_last_error ()
	{
		{
			int		err = GetLastError ();
			LPTSTR	s;
			
			if (err != 0)
			{
				if (FormatMessage ( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
								   0,
								   err,
								   0,
								   (LPTSTR) &s,
								   0,
								   0) == 0)
				{
					printf ("error: unknown %i (win32)\n", err);
				}
				else
				{
					printf ("error: %s (win32)\n", s);
					LocalFree (s);
				}
			}
		}
		
		{
			int err = glGetError ();
			if (err != GL_NO_ERROR)
			{
				printf ("error: %s (gl)\n", gluErrorString (err));
			}
		}
	}
	
	LRESULT CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
		return DefWindowProc (hwnd, msg, wParam, lParam);
	}

	bool engine::create_window ()
	{
		WNDCLASSEX	wcex;

		m_instance = GetModuleHandle (0);

		wcex.cbSize = sizeof (WNDCLASSEX); 

		wcex.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_DBLCLKS;
		wcex.lpfnWndProc	= (WNDPROC) WndProc;
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= m_instance;
		wcex.hIcon			= NULL;;
		wcex.hCursor		= LoadCursor (0, (LPCTSTR) IDC_ARROW);
		wcex.hbrBackground	= NULL;
		wcex.lpszMenuName	= NULL;
		wcex.lpszClassName	= "GPU_SIMX_CLASS";
		wcex.hIconSm		= NULL;

		if (!RegisterClassEx (&wcex))
		{
			check_last_error ();
			return false;
		}

		m_hwnd = CreateWindowEx (0,
								 "GPU_SIMX_CLASS",
								 "GPU Simplex Method",
								 WS_OVERLAPPEDWINDOW,
								 CW_USEDEFAULT,
								 0,
								 100,
								 100,
								 NULL,
								 NULL,
								 m_instance,
								 NULL);

		if (!m_hwnd)
		{
			check_last_error ();
			return false;
		}

		GLuint uiPxlFmt;
		static PIXELFORMATDESCRIPTOR pfd =	
		{
			sizeof(PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			32,
			8, 0, 8, 0, 8, 0, 8, 0, 0, 0, 0, 0, 0,
			16,
			0,
			0,
			PFD_MAIN_PLANE,
			0, 0, 0, 0
		};

		m_hdc = GetDC (m_hwnd);
		if (!m_hdc)	
		{
			check_last_error ();
			destroy_window ();
			return false;
		}

		uiPxlFmt = ChoosePixelFormat (m_hdc, &pfd);
		if (!uiPxlFmt || !SetPixelFormat (m_hdc, uiPxlFmt, &pfd))
		{
			check_last_error ();
			destroy_window ();
			return false;
		}
		else
		{
			DescribePixelFormat (m_hdc, uiPxlFmt, sizeof (pfd), &pfd);
		}

		m_hrc = wglCreateContext(m_hdc);
		if (!m_hrc || !wglMakeCurrent (m_hdc, m_hrc))
		{
			check_last_error ();
			destroy_window ();
			return false;
		}

		return true;
	}

	void engine::destroy_window ()
	{
		if (m_hrc)
		{
			if (!wglMakeCurrent (0, 0))
				check_last_error ();

			if (!wglDeleteContext (m_hrc))
				check_last_error ();

			m_hrc = 0;
		}

		if (m_hdc && !ReleaseDC (m_hwnd, m_hdc))
		{
			check_last_error ();
			m_hdc = 0;
		}

		if (m_hwnd && !DestroyWindow (m_hwnd))
		{
			check_last_error ();
			m_hwnd = 0;
		}

		if (!UnregisterClass ("GPU_SIMX_CLASS", m_instance))
		{
			check_last_error ();
			m_instance = 0;
		}
	}
#else
	void check_last_error ()
	{
		{
			int err = errno;
			errno = 0;
			if (err != 0)
			{
				printf ("error: %i (linux)\n", err);
			}
		}

		{
			int err = glGetError ();
			if (err != GL_NO_ERROR)
			{
				printf ("error: %s (gl)\n", gluErrorString (err));
			}
		}
	}
	

	bool engine::create_window ()
	{
		char const* dsp_env = getenv ("DISPLAY");
		if (!dsp_env || *dsp_env == '\0')
		{
			printf ("warning: DISPLAY not set, trying ':0'\n");
			dsp_env = ":0";
		}
		
		m_dsp = XOpenDisplay (dsp_env);
		int screen = DefaultScreen (m_dsp);

		errno = 0;
/*
		{
			int min, maj;
			XF86VidModeQueryVersion (m_dsp, &maj, &min);
			printf("XF86VidModeExtension-Version %d.%d\n", maj, min);
		}
*/		
		XVisualInfo* vi = 0;
		{
			int attrListSgl [] =
			{
				GLX_RGBA,
				GLX_RED_SIZE, 8,
				GLX_GREEN_SIZE, 8,
				GLX_BLUE_SIZE, 8,
				None
			};
			
			vi = glXChooseVisual (m_dsp, screen, attrListSgl);
			if (!vi)
			{
				printf ("error: no visual found.\n");
				check_last_error ();
				destroy_window ();
				return false;
			}
		}
/*
		{
			int min, maj;
			glXQueryVersion (m_dsp, &maj, &min);
			printf("glX-Version: %d.%d\n", maj, min);
		}
*/		
		m_ctx = glXCreateContext (m_dsp, vi, 0, GL_TRUE);
		if (!m_ctx)
		{
			printf ("error: unable to create context.");
			check_last_error ();
			destroy_window ();
			return false;
		}
		
		XSetWindowAttributes attr = {};
		attr.colormap = XCreateColormap (m_dsp, RootWindow (m_dsp, vi->screen), vi->visual, AllocNone);
		attr.event_mask = ExposureMask | KeyPressMask | ButtonPressMask;
		m_wnd = XCreateWindow (m_dsp,
							   RootWindow (m_dsp, vi->screen),
							   0,
							   0,
							   100,
							   100,
							   0,
							   vi->depth,
							   InputOutput,
							   vi->visual,
							   CWBorderPixel | CWColormap | CWEventMask,
							   &attr);
		if (!m_wnd)
		{
			printf ("error: unable to create window.");
			check_last_error ();
			destroy_window ();
			return false;
		}
		
		{
			Atom wmDelete = XInternAtom (m_dsp, "WM_DELETE_WINDOW", True);
			XSetWMProtocols (m_dsp, m_wnd, &wmDelete, 1);
			XSetStandardProperties (m_dsp, m_wnd, "GPU Simplex Method", "GPU Simplex Method", None, NULL, 0, NULL);
			XMapRaised (m_dsp, m_wnd);
		}
		
		if (!glXMakeCurrent (m_dsp, m_wnd, m_ctx))
		{
			printf ("error: unable to make context current.");
			check_last_error ();
			destroy_window ();
			return false;
		}
		
		if (!glXIsDirect(m_dsp, m_ctx))
		{
			printf ("error: no direct rendering available.");
			check_last_error ();
			destroy_window ();
			return false;
		}

		errno = 0;
		return true;    
	}
	
	void engine::destroy_window ()
	{
		if (!m_ctx)
		{
			if (!glXMakeCurrent ( m_dsp, None, NULL))
			{
				printf("Could not release drawing context.\n");
			}
			glXDestroyContext (m_dsp, m_ctx);
			m_ctx = NULL;
		}
		XCloseDisplay (m_dsp);
	}	
#endif

	engine::engine ()
	{
		environment::pengine = this;

		//	set up OpenGL
		if (!create_window ())
			throw std::runtime_error ("Unable to initialize OpenGL.");

		printf ("Renderer:     %s\n", glGetString (GL_RENDERER));
		printf ("Vendor:       %s\n", glGetString (GL_VENDOR));
		printf ("Version:      %s\n", glGetString (GL_VERSION));
		printf ("GLSL Version: %s\n", glGetString (GL_SHADING_LANGUAGE_VERSION));
		printf ("Extensions:   %s\n", glGetString (GL_EXTENSIONS));

		int max_tex_size = 0;
		glGetIntegerv (GL_MAX_TEXTURE_SIZE, &max_tex_size);
		printf ("Texture Size: %i\n", max_tex_size);

		glDisable (GL_LIGHTING);
		glDisable (GL_DEPTH_TEST);
		glDisable (GL_ALPHA_TEST);
		glDisable (GL_NORMALIZE);
		glPolygonMode (GL_FRONT, GL_FILL);
		glClearColor (0, 0, 0, 0);

		if (!frame_buffer::is_supported () ||
			!multitexture::is_supported () ||
			!gpu::environment::pixel_buffer::is_supported () ||
			!extensions::is_supported ("GL_ARB_texture_rectangle") ||
			!extensions::is_supported (EXT_texture_float))
		{
			printf ("Not supported: ");
			if (!frame_buffer::is_supported ())
				printf ("frame_buffer ");
			if (!multitexture::is_supported ())
				printf ("multitexture ");
			if (!gpu::environment::pixel_buffer::is_supported ())
				printf ("pixel_buffer ");
			if (!extensions::is_supported ("GL_ARB_texture_rectangle"))
				printf ("texture_rectangle ");
			if (!extensions::is_supported (EXT_texture_float))
				printf ("%s ", EXT_texture_float);
			printf ("\n");
			throw std::runtime_error ("GPU capabilities insufficient.");
		}
/*
		if (gpu_affinity::is_supported ())
		{
			printf ("Extension WGL_NV_gpu_affinity found:\n");

			std::vector <HGPUNV> hgpus (1);
			std::vector <GPU_DEVICE> gpu_devs (1);
			while (wglEnumGpusNV (hgpus.size () - 1, &hgpus.back ()))
				hgpus.push_back (HGPUNV());
			hgpus.pop_back ();

			printf ("  %i GPUs found\n", hgpus.size ());
			for (std::vector <HGPUNV>::iterator i = hgpus.begin (); i != hgpus.end (); ++ i)
			{
				gpu_devs.back ().cb = sizeof (GPU_DEVICE);
				wglEnumGpuDevicesNV (*i, 0, &gpu_devs.back ());
				gpu_devs.push_back (GPU_DEVICE ());
			}
		}
*/
		glEnableClientState (GL_VERTEX_ARRAY);
		glClientActiveTextureARB (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glClientActiveTextureARB (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);

		//	set up the buffers
		texCoords0 [0] =
		texCoords0 [1] =
		texCoords0 [3] =
		texCoords0 [6] =
		texCoords1 [0] =
		texCoords1 [1] =
		texCoords1 [3] =
		texCoords1 [6] = 0.0f;
		texCoords0 [2] =
		texCoords0 [4] =
		texCoords0 [5] =
		texCoords0 [7] =
		texCoords1 [2] =
		texCoords1 [4] =
		texCoords1 [5] =
		texCoords1 [7] = 1.0f;

		glVertexPointer (2, GL_FLOAT, 0, vertices);

		m_fbo = new frame_buffer ();
		m_fbo->activate ();
		m_fbo->set_buffers (GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT0_EXT);

		m_rb_res = new render_buffer (1, 1, GL_RGBA32F_ARB);

		m_pbo = new environment::pixel_buffer (4);

		//	set up once only
		glMatrixMode (GL_PROJECTION);
		glLoadIdentity ();
		gluOrtho2D (0, 1, 0, 1);	//	bottom/left is (0,0)

		glMatrixMode (GL_MODELVIEW);
		
		shader_manager = new shader::manager;
		if (shader_manager->find_config ())
			shader_manager->load_shaders ();
		else
			printf ("error: unable to find suitable shader configuration");
		
		glFlush ();
		check_last_error ();
	}

	engine::~engine ()
	{
		check_last_error ();

		delete m_rb_res;

		m_fbo->deactivate ();
		delete m_fbo;
		check_last_error ();

		delete shader_manager;
		check_last_error ();

		destroy_window ();
	}

	engine& engine::instance ()
	{
		if (environment::pengine)
			return *environment::pengine;

		try
		{
			static engine e;
			return e;
		}
		catch (std::runtime_error const& e)
		{
			printf ("fatal: %s\n", e.what ());
			throw;
		}
	}

	void engine::precache_matrix (matrix const& cache)
	{
		render_buffer rb (cache.bounds.width, cache.bounds.height, cache.texture->data.internal_format);

		shader& s = shader_manager->get_shader (shader::PASSTHRU);
		s.bind ();
		s.set_parameter ("tex", cache.texture->data);

		set_tex_coords (texCoords0, cache.bounds);

		glClientActiveTextureARB (GL_TEXTURE0);
		glTexCoordPointer (2, GL_FLOAT, 0, texCoords0);

		set_viewport (0, 0, 1, 1);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, rb);
		glDrawArrays (GL_QUADS, 0, 4);
		glReadPixels (0, 0, 1, 1, GL_RGBA, GL_FLOAT, m_rd_buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);

	}

	void engine::precache_run ()
	{
		set_tex_coords (texCoords0, 1, 1);
		set_tex_coords (texCoords1, 1, 1);
		render_arrays_fetch ();
	}

	void engine::zero (texture& tex)
	{
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, tex);
		glClearColor (0, 0, 0, 0);
		glClear (GL_COLOR_BUFFER_BIT);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::zero (texture& tex, int x, int y, int w, int h)
	{
		glEnable (GL_SCISSOR_TEST);
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, tex);
		glClearColor (0, 0, 0, 0);
		clear_region (x, y, w, h);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
		glDisable (GL_SCISSOR_TEST);
	}

	void engine::get_element (texture const& tex, int x, int y, float* buff)
	{
		//	const_cast, because we do not semantically alter the texture, we just read from it
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, const_cast <texture&> (tex));
		glReadPixels (x, y, 1, 1, GL_RED, GL_FLOAT, buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::get_elements (texture const& tex, int x, int y, int w, int h, float* buff)
	{
		//	const_cast, because we do not semantically alter the texture, we just read from it
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, const_cast <texture&> (tex));
		glReadPixels (x, y, w, h, GL_RED, GL_FLOAT, buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}
#ifdef USE_1CHN_TEXTURE
	void engine::get_elements (texture4f const& tex, int x, int y, int w, int h, float* buff)
	{
		//	const_cast, because we do not semantically alter the texture, we just read from it
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, const_cast <texture4f&> (tex));
		glReadPixels (x, y, w, h, GL_RGBA, GL_FLOAT, buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}
#endif
	void engine::get_elements (texture4f const& tex, int chn, int x, int y, int w, int h, float* buff)
	{
		//	const_cast, because we do not semantically alter the texture, we just read from it
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, const_cast <texture4f&> (tex));
		glReadPixels (x, y, w, h, chn, GL_FLOAT, buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::set_element (texture& res, float value, int x, int y)
	{
		m_pbo->activate ();
		float* tmp = m_pbo->get_buffer ();
		*tmp = value;
		m_pbo->release_buffer ();
		res.fill_rect (x, y, 1, 1, GL_RED, *m_pbo);
		m_pbo->deactivate ();
/*
		glEnable (GL_SCISSOR_TEST);
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, res);
		m_fbo->status ();

//		y += res.bounds.yoffset;
		glClearColor (value, value, value, value);

		//if (tex.channelUsage == texture.Usage.COMPACT)
		//{
		//	final int chn = x % 4;
		//	x /= 4;

		//	switch (chn)
		//	{
		//		case 0: glColorMask (true, false, false, false); break;
		//		case 1: glColorMask (false, true, false, false); break;
		//		case 2: glColorMask (false, false, true, false); break;
		//		case 3: glColorMask (false, false, false, true); break;
		//	}
		//}

		clear_region (x, y, 1, 1);

		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
		glDisable (GL_SCISSOR_TEST);
		//glColorMask (true, true, true, true);*/
	}

	void engine::set_elements (texture& res, float value, int x, int y, int w, int h)
	{
		glEnable (GL_SCISSOR_TEST);
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, res);
		m_fbo->status ();

//		y += res.bounds.yoffset;
		glClearColor (value, value, value, value);

		//if (tex.channelUsage == texture.Usage.COMPACT)
		//{
		//	final int chn1 = x % 4;
		//	final int chn2 = (x + w) % 4;
		//	if (chn1 > 0)
		//	{
		//		switch (chn1)
		//		{
		//			case 1: glColorMask (false, true, true, true); break;
		//			case 2: glColorMask (false, false, true, true); break;
		//			case 3: glColorMask (false, false, false, true); break;
		//		}
		//		clear_region (x / 4, y, 1);
		//	}

		//	clear_region ((int) (x / 4.0 + 0.5) , y, w / 4, 1);

		//	if (chn2 > 0)
		//	{
		//		switch (chn2)
		//		{
		//			case 1: glColorMask (true, false, false, false); break;
		//			case 2: glColorMask (true, true, false, false); break;
		//			case 3: glColorMask (true, true, true, false); break;
		//		}
		//		clear_region ((x + w) / 4, y, 1, 1);
		//	}

		//	glColorMask (true, true, true, true);
		//}
		//else
			clear_region (x, y, w, h);

		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
		glDisable (GL_SCISSOR_TEST);
	}

	void engine::find_leaving_bounded_reduce (matrix const& lhs, matrix const& rhs, matrix const& chi,
											  matrix const& x, float enter, matrix4f pingpong [2], float maxDiff)
	{
		bool const wantPositive = maxDiff > 0;

		//	init phase
		pingpong [0].bounds.yoffset = 0;
		pingpong [0].bounds.width = lhs.bounds.tex_height ();
		pingpong [0].bounds.height = 1;
		pingpong [1].bounds.yoffset = 0;
		pingpong [1].bounds.height = 1;

		shader& s = shader_manager->get_shader (wantPositive ?
												shader::REDUCE_INIT_LEAVEBND_POS :
												shader::REDUCE_INIT_LEAVEBND_NEG);
		s.bind ();
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("chi", chi.texture->data);
		s.set_parameter ("ux", x.texture->data);
		set_tex_coords (texCoords0, lhs.bounds.tex_height (), lhs.bounds.tex_width ());

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong [0].texture->data);
		m_fbo->attach (GL_COLOR_ATTACHMENT1_EXT, pingpong [1].texture->data);
		render_arrays_raw (0, 0, lhs.bounds.tex_height (), 1);

		if (wantPositive)
		{
			int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MINY_SQW), pingpong, 999999.0);
			lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MINY_REST_BND), maxDiff, pingpong [read]);
		}
		else
		{
			int const read = lhs_rhs_op_reduce_step (shader_manager->get_shader (shader::REDUCE_MAXY_SQW), pingpong, -999999.0);
			lhs_rhs_op_reduce_final (shader_manager->get_shader (shader::REDUCE_MAXY_REST_BND), maxDiff, pingpong [read]);
		}
	}

	struct rgba { float x,y,z,_w; };
	void engine::lhs_rhs_op_reduce_init (shader& s, matrix const& mat, int DIM, matrix4f pingpong [2])
	{
		pingpong [0].bounds.yoffset = 0;
		pingpong [0].bounds.width = DIM;
		pingpong [0].bounds.height = 1;
		pingpong [1].bounds.yoffset = 0;
		pingpong [1].bounds.height = 1;

		s.bind ();
		s.set_parameter ("mat", mat.texture->data);
		set_tex_coords (texCoords0, DIM, 1);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong [0].texture->data);
		m_fbo->attach (GL_COLOR_ATTACHMENT1_EXT, pingpong [1].texture->data);
		render_arrays_raw (0, 0, DIM, 1);
	}

	void engine::lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, int DIM, matrix4f pingpong [2])
	{
		pingpong [0].bounds.yoffset = 0;
		pingpong [0].bounds.width = DIM;
		pingpong [0].bounds.height = 1;
		pingpong [1].bounds.yoffset = 0;
		pingpong [1].bounds.height = 1;

		s.bind ();
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		set_tex_coords (texCoords0, DIM, 1);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong [0].texture->data);
		m_fbo->attach (GL_COLOR_ATTACHMENT1_EXT, pingpong [1].texture->data);
		render_arrays_raw (0, 0, DIM, 1);

#ifdef _DEBUG
		std::vector <rgba> tmp1 (DIM);
		glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
		glReadPixels (0, 0, DIM, 1, GL_RGBA, GL_FLOAT, &tmp1.front ());
#endif
	}

	void engine::lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, matrix const& gamma, int DIM, matrix4f pingpong [2])
	{
		pingpong [0].bounds.yoffset = 0;
		pingpong [0].bounds.width = DIM;
		pingpong [0].bounds.height = 1;
		pingpong [1].bounds.yoffset = 0;
		pingpong [1].bounds.height = 1;

		s.bind ();
		s.set_parameter ("lhs", lhs.texture->data);
		s.set_parameter ("rhs", rhs.texture->data);
		s.set_parameter ("gamma", gamma.texture->data);
		set_tex_coords (texCoords0, DIM, 1);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong [0].texture->data);
		m_fbo->attach (GL_COLOR_ATTACHMENT1_EXT, pingpong [1].texture->data);
		render_arrays_raw (0, 0, DIM, 1);
	}

	int engine::lhs_rhs_op_reduce_step (shader& s, matrix4f pingpong [2], float ydummy)
	{
		int const	MIN_PINGPONG_SIZE = 8;
		int			w = pingpong [0].bounds.width;
		int			read = 0;
		int			write = 1;

#ifdef _DEBUG
		std::vector <rgba> tmp1 (w);
		glReadBuffer (GL_COLOR_ATTACHMENT0_EXT + read);
		glReadPixels (0, 0, w, 1, GL_RGBA, GL_FLOAT, &tmp1.front ());
#endif

		s.bind ();
		while (w >= MIN_PINGPONG_SIZE * 2)
		{
			if (w % 2 == 1)
			{
				//	insert dummy element which does
				//	not change our solution
				glEnable (GL_SCISSOR_TEST);
				glClearColor (-1, ydummy, 0, 0);
				clear_region (w, 0, 1, 1);
				glDisable (GL_SCISSOR_TEST);

				++ w;

#ifdef _DEBUG
				std::vector <rgba> tmp2 (w);
				glReadBuffer (GL_COLOR_ATTACHMENT0_EXT + read);
				glReadPixels (0, 0, w, 1, GL_RGBA, GL_FLOAT, &tmp2.front ());
				glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
#endif
			}

			s.set_parameter ("tex", pingpong [read].texture->data);
			set_tex_coords (texCoords0, w, 1);

			int const new_w = w / 2;

			glDrawBuffer (GL_COLOR_ATTACHMENT0_EXT + write);
			render_arrays_raw (0, 0, new_w, 1);

			w = new_w;
			std::swap (read, write);

#ifdef _DEBUG
			std::vector <rgba> tmp2 (w);
			glDrawBuffer (GL_COLOR_ATTACHMENT0_EXT + write);
			glReadBuffer (GL_COLOR_ATTACHMENT0_EXT + read);
			glReadPixels (0, 0, w, 1, GL_RGBA, GL_FLOAT, &tmp2.front ());
			glDrawBuffer (GL_COLOR_ATTACHMENT0_EXT + read);
			glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
#endif
		}
		m_fbo->detach (GL_COLOR_ATTACHMENT1_EXT);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
		glDrawBuffer (GL_COLOR_ATTACHMENT0_EXT);
		pingpong [read].bounds.width = w;

#ifdef _DEBUG
		std::vector <rgba> tmp2 (8);
		glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong [read].texture->data);
		glReadPixels (0, 0, 8, 1, GL_RGBA, GL_FLOAT, &tmp2.front ());
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
#endif

		return read;
	}

	void engine::lhs_rhs_op_reduce_final (shader& s, matrix4f const& pingpong)
	{
		s.set_special_parameter ("DIM", pingpong.bounds.tex_width ());
		s.set_parameter ("tex", pingpong.texture->data);
		render_arrays_fetch ();
	}

	void engine::lhs_rhs_op_reduce_final (shader& s, float maxDiff, matrix4f const& pingpong)
	{
#ifdef _DEBUG
		std::vector <rgba> tmp2 (13);
		glReadBuffer (GL_COLOR_ATTACHMENT0_EXT);
		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, pingpong.texture->data);
		glReadPixels (0, 0, 13, 1, GL_RGBA, GL_FLOAT, &tmp2.front ());
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
#endif

		s.set_special_parameter ("DIM", pingpong.bounds.tex_width ());
		s.set_parameter ("tex", pingpong.texture->data);
		s.set_parameter ("max_diff", maxDiff);
		render_arrays_fetch ();
	}



	//	public void compact (texture tex, bounds_t const& tex_bounds,
	//						 texture res, bounds_t const& res_bounds)
	//	{
	//		assert (tex.channelUsage == texture.Usage.REGULAR &&
	//				res.channelUsage == texture.Usage.COMPACT);
	//
	//		res_bounds.assign (tex_bounds);
	//		res_bounds.updateRealBounds (res.channelUsage);
	//
	//		shader& s = shader_manager->get_shader (shader::COMPACT, texture.Usage.REGULAR);
	//		s.bind ();
	//		s.set_parameter ("tex", tex);
	//
	//		set_tex_coords (texCoords0, res_bounds);
	//		render_arrays (res, res_bounds);
	//	}


	void engine::render_arrays_raw (int x, int y, int w, int h)
	{
		glClientActiveTextureARB (GL_TEXTURE1);
		glTexCoordPointer (2, GL_FLOAT, 0, texCoords1);
		glClientActiveTextureARB (GL_TEXTURE0);
		glTexCoordPointer (2, GL_FLOAT, 0, texCoords0);

		set_viewport (x, y, w, h);
		glDrawArrays (GL_QUADS, 0, 4);
	//	glFinish ();
	}

	void engine::render_arrays (matrix& res)
	{
		int const x = 0;
		int const y = res.bounds.yoffset;
		int const w = res.bounds.tex_width ();
		int const h = res.bounds.tex_height ();

		check_output_rect (res.texture->data, x + w, y + h);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, res.texture->data);
		render_arrays_raw (x, y, w, h);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::render_arrays (texture& res, int x, int y, int w, int h)
	{
		check_output_rect (res, x + w, y + h);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, res);
		render_arrays_raw (x, y, w, h);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::render_arrays_fetch ()
	{
		glClientActiveTextureARB (GL_TEXTURE1);
		glTexCoordPointer (2, GL_FLOAT, 0, texCoords1);
		glClientActiveTextureARB (GL_TEXTURE0);
		glTexCoordPointer (2, GL_FLOAT, 0, texCoords0);

		set_viewport (0, 0, 1, 1);

		m_fbo->attach (GL_COLOR_ATTACHMENT0_EXT, *m_rb_res);
		glDrawArrays (GL_QUADS, 0, 4);
		glReadPixels (0, 0, 1, 1, GL_RGBA, GL_FLOAT, m_rd_buff);
		m_fbo->detach (GL_COLOR_ATTACHMENT0_EXT);
	}

	void engine::set_tex_coords (float* buff, int w, int h)
	{
		buff [2] = (float) w;
		buff [4] = (float) w;
		buff [5] = (float) h;
		buff [7] = (float) h;
	}

	void engine::set_tex_coords (float* buff, int x, int y, int w, int h)
	{
		buff [0] =
		buff [6] = (float) x;

		buff [1] =
		buff [3] = (float) y;

		buff [2] =
		buff [4] = (float) (x + w);

		buff [5] =
		buff [7] = (float) (y + h);
	}

	void engine::set_tex_coords (float* buff, bounds_t const& tex_bounds)
	{
//		buff [0] =
//		buff [6] = xoffset;

		buff [1] =
		buff [3] = (float) tex_bounds.yoffset;

		buff [2] =
		buff [4] = (float) tex_bounds.tex_width ();	//	+ xoffset

		buff [5] =
		buff [7] = (float) tex_bounds.tex_height () + tex_bounds.yoffset;
	}

	void engine::set_tex_coords_transposed (float* buff, bounds_t const& tex_bounds)
	{
		buff [1] =
		buff [3] = (float) tex_bounds.yoffset;

		buff [2] =
		buff [4] = (float) tex_bounds.tex_height () + tex_bounds.yoffset;

		buff [5] =
		buff [7] = (float) tex_bounds.tex_width ();
	}

	void engine::reset_tex_coord_offsets (float* buff)
	{
		//	reset x-offset
		buff [0] =
		buff [6] = 0.0f;

		//	reset y-offset
		buff [1] =
		buff [3] = 0.0f;
	}

	void engine::clear_region (int x, int y, int w, int h)
	{
		set_viewport (x, y, w, h);
		glScissor (x, y, w, h);
		glClear (GL_COLOR_BUFFER_BIT);
	}

	void engine::set_viewport (int x, int y, int w, int h)
	{
		glLoadIdentity ();
		glViewport (x, y, w, h);
	}

};
