/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011


#pragma once

#ifndef WIN32
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/glx.h>
#endif

#include "../bounds.h"
#include "settings.h"
#include "shader_gpu.h"

namespace gpu
{

	class engine;

	struct environment
	{
		typedef float				value_type;
		typedef gpu::engine			engine;
		typedef gpu::shader			shader;
		typedef gpu::texture_rect1f	texture;
		typedef gpu::texture_rect4f	texture4f;

		typedef ::matrix <environment, texture>		matrix;
		typedef ::matrix <environment, texture4f>	matrix4f;

//	HACK: PBO do not work properly with RGBA float textures. They do
//		  work when uploading full RGBA pixels to VRAM, but you cannot
//		  specify to upload only the red channel. An option may be to
//		  upload the data into RGBA and use shaders to pull the data
//		  apart. Also, using PBOs *will* be an option again when using
//		  compact texture representations!
#if defined(USE_1CHN_TEXTURE) && defined(USE_PIXEL_BUFFER)
		typedef ::pixel_buffer <float, usage::write_once> pixel_buffer;
#else
		struct pixel_buffer
		{
			pixel_buffer (size_t sz) : m_data (sz) {}
			operator float const* () const { return &m_data.front (); }
			void resize (size_t sz) { if (m_data.size () < sz) m_data.resize (sz); }
			float* get_buffer () { return &m_data.front (); }
			void release_buffer () {}
			void activate () {}
			void deactivate () {}

			void* pointer (int offset) { return (float*) &m_data.front () + offset; }
			static bool is_supported () { return true; }

		private:
			std::vector <float>	m_data;
		};
#endif
		static engine* pengine;
	};

	/**
	 *
	 * @author Jakob Bieling
	 */
	class engine : public engine_base <environment>
	{
		frame_buffer*				m_fbo;
		render_buffer*				m_rb_res;
		environment::pixel_buffer*	m_pbo;

#ifdef WIN32
		HINSTANCE				m_instance;
		HWND					m_hwnd;
		HDC						m_hdc;
		HGLRC 					m_hrc;
#else
		Display*				m_dsp;
		Window					m_wnd;
		GLXContext				m_ctx;
#endif
		static float const		vertices [];

		bool create_window ();
		void destroy_window ();

		engine ();

		engine (engine const&);
		engine& operator= (engine const&);

	public:

		~engine ();

		static engine& instance ();

		float get_last_result (int chn)
		{
			return m_rd_buff [chn];
		}

		void precache_matrix (matrix const& res);
		void precache_run ();


		void zero (texture& tex);
		void zero (texture& tex, int x, int y, int w, int h);

		void get_element (texture const& tex, int x, int y, float* buff);
		void get_elements (texture const& tex, int x, int y, int w, int h, float* buff);
#ifdef USE_1CHN_TEXTURE
		void get_elements (texture4f const& tex, int x, int y, int w, int h, float* buff);
#endif
		void get_elements (texture4f const& tex, int chn, int x, int y, int w, int h, float* buff);

		void set_element (texture& res, float value, int x, int y);
		void set_elements (texture& res, float value, int x, int y, int w, int h);

		void find_leaving_bounded_reduce (matrix const& lhs, matrix const& rhs, matrix const& chi,
										  matrix const& ux, float enter, matrix4f pingpong [2], float maxDiff);

	private:

		void lhs_rhs_op_reduce_init (shader& s, matrix const& mat, int DIM, matrix4f pingpong [2]);
		void lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, int DIM, matrix4f pingpong [2]);
		void lhs_rhs_op_reduce_init (shader& s, matrix const& lhs, matrix const& rhs, matrix const& gamma, int DIM, matrix4f pingpong [2]);
		int  lhs_rhs_op_reduce_step (shader& s, matrix4f pingpong [2], float ydummy);
		void lhs_rhs_op_reduce_final (shader& s, matrix4f const& pingpong);
		void lhs_rhs_op_reduce_final (shader& s, float maxDiff, matrix4f const& pingpong);

		void render_arrays_raw (int x, int y, int w, int h);
		
		void render_arrays (matrix& res);
		void render_arrays (texture& res, int x, int y, int w, int h);
		void render_arrays_fetch ();

		void set_tex_coords (float* buff, int w, int h);
		void set_tex_coords (float* buff, int x, int y, int w, int h);

		void set_tex_coords (float* buff, bounds_t const& tex_bounds);
		void set_tex_coords_transposed (float* buff, bounds_t const& tex_bounds);
		void reset_tex_coord_offsets (float* buff);

		/**
		 * Clear a specified rectangle of the frame buffer. The respective portion
		 * of the bound texture will be set to the values previously specified in
		 * a call to glClearColor.
		 *
		 * @param x		x-offset of the recctangle.
		 * @param y		y-offset of the recctangle.
		 * @param w		Width of the recctangle.
		 * @param h		Height of the recctangle.
		 */
		void clear_region (int x, int y, int w, int h);

		/**
		 * Sets viewport of render target
		 *
		 * @param x x-coordinate of the viewport origin.
		 * @param y y-coordinate of the viewport origin.
		 * @param w Viewport width in pixels.
		 * @param h Viewport height in pixels.
		 */
		void set_viewport (int x, int y, int w, int h);
	};

};
