/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#include <algorithm>

#include "frame_buffer.h"

#ifndef __APPLE__
static PFNGLISRENDERBUFFEREXTPROC						glIsRenderbufferEXT;
static PFNGLBINDRENDERBUFFEREXTPROC						glBindRenderbufferEXT;
static PFNGLDELETERENDERBUFFERSEXTPROC					glDeleteRenderbuffersEXT;
static PFNGLGENRENDERBUFFERSEXTPROC						glGenRenderbuffersEXT;
static PFNGLRENDERBUFFERSTORAGEEXTPROC					glRenderbufferStorageEXT;
static PFNGLGETRENDERBUFFERPARAMETERIVEXTPROC			glGetRenderbufferParameterivEXT;

static PFNGLISFRAMEBUFFEREXTPROC						glIsFramebufferEXT;
static PFNGLBINDFRAMEBUFFEREXTPROC						glBindFramebufferEXT;
static PFNGLDELETEFRAMEBUFFERSEXTPROC					glDeleteFramebuffersEXT;
static PFNGLGENFRAMEBUFFERSEXTPROC						glGenFramebuffersEXT;
static PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC				glCheckFramebufferStatusEXT;

static PFNGLFRAMEBUFFERTEXTURE1DEXTPROC					glFramebufferTexture1DEXT;
static PFNGLFRAMEBUFFERTEXTURE2DEXTPROC					glFramebufferTexture2DEXT;
static PFNGLFRAMEBUFFERTEXTURE3DEXTPROC					glFramebufferTexture3DEXT;
static PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC				glFramebufferRenderbufferEXT;
static PFNGLGETFRAMEBUFFERATTACHMENTPARAMETERIVEXTPROC	glGetFramebufferAttachmentParameterivEXT;
static PFNGLGENERATEMIPMAPEXTPROC						glGenerateMipmapEXT;
#endif



#ifdef _DEBUG
#define check_error()	check_error_dbg (false, __LINE__, __FILE__)

#include <cassert>
#include <cstdio>

static void check_error_dbg (bool doassert, int line, char const* file)
{
	GLenum e = glGetError ();

	if (e != GL_NO_ERROR)
	{
		char tmp [2048];

		sprintf (tmp, "%s(%i): ", file, line);

		switch (e)
		{
			case GL_INVALID_ENUM: strcat (tmp, "GL_INVALID_ENUM\n"); break;
			case GL_INVALID_VALUE: strcat (tmp, "GL_INVALID_VALUE\n"); break;
			case GL_INVALID_OPERATION: strcat (tmp, "GL_INVALID_OPERATION\n"); break;
			case GL_STACK_OVERFLOW: strcat (tmp, "GL_STACK_OVERFLOW\n"); break;
			case GL_STACK_UNDERFLOW: strcat (tmp, "GL_STACK_UNDERFLOW\n"); break;
			case GL_OUT_OF_MEMORY : strcat (tmp, "GL_OUT_OF_MEMORY\n"); break;
			default: strcat (tmp, "unkown error\n");
		}

#ifdef WIN32
		OutputDebugString (tmp);
#else
		printf ("%s", tmp);
#endif
		if (doassert)
			assert (false);
	}
}
#else
void check_error () {}
#endif




render_buffer::render_buffer (int width, int height, int internalFormat)
	:	m_width (width),
		m_height (height)
{
	glGenRenderbuffersEXT (1, &m_rb);

	bind ();
	glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, internalFormat, width, height);
	check_error ();
	unbind ();
}

render_buffer::~render_buffer ()
{
	glDeleteRenderbuffersEXT (1, &m_rb);
}

void render_buffer::bind () const
{
	glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, m_rb);
}

void render_buffer::unbind () const
{
	glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, 0);
}





frame_buffer::frame_buffer ()
{
	glGenFramebuffersEXT (1, &m_fbo);
	check_error ();
}

frame_buffer::~frame_buffer ()
{
	glDeleteFramebuffersEXT (1, &m_fbo);
}

void frame_buffer::attach (int attachment_point, render_buffer& rb)
{
	glFramebufferRenderbufferEXT (GL_FRAMEBUFFER_EXT, attachment_point, GL_RENDERBUFFER_EXT, rb.handle ());
	check_error ();
}

void frame_buffer::attach (int attachment_point, int target, int handle)
{
	glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, attachment_point, target, handle, 0);
	check_error ();
}

void frame_buffer::detach (int attachment_point)
{
	//	The target may be GL_TEXTURE_2D, regardless of the initial target, because
	//	we actually re-assign the attachment point and assign 'nothing' to it.

	glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, attachment_point, GL_TEXTURE_2D, 0, 0);
	check_error ();
}

void frame_buffer::set_buffers (GLenum draw_buffer, GLenum read_buffer)
{
	glDrawBuffer (draw_buffer);
	glReadBuffer (read_buffer); 
	check_error ();
}

int frame_buffer::status () const
{
#ifdef _DEBUG
	GLint	err = glCheckFramebufferStatusEXT (GL_FRAMEBUFFER_EXT);
	char	tmp [128];

	strcpy (tmp, "Framebuffer Status: ");

	switch (err)
	{
		case GL_FRAMEBUFFER_COMPLETE_EXT:						return err;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:			strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT\n"); break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:	strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT\n"); break;
		case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:			strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT\n"); break;
		case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:				strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT\n"); break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:			strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT\n"); break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:			strcat (tmp, "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT\n"); break;
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:					strcat (tmp, "GL_FRAMEBUFFER_UNSUPPORTED_EXT\n"); break;
		default:												strcat (tmp, "(unknown)\n"); break;
	}

#ifdef WIN32
	OutputDebugString (tmp);
#else
	printf ("%s", tmp);
#endif
	return err;
#else
	return glCheckFramebufferStatusEXT (GL_FRAMEBUFFER_EXT);
#endif
}

void frame_buffer::activate ()
{
	glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, m_fbo);
}

void frame_buffer::deactivate ()
{
	glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
}

bool frame_buffer::is_supported ()
{
	static bool is_supp = false,
				is_checked = false;

	if (!is_checked)
	{
		if (extensions::is_supported ("GL_EXT_framebuffer_object"))
		{
#ifndef __APPLE__
			glIsRenderbufferEXT = (PFNGLISRENDERBUFFEREXTPROC) wglXGetProcAddress ("glIsRenderbufferEXT");
			glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC) wglXGetProcAddress ("glBindRenderbufferEXT");
			glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC) wglXGetProcAddress ("glDeleteRenderbuffersEXT");
			glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC) wglXGetProcAddress ("glGenRenderbuffersEXT");
			glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC) wglXGetProcAddress ("glRenderbufferStorageEXT");
			glGetRenderbufferParameterivEXT = (PFNGLGETRENDERBUFFERPARAMETERIVEXTPROC) wglXGetProcAddress ("glGetRenderbufferParameterivEXT");
			glIsFramebufferEXT = (PFNGLISFRAMEBUFFEREXTPROC) wglXGetProcAddress ("glIsFramebufferEXT");
			glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC) wglXGetProcAddress ("glBindFramebufferEXT");
			glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC) wglXGetProcAddress ("glDeleteFramebuffersEXT");
			glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC) wglXGetProcAddress ("glGenFramebuffersEXT");
			glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC) wglXGetProcAddress ("glCheckFramebufferStatusEXT");
			glFramebufferTexture1DEXT = (PFNGLFRAMEBUFFERTEXTURE1DEXTPROC) wglXGetProcAddress ("glFramebufferTexture1DEXT");
			glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC) wglXGetProcAddress ("glFramebufferTexture2DEXT");
			glFramebufferTexture3DEXT = (PFNGLFRAMEBUFFERTEXTURE3DEXTPROC) wglXGetProcAddress ("glFramebufferTexture3DEXT");
			glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC) wglXGetProcAddress ("glFramebufferRenderbufferEXT");
			glGetFramebufferAttachmentParameterivEXT = (PFNGLGETFRAMEBUFFERATTACHMENTPARAMETERIVEXTPROC) wglXGetProcAddress ("glGetFramebufferAttachmentParameterivEXT");
			glGenerateMipmapEXT = (PFNGLGENERATEMIPMAPEXTPROC) wglXGetProcAddress ("glGenerateMipmapEXT");
#endif
			is_supp =	glIsRenderbufferEXT &&
						glBindRenderbufferEXT &&
						glDeleteRenderbuffersEXT &&
						glGenRenderbuffersEXT &&
						glRenderbufferStorageEXT &&
						glGetRenderbufferParameterivEXT &&
						glIsFramebufferEXT &&
						glBindFramebufferEXT &&
						glDeleteFramebuffersEXT &&
						glGenFramebuffersEXT &&
						glCheckFramebufferStatusEXT &&
						glFramebufferTexture1DEXT &&
						glFramebufferTexture2DEXT &&
						glFramebufferTexture3DEXT &&
						glFramebufferRenderbufferEXT &&
						glGetFramebufferAttachmentParameterivEXT &&
						glGenerateMipmapEXT;
		}
		is_checked = true;
	}

	return is_supp;
}
