/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#pragma once

#include "extensions.h"
#include "multitexture.h"
#include "pixel_buffer.h"

#include <algorithm>
#include <cassert>



template <typename T> struct texture_type { };
template <> struct texture_type <unsigned char> { static int const name = GL_UNSIGNED_BYTE; };
template <> struct texture_type <unsigned short> { static int const name = GL_UNSIGNED_SHORT; };
template <> struct texture_type <unsigned int> { static int const name = GL_UNSIGNED_INT; };
template <> struct texture_type <char> { static int const name = GL_BYTE; };
template <> struct texture_type <short> { static int const name = GL_BYTE; };
template <> struct texture_type <int> { static int const name = GL_INT; };
template <> struct texture_type <float> { static int const name = GL_FLOAT; };

template <typename T> struct conversion
{
	typedef T converted_type;
	static converted_type* convert (T const* t,size_t) { return 0; }
	static void free (converted_type*) { }
};

template <> struct conversion <double>
{
	typedef float converted_type;
	static converted_type* convert (double const* ptr, size_t sz)
	{
		converted_type* tmp = new converted_type [sz];
		for (size_t i = 0; i < sz; ++ i)
			tmp [i] = static_cast <converted_type> (ptr [i]);
		return tmp;
	}
	
	static void free (converted_type* mem)
	{
		delete [] mem;
	}
};

template <GLenum target_, GLenum format_, GLenum internalFormat_, typename value_type_>
class texture
{
	static int _components (int format)
	{
		switch (format)	
		{
			case GL_RED: return 1;
			case GL_GREEN: return 1;
			case GL_BLUE: return 1;
			case GL_ALPHA: return 1;
			case GL_RGB: return 3;
			case GL_RGBA: return 4;
			case GL_BGR_EXT: return 3;
			case GL_BGRA_EXT: return 4;
			case GL_LUMINANCE: return 1;
			case GL_LUMINANCE_ALPHA: return 2;
		}

		assert ("Unknown format." && false);
		return 1;
	}


public:

	static int const	components;
	static GLenum const target = target_;
	static GLenum const format = format_;
	static GLenum const internal_format = internalFormat_;
	typedef value_type_ value_type;

private:

	GLuint	m_texid;
	bool	m_valid;

	int		m_width,
			m_height;

	bool common_load (size_t w, size_t h, GLint f)
	{
		glTexParameteri (target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri (target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri (target, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri (target, GL_TEXTURE_WRAP_T, GL_CLAMP);

		unbind ();
		m_valid = glGetError () == GL_NO_ERROR;
		m_width = static_cast <int> (w);
		m_height = static_cast <int> (h);

		return m_valid;
	}

	texture (texture const&);
	texture& operator= (texture const&);

public:

	texture ()
		:	m_valid (false),
			m_width (-1),
			m_height (-1)
	{
		glGenTextures (1, &m_texid);
	}

	~texture ()
	{
		glDeleteTextures (1, &m_texid);
	}

	static void enable_multi (int i) { glActiveTextureARB (GL_TEXTURE0_ARB + i); glEnable (target); }
	static void disable_multi (int i) { glActiveTextureARB (GL_TEXTURE0_ARB + i); glDisable (target); }

	static void enable () { glEnable (target); }
	static void disable () { glDisable (target); }

	bool create (int width, int height)
	{
//		if (m_valid)
//			return false;

		bind ();
		glTexImage2D (target, 0, internal_format, width, height, 0, format, texture_type <value_type_>::name, 0);

		glTexParameteri (target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri (target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri (target, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri (target, GL_TEXTURE_WRAP_T, GL_CLAMP);

//		unbind ();
		m_valid = glGetError () == GL_NO_ERROR;
		m_width = width;
		m_height = height;

		return m_valid;
	}

	template <typename T>
	bool load (size_t w, size_t h, T const* data, int f = format)	//	'format' is 'GL_RED' or 'GL_RGBA' etc
	{
		bind ();
		typename conversion <T>::converted_type* const tmp = conversion <T>::convert (data, w * h);
		glTexImage2D (target, 0, internal_format, (int) w, (int) h, 0, format, texture_type <typename conversion <T>::converted_type>::name, (void*) (tmp ? tmp : data));
		conversion <T>::free (tmp);

		return common_load (w, h, format);
	}

	void	bind_multi (int i) const { glActiveTextureARB (GL_TEXTURE0_ARB + i); glBindTexture (target, m_texid); }
	void	unbind_multi (int i) const { glActiveTextureARB (GL_TEXTURE0_ARB + i); glBindTexture (target, 0); }

	void	bind () const { glBindTexture (target, m_texid); }
	void	unbind () const { glBindTexture (target, 0); }

	int		width () const { return m_width; }
	int		height () const { return m_height; }

	GLuint	handle () const { return m_texid; }

	void	set_texel (float const* fv, int x, int y) const
	{
		bind ();
		//	use optimization here?!
		glTexSubImage2D (target, 0, x, y, 1, 1, format, texture_type <value_type>::name, fv);
	}

	void	fill_rect (int x, int y, int w, int h, int f, value_type const* data)
	{
		bind ();
		glTexSubImage2D (target, 0, x, y, w, h, f, texture_type <value_type>::name, data);
	}

	void	fill_rect (int x, int y, int w, int h, int f, pixel_buffer <value_type, usage::write_once> const& data)
	{
		bind ();
		glTexSubImage2D (target, 0, x, y, w, h, f, texture_type <value_type>::name, data.pointer (0));
	}
};

template <GLenum target_, GLenum format_, GLenum internalFormat_, typename value_type_>
int const texture <target_, format_, internalFormat_, value_type_>::components = texture <target_, format_, internalFormat_, value_type_>::_components (format_);


class render_buffer
{
	GLuint	m_rb;

	int		m_width,
			m_height;

	render_buffer (render_buffer const&);
	render_buffer& operator= (render_buffer const&);

public:

	render_buffer (int width, int height, int format);
	~render_buffer ();

	void	bind () const;
	void	unbind () const;

	int		width () const { return m_width; }
	int		height () const { return m_height; }

	GLuint	handle () { return m_rb; }
};


class frame_buffer
{
	GLuint	m_fbo;

	void attach (int attachment_point, int target, int handle);

	frame_buffer (frame_buffer const&);
	frame_buffer& operator= (frame_buffer const&);

public:

	frame_buffer ();
	~frame_buffer ();

	void	attach (int attachment_point, render_buffer& rb);
	template <GLenum target, GLenum f, GLenum iF, typename t>
	void	attach (int attachment_point, texture <target, f, iF, t>& t) { attach (attachment_point, target, t.handle ()); }
	void	detach (int attachment_point);

	void	set_viewport (int x, int y, int w, int h);
	void	set_buffers (GLenum draw_buffer, GLenum read_buffer);

	void	activate ();
	void	deactivate ();

	int		status () const;

	static bool is_supported ();
};
