/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#pragma once

#include "extensions.h"

#include <cstring>
#include <cassert>



namespace usage
{
	struct read_once
	{
		static GLenum const buffer_target = GL_PIXEL_PACK_BUFFER_ARB;
		static GLenum const stream_operation = GL_STREAM_READ;
		static GLenum const optimize_operation = GL_READ_ONLY;
	};

	struct write_once
	{
		static GLenum const buffer_target = GL_PIXEL_UNPACK_BUFFER_ARB;
		static GLenum const stream_operation = GL_STREAM_DRAW;
		static GLenum const optimize_operation = GL_WRITE_ONLY;
	};

	//	More usage patterns possible. For an explaination of possible uses, see
	//	http://www.opengl.org/sdk/docs/man/xhtml/glBufferData.xml
};

#ifndef __APPLE__
extern PFNGLGENBUFFERSARBPROC		glGenBuffersARB;
extern PFNGLDELETEBUFFERSARBPROC	glDeleteBuffersARB;
extern PFNGLBINDBUFFERARBPROC		glBindBufferARB;
extern PFNGLBUFFERDATAARBPROC		glBufferDataARB;
extern PFNGLMAPBUFFERARBPROC		glMapBufferARB;
extern PFNGLUNMAPBUFFERARBPROC		glUnmapBufferARB;
#endif

template <typename T, typename U>
class pixel_buffer
{
	GLuint	m_pbo;
	T*		m_ptr;
	size_t	m_size;

	pixel_buffer (pixel_buffer const&);
	pixel_buffer& operator= (pixel_buffer const&);

public:

	pixel_buffer (size_t sz) : m_ptr (0), m_size (0)
	{
		glGenBuffersARB (1, &m_pbo);
		resize (sz);
	}

	~pixel_buffer ()
	{
		glDeleteBuffersARB (1, &m_pbo);
	}

	void resize (size_t sz)
	{
		if (sz <= m_size)
			return;

		m_size = sz;

		glBindBufferARB (U::buffer_target, m_pbo);
		glBufferDataARB (U::buffer_target, m_size * sizeof (T), 0, U::stream_operation);
		glBindBufferARB (U::buffer_target, 0);
	}

	T* get_buffer ()
	{
		if (!m_ptr)
		{
			m_ptr = (T*) glMapBufferARB (U::buffer_target, U::optimize_operation);
			assert (m_ptr);
		}
		return m_ptr;
	}

	void release_buffer ()
	{
		if (m_ptr)
			glUnmapBufferARB (U::buffer_target);
		m_ptr = 0;
	}

	void activate ()
	{
		glBindBufferARB (U::buffer_target, m_pbo);
	}

	void deactivate ()
	{
		glBindBufferARB (U::buffer_target, 0);
	}

	static void* pointer (int offset) { return (char /* T? */*) 0 + offset; }
	static bool is_supported ()
	{
		bool pixel_buffer_supported ();
		return pixel_buffer_supported ();
	}
};
