/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#include "../matrix.h"
#include "../engine_base.h"
#include "engine_gpu.h"
#include "shader_gpu.h"

#include <limits>



#ifdef WIN32
#include <windows.h>
#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif

std::string load_resource (std::string name)
{
	std::replace (name.begin (), name.end (), '/', '_');
	std::replace (name.begin (), name.end (), '.', '_');

	HMODULE	g_instance = GetModuleHandle (0);
	HRSRC			hRes = FindResource (g_instance, name.c_str (), "SHADER");
	HGLOBAL			res_gl = LoadResource (g_instance, hRes);
	std::string		ret ((char*) LockResource (res_gl), SizeofResource (g_instance, hRes));

	FreeResource (res_gl);
	return ret;
}
#else
#include "shader/passthru.cg.h"
#include "shader/selftest.cg.h"
#include "shader/1/addcol.cg.h"
#include "shader/1/art.cg.h"
#include "shader/1/copy.cg.h"
#include "shader/1/enterbnd.cg.h"
#include "shader/1/enterbnd_se.cg.h"
#include "shader/1/enter.cg.h"
#include "shader/1/id.cg.h"
#include "shader/1/leavebnd.cg.h"
#include "shader/1/leave.cg.h"
#include "shader/1/lookup.cg.h"
#include "shader/1/mul.cg.h"
#include "shader/1/mulcol.cg.h"
#include "shader/1/mulrow.cg.h"
#include "shader/1/mulsq.cg.h"
#include "shader/1/mult1.cg.h"
#include "shader/1/mult2.cg.h"
#include "shader/1/norm2.cg.h"
#include "shader/1/pivot.cg.h"
#include "shader/1/pmul.cg.h"
#include "shader/1/reduce_add.cg.h"
#include "shader/1/reduce_add_rest.cg.h"
#include "shader/1/reduce_init_enterbnd.cg.h"
#include "shader/1/reduce_init_enterbnd_se.cg.h"
#include "shader/1/reduce_init_enter.cg.h"
#include "shader/1/reduce_init_leavebnd_neg.cg.h"
#include "shader/1/reduce_init_leavebnd_pos.cg.h"
#include "shader/1/reduce_init_leave.cg.h"
#include "shader/1/reduce_maxy_rest_bnd.cg.h"
#include "shader/1/reduce_maxy_sqw.cg.h"
#include "shader/1/reduce_miny.cg.h"
#include "shader/1/reduce_minyneg_rest.cg.h"
#include "shader/1/reduce_miny_rest_bnd.cg.h"
#include "shader/1/reduce_miny_rest.cg.h"
#include "shader/1/reduce_miny_sqw.cg.h"
#include "shader/1/resbnd.cg.h"
#include "shader/1/res.cg.h"
#include "shader/1/scatter.cg.h"
#include "shader/1/sid.cg.h"
#include "shader/1/snorm2.cg.h"
#include "shader/1/square.cg.h"
#include "shader/1/sub.cg.h"
#include "shader/1/subcol.cg.h"
#include "shader/1/update_g.cg.h"
#include "shader/1/update_g2.cg.h"
#include "shader/1/updatexn.cg.h"
#include "shader/1/waddsub.cg.h"
#include "shader/1/wsub.cg.h"

std::string load_resource (std::string name)
{
	static std::map <std::string, std::string> shaders;

	shaders ["cg/selftest.cg"] = selftest_cg;
	shaders ["cg/passthru.cg"] = passthru_cg;
	shaders ["cg/1/addcol.cg"] = addcol_cg;
	shaders ["cg/1/art.cg"] = art_cg;
	shaders ["cg/1/copy.cg"] = copy_cg;
	shaders ["cg/1/enterbnd.cg"] = enterbnd_cg;
	shaders ["cg/1/enterbnd_se.cg"] = enterbnd_se_cg;
	shaders ["cg/1/enter.cg"] = enter_cg;
	shaders ["cg/1/id.cg"] = id_cg;
	shaders ["cg/1/leavebnd.cg"] = leavebnd_cg;
	shaders ["cg/1/leave.cg"] = leave_cg;
	shaders ["cg/1/lookup.cg"] = lookup_cg;
	shaders ["cg/1/mul.cg"] = mul_cg;
	shaders ["cg/1/mulcol.cg"] = mulcol_cg;
	shaders ["cg/1/mulrow.cg"] = mulrow_cg;
	shaders ["cg/1/mulsq.cg"] = mulsq_cg;
	shaders ["cg/1/mult1.cg"] = mult1_cg;
	shaders ["cg/1/mult2.cg"] = mult2_cg;
	shaders ["cg/1/norm2.cg"] = norm2_cg;
	shaders ["cg/1/pivot.cg"] = pivot_cg;
	shaders ["cg/1/pmul.cg"] = pmul_cg;
	shaders ["cg/1/reduce_add.cg"] = reduce_add_cg;
	shaders ["cg/1/reduce_add_rest.cg"] = reduce_add_rest_cg;
	shaders ["cg/1/reduce_init_enterbnd.cg"] = reduce_init_enterbnd_cg;
	shaders ["cg/1/reduce_init_enterbnd_se.cg"] = reduce_init_enterbnd_se_cg;
	shaders ["cg/1/reduce_init_enter.cg"] = reduce_init_enter_cg;
	shaders ["cg/1/reduce_init_leavebnd_neg.cg"] = reduce_init_leavebnd_neg_cg;
	shaders ["cg/1/reduce_init_leavebnd_pos.cg"] = reduce_init_leavebnd_pos_cg;
	shaders ["cg/1/reduce_init_leave.cg"] = reduce_init_leave_cg;
	shaders ["cg/1/reduce_maxy_rest_bnd.cg"] = reduce_maxy_rest_bnd_cg;
	shaders ["cg/1/reduce_maxy_sqw.cg"] = reduce_maxy_sqw_cg;
	shaders ["cg/1/reduce_miny.cg"] = reduce_miny_cg;
	shaders ["cg/1/reduce_minyneg_rest.cg"] = reduce_minyneg_rest_cg;
	shaders ["cg/1/reduce_miny_rest_bnd.cg"] = reduce_miny_rest_bnd_cg;
	shaders ["cg/1/reduce_miny_rest.cg"] = reduce_miny_rest_cg;
	shaders ["cg/1/reduce_miny_sqw.cg"] = reduce_miny_sqw_cg;
	shaders ["cg/1/resbnd.cg"] = resbnd_cg;
	shaders ["cg/1/res.cg"] = res_cg;
	shaders ["cg/1/scatter.cg"] = scatter_cg;
	shaders ["cg/1/sid.cg"] = sid_cg;
	shaders ["cg/1/snorm2.cg"] = snorm2_cg;
	shaders ["cg/1/square.cg"] = square_cg;
	shaders ["cg/1/sub.cg"] = sub_cg;
	shaders ["cg/1/subcol.cg"] = subcol_cg;
	shaders ["cg/1/update_g.cg"] = update_g_cg;
	shaders ["cg/1/update_g2.cg"] = update_g2_cg;
	shaders ["cg/1/updatexn.cg"] = updatexn_cg;
	shaders ["cg/1/waddsub.cg"] = waddsub_cg;
	shaders ["cg/1/wsub.cg"] = wsub_cg;

	std::map <std::string, std::string>::iterator i = shaders.find (name);
	if (i != shaders.end ())
		return i->second;
	else
		return std::string ();
}
#endif



namespace gpu
{

	char const* const shader::manager::globalDefines = "#define eps			0.0001f\n"
													   "#define FLOAT		float\n"
													   "#define FLOAT2		float2\n"
													   "#define FLOAT4		float4\n"
													   "#define FLOAT44		float4x4\n"
													   "#define F4_TEX_RECT	f4texRECT\n"
													   "\n"
													   "inline FLOAT cval (samplerRECT tex, FLOAT fidx) { return f4texRECT(tex, half2(fidx,0.5)).r; }\n"
													   "inline FLOAT xval (samplerRECT tex, FLOAT fidx) { return f4texRECT(tex, half2(fidx,1.5)).r; }\n"
													   "inline FLOAT uval (samplerRECT tex, FLOAT fidx) { return f4texRECT(tex, half2(fidx,2.5)).r; }\n"
													   "inline FLOAT lval (samplerRECT tex, FLOAT fidx) { return f4texRECT(tex, half2(fidx,3.5)).r; }\n"
													   "\n"
													   "inline bool is_basic (samplerRECT tex, FLOAT fidx) { return fval (tex, fidx) == 1; }\n"
													   "\n";
	std::string shader::manager::specialNumbers;

	shader::manager::manager ()
	{
		//	set up Cg
		GPUtype = cgGLIsProfileSupported (CG_PROFILE_GPU_FP) ? 1 : 0;
//		GPUtype = 0;	//	force using old ARB shaders

		if (GPUtype == 1)
			fragmentProfile = CG_PROFILE_GPU_FP;
		else if (cgGLIsProfileSupported (CG_PROFILE_ARBFP1))
			fragmentProfile = CG_PROFILE_ARBFP1;	// CG_PROFILE_FP40;	//	if we use FP4, check the inf and NaN stuff
		else
			throw std::runtime_error ("GPU not suffienctly programmable.");

		cgContext = cgCreateContext ();
		cgGLEnableProfile (fragmentProfile);
		cgGLSetOptimalOptions (fragmentProfile);
	}

	bool shader::manager::find_config ()
	{
		engine& pengine = gpu::engine::instance ();

		if (GPUtype == 1)
		{	//	NOTE: if things get weird, this *might* be the source, depending
			//		  on how well the IEEE float standard is actually implemented
			specialNumbers = "static float const inf = 1.0 / 0.0;\n\n";
		}
		else
		{	//	HACK: maybe we find a better solution than this approximate
			specialNumbers = "static float inf = 1e38;\n\n";
		}

		shader* st = new shader (*this, "cg/selftest.cg");
		st->compile ();
		st->set_parameter ("test_nan", std::numeric_limits <float>::quiet_NaN ());
		st->set_parameter ("test_inf", std::numeric_limits <float>::infinity ());
		st->bind ();

		pengine.precache_run ();

		delete st;

		if (!pengine.get_last_result (0))
			printf ("warning: self-test failed, infinity set to 1e38 instead\n");
		if (!pengine.get_last_result (1))
			printf ("warning: self-test failed, infinity comparisons inconsistent\n");
		if (!pengine.get_last_result (2))
			printf ("warning: self-test failed, NaN not uploaded properly\n");
		if (!pengine.get_last_result (3))
			printf ("warning: self-test failed, NaN comparisons inconsistent\n");

		//	we must have the NaN stuff .. but we can live with inf = 1e38
		return pengine.get_last_result (2) && pengine.get_last_result (3);
	}

	void shader::manager::load_shaders ()
	{
//		shaders.push_back (std::make_pair (new shader (*this, "cg/merge.cg"),						(shader*) 0));
//		shaders.push_back (std::make_pair (new shader (*this, "cg/compact.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/passthru.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/id.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/sid.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/art.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/copy.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/sub.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/wsub.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/update_g.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/update_g2.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/mul.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/mulcol.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/mulrow.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/mult1.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/mult2.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/pmul.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/norm2.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/square.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/leave.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/leavebnd.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/enter.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/enterbnd.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/enterbnd_se.cg"),				(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_add.cg"),				(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_add_rest.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_enter.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_enterbnd.cg"),		(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_enterbnd_se.cg"),	(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_leave.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_leavebnd_pos.cg"),	(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_init_leavebnd_neg.cg"),	(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_miny.cg"),				(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_miny_sqw.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_maxy_sqw.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_miny_rest.cg"),			(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_miny_rest_bnd.cg"),		(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_maxy_rest_bnd.cg"),		(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/reduce_minyneg_rest.cg"),		(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/res.cg"),						(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/resbnd.cg"),					(shader*) 0));
		shaders.push_back (std::make_pair (new shader (*this, "cg/1/lookup.cg"),					(shader*) 0));

		//	Compile those shaders that do not depend on #defines.
		//	The gp4fp shaders do not use #defines at all, so they
		//	will all compile. Some ARB shaders need recompilation
		//	tho, because we need to add individual #defines for
		//	each operation. This is due to the limits of the profile,
		//	most importantly the fact that we cannot use variable
		//	loops.
		for (shader_list_t::iterator i = shaders.begin (); i != shaders.end (); ++ i)
		{
			if (i->first != 0)	i->first->compile ();
			if (i->second != 0)	i->second->compile_silent ();
		}
	}

	shader::manager::~manager ()
	{
		for (shader_list_t::iterator i = shaders.begin (); i != shaders.end (); ++ i)
		{
			delete i->first;
			delete i->second;
		}

		//	free Cg
		cgDestroyContext (cgContext);
		cgContext = 0;
	}

	void shader::manager::precache () const
	{
		for (shader_list_t::const_iterator i = shaders.begin (); i != shaders.end (); ++ i)
			i->first->precache ();
	}




	void shader::precache () const
	{
		texture_rect1f	tmp;
		CGenum	const	types [] = { CG_GLOBAL, CG_PROGRAM };

		tmp.create (1, 1);

		bind ();
		for (int j = 0; j < 2; ++ j)
		{
			CGparameter i = cgGetFirstParameter (program, types [j]);
			if (!i)
				return;

			do
			{
				if (cgGetParameterVariability (i) == CG_UNIFORM)
				{
					if (cgGetParameterClass(i) == CG_PARAMETERCLASS_SAMPLER)
						set_parameter (cgGetParameterName (i), tmp);
					else
						set_parameter (cgGetParameterName (i), 0.0f);
				}
			}
			while ((i = cgGetNextParameter (i)) != 0);
			checkCgError ();
		}
		gpu::environment::pengine->precache_run ();
	}

	void shader::compile_silent ()
	{
		std::string tmpSource (manager::globalDefines);
		tmpSource.append (manager::specialNumbers);
		tmpSource.append (globals);
		tmpSource.append (source);

		//	entry point omitted, thus it is always "main"
		CGprogram fp = cgCreateProgram (mgr.cgContext,
										CG_SOURCE,
										tmpSource.c_str (),
										mgr.fragmentProfile, 0, 0);
		if (fp != 0)
		{
			if (program != 0)
			{
				cgDestroyProgram (program);
			}

			cgGLLoadProgram (fp);
			program = fp;

			if (!mgr.is_old_gpu ())
			{
				load_parameters (CG_GLOBAL);
				load_parameters (CG_PROGRAM);
			}
		}
		else
		{
			printf ("warning: shader initialization delayed (%s)\n", file.c_str ());
		}
	}

	void shader::compile ()
	{
		std::string tmpSource (manager::globalDefines);
		tmpSource.append (manager::specialNumbers);
		tmpSource.append (globals);
		tmpSource.append (source);

		//	entry point omitted, thus it is always "main"
		CGprogram fp = cgCreateProgram (mgr.cgContext,
										CG_SOURCE,
										tmpSource.c_str (),
										mgr.fragmentProfile, 0, 0);
		if (fp == 0)
		{
			printf ("%s:\n", file.c_str ());
			printf ("%s\n", cgGetLastListing (mgr.cgContext));
			checkCgError ();
		}
		else
		{
			if (program != 0)
			{
				cgDestroyProgram (program);
			}

			cgGLLoadProgram (fp);
			program = fp;

			if (!mgr.is_old_gpu ())
			{
				load_parameters (CG_GLOBAL);
				load_parameters (CG_PROGRAM);
			}
		}
	}

};
