/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#pragma once

#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <stdexcept>

#include <Cg/cg.h>
#include <Cg/cgGL.h>



std::string load_resource (std::string name);

namespace gpu
{

	/**
	 * Check Cg runtime for errors and throw std::runtime_error in case of an error.
	 */
	inline void checkCgError ()
	{
		CGerror cgError = cgGetError ();
		if (cgError != 0)
		{
			throw std::runtime_error (cgGetErrorString (cgError));
		}
	}

	inline char const* c_str (int i)
	{
		static char tmp [32];
		sprintf (tmp, "%i", i);
		return tmp;
	}

	/**
	 *
	 * @author Jakob Bieling
	 */
	class shader
	{
	public:

		enum operation
		{
			PASSTHRU,
			IDENTITY,
			SIGNED_IDENTITY,
			ARTIFICIALS,
			COPY,
			SUB,
			WSUB,
			UPDATE_G,
			UPDATE_G2,
			MUL,
			MUL_COL,
			MUL_ROW,
			MUL_T1,
			MUL_T2,
			PIVOT_MUL,
			NORM2,
			SQUARE,
			LEAVE,
			LEAVE_BND,
			ENTER,
			ENTER_BND,
			ENTER_BND_SE,
			REDUCE_ADD,
			REDUCE_ADD_REST,
			REDUCE_INIT_ENTER,
			REDUCE_INIT_ENTERBND,
			REDUCE_INIT_ENTERBND_SE,
			REDUCE_INIT_LEAVE,
			REDUCE_INIT_LEAVEBND_POS,
			REDUCE_INIT_LEAVEBND_NEG,
			REDUCE_MINY,
			REDUCE_MINY_SQW,
			REDUCE_MAXY_SQW,
			REDUCE_MINY_REST,
			REDUCE_MINY_REST_BND,
			REDUCE_MAXY_REST_BND,
			REDUCE_MINYNEG_REST,
			RES,
			RES_BND,
			LOOKUP
		};

		class manager
		{
			typedef std::vector < std::pair <shader*, shader*> >	shader_list_t;

			static char const* const	globalDefines;
			static std::string			specialNumbers;

			CGcontext		cgContext;
			CGprofile		fragmentProfile;
			int				GPUtype;
			shader_list_t	shaders;

		public:

			manager ();
			~manager ();

			void precache () const;

			bool find_config ();
			void load_shaders ();

			bool is_old_gpu () const
			{
				return GPUtype == 0;
			}

			shader& get_shader (shader::operation op/*, Texture::Usage uc*/) const
			{
	//			if (uc == Texture::REGULAR)
					return *shaders [op].first;
	//			else
	//				return *shaders [op].second;
			}

			friend class shader;
		};

	private:

		shader::manager const&	mgr;
		std::string				globals;
		std::string				file;
		std::string				source;
		CGprogram				program;
		std::map <std::string, CGparameter> params;

		void load_parameters (CGenum type)
		{
			CGparameter i = cgGetFirstParameter (program, type);
			if (!i)
				return;

			do
			{
				if (cgGetParameterVariability (i) == CG_UNIFORM)
					params [cgGetParameterName (i)] = i;
			}
			while ((i = cgGetNextParameter (i)) != 0);
		}

		CGparameter get_parameter (std::string const& n) const
		{
			if (mgr.is_old_gpu ())
			{
				return cgGetNamedParameter (program, n.c_str ());
			}
			else
				return params.find (n)->second;
		}

		shader (shader::manager const& parent, std::string const& s) :
			mgr (parent),
			file (s),
			program (0)
		{
			source = load_resource (s);
		}

	public:

		void precache () const;

		/**
		 * Compile the shader using the Cg runtime without producing output
		 */
		void compile_silent ();

		/**
		 * Compile the shader using the Cg runtime.
		 */
		void compile ();

		/**
		 * Bind the shader program as active on GPU.
		 */
		void bind () const
		{
	//		printf ("shader source for %s:\n%s\n", file.c_str (), source.c_str ());

			cgGLBindProgram (program);
		}
#ifdef USE_1CHN_TEXTURE
		void set_parameter (char const* name, texture_rect1f const& tex) const
		{
			CGparameter param = get_parameter (name);
			cgGLSetTextureParameter (param, tex.handle ());
			cgGLEnableTextureParameter (param);
		}
#endif
		void set_parameter (char const* name, texture_rect4f const& tex) const
		{
			CGparameter param = get_parameter (name);
			cgGLSetTextureParameter (param, tex.handle ());
			cgGLEnableTextureParameter (param);
		}

		void set_parameter (char const* name, int i) const
		{
			cgSetParameter1i (get_parameter (name), i);
		}

		void set_parameter (char const* name, float f) const
		{
			cgSetParameter1f (get_parameter (name), f);
		}

		void set_parameter (char const* name, int i1, int i2) const
		{
			cgSetParameter2i (get_parameter (name), i1, i2);
		}

		void set_parameter (char const* name, float f1, float f2) const
		{
			cgSetParameter2f (get_parameter (name), f1, f2);
		}

		void set_parameter (char const* name, int i1, int i2, int i3, int i4) const
		{
			cgSetParameter4i (get_parameter (name), i1, i2, i3, i4);
		}

		void set_parameter (char const* name, int const* i) const
		{
			cgSetParameter4iv (get_parameter (name), i);
		}

		void set_parameter (char const* name, float const* f) const
		{
			cgSetParameter4fv (get_parameter (name), f);
		}

		void set_special_parameter (char const* name, int i)
		{
			if (mgr.is_old_gpu ())
			{
				globals.assign ("#define ARBFP1\n\nstatic const int ").
						append (name).
						append (" = ").
						append (c_str (i)).
						append (";\n");

				compile ();
				bind ();
			}
			else
			{
				bind ();
				set_parameter (name, i);
			}
		}

		void set_special_parameters (char const* name, int i1, int i2)
		{
			if (mgr.is_old_gpu ())
			{
				globals.assign ("#define ARBFP1\n\nstatic const int2 ").
						append (name).
						append (" = int2(").
						append (c_str (i1)).
						append (", ").
						append (c_str (i2)).
						append (");\n");

				compile ();
				bind ();
			}
			else
			{
				bind ();
				set_parameter (name, i1, i2);
			}
		}
	};

};
