/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 06, 2009


#pragma once

#include "bounds.h"
#include "engine_base.h"

#ifdef _DEBUG
#include <vector>
#include <sstream>
#include <iomanip>
#include <ios>
#endif


template <typename environment_t, typename storage_t>
struct matrix
{
private:

	struct ptr_storage
	{
		storage_t	data;
		size_t		shared;
	};

	void free ()
	{
		if (texture)
		{
			-- texture->shared;
			if (!texture->shared) delete texture;
		}
	}

	matrix (matrix const&);
	matrix& operator= (matrix const&);

public:

	ptr_storage*	texture;
	bounds_t		bounds;

	matrix () : texture (0)
	{
	}

	~matrix ()
	{
		free ();
	}

	void swap (matrix& other)
	{
		std::swap (texture, other.texture);
		std::swap (bounds, other.bounds);
	}

	void resize (int w, int h)
	{
		if (!texture)
		{
			texture = new ptr_storage;
			texture->shared = 1;
		}

		if (texture->data.width () < w || texture->data.height () < h)
			texture->data.create (w, h);
		bounds.width = w;
		bounds.height = h;
	}

	void share (matrix& other)
	{
		free ();

		texture = other.texture;
		++ texture->shared;

		bounds = other.bounds;
	}


	inline float get_element (int x, int y) const
	{
		float tmp [4];
		environment_t::pengine->get_element (texture->data, x, y + bounds.yoffset, tmp);

		return tmp [0];
	}

	inline void get_elements (int x, int y, int w, int h, float* f) const
	{
		environment_t::pengine->get_elements (texture->data, x, y + bounds.yoffset, w, h, f);
	}

	inline void set_element (float f, int x, int y)
	{
		float const tmp [] = { f, 0.0f, 0.0f, 0.0f };
//		texture->data.set_texel (tmp, x, y + bounds.yoffset);
		//	this below is faster!
		environment_t::pengine->set_element (texture->data, f, x, y + bounds.yoffset);
	}

	inline void set_elements (typename environment_t::pixel_buffer const& pbo, int x, int y, int w, int h)
	{
		texture->data.fill_rect (x, y + bounds.yoffset, w, h, GL_RED, pbo);
	}

	inline void set_elements (float const* f, int x, int y, int w, int h)
	{
		texture->data.fill_rect (x, y + bounds.yoffset, w, h, GL_RED, f);
	}

	inline void set_elements (float f, int x, int y, int w, int h)
	{
		environment_t::pengine->set_elements (texture->data, f, x, y + bounds.yoffset, w, h);
	}

	inline void set_zero ()
	{
		environment_t::pengine->zero (texture->data);
	}

	inline void set_zero (int x, int y, int w, int h)
	{
		environment_t::pengine->zero (texture->data, x, y + bounds.yoffset, w, h);
	}

	inline void set_identity (int xstart, int ystart)
	{
		environment_t::pengine->identity (*this, xstart, ystart);
	}

	inline void set_signed_identity (int xstart, int ystart, matrix const& signs)
	{
		environment_t::pengine->identity (*this, xstart, ystart, signs);
	}

	inline void set_artificials (int xstart, int ystart, matrix const& signs)
	{
		environment_t::pengine->artificials (*this, xstart, ystart, signs);
	}

	inline void lookup (matrix const& src, matrix const& idx)
	{
		environment_t::pengine->lookup (src, idx, *this);
	}

#ifdef _DEBUG
	inline void print (FILE* f, char const* prefix) const
	{
		int w = bounds.tex_width ();
		int h = bounds.tex_height ();

		std::vector <float> data (w * h * texture->data.components);

		environment_t::pengine->get_elements (texture->data, 0, bounds.yoffset, w, h, &data.front ());

		std::stringstream ss;
		ss.flags (std::ios::right);
		ss.width (8);
		ss.precision (4); 
		ss << std::fixed << std::setprecision (4);

		if (w == 1)
			std::swap (h, w);
		
		for (int y = 0; y < h; ++ y)
		{
			for (int x = 0; x < w; ++ x)
			{
				if (data [x + w * y] != data [x + w * y])
					ss << "     nan ";
				else
				{
					ss.width (8);
					ss << data [x + w * y] << " ";
				}
			}
			ss << std::endl;
		}
		ss << std::endl;

		fprintf (f, "%s%ix%i\n", prefix, h, w);

		std::string const& str (ss.str ());
		fwrite (str.c_str (), 1, str.size (), f);
	}
#endif
};
