/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	June 07, 2009

#pragma comment (lib, "opengl32.lib")
#pragma comment (lib, "glu32.lib")
#pragma comment (lib, "cg.lib")
#pragma comment (lib, "cgGL.lib")


#define CPP
#include "revised_simplex.h"
#include "../timer.h"

#include <vector>
#include <cmath>
#include <ctime>
#include <limits>

#define SANITY_CHECK 0
#define DEBUG_VERBOSE 2

#define dbg_log(a,i) {FILE* f = fopen ("dbg.log", "a"); fprintf (f, a "%i\n", i); fclose (f); }

static bool precache_run = false;



#ifdef _DEBUG
	#define dbg_print_matrix(p,m) if (!precache_run) m.print (stdout, p)
	#define dbg_print if (!precache_run) printf
#else
	#define dbg_print_matrix(p,m) {}
	void dbg_print(char const* p,...) {}
#endif



enum solve_type_e
{
	not_set,
	upper_bound_1,
	no_upper_bound
};

enum implementation_e
{
	simple,
	regular,
	bounded
};

static float const NaN = std::numeric_limits <float>::quiet_NaN ();
static float const inf = std::numeric_limits <float>::infinity ();


template <typename env>
class revised_simplex_impl
{
	typedef typename env::matrix		matrix;
	typedef typename env::matrix4f		matrix4f;
	typedef typename env::pixel_buffer	pixel_buffer;

	solve_type_e	solve_as;

	matrix			c,		//	objective function
					Ao,		//	constraint matrix
					b,		//	b-vector

					Binv,	//	inverse of basis matrix B
					xB,		//	basic solution
					cB,		//	basic part of objective function
					chi,	//	cB indices into c, stored in the same texture as cB
							//	but in the second row

					g,		//	steepest-edge weights
					x,		//	values of all variables (basic ones are NaN)

					ub,		//	upper bounds for x, stored in the same texture as c.
					lb;		//	lower bounds for x, also stored in the same texture as c
							//	if a specific bound is not present, we store NaN

	matrix			pi,
					d;

	mutable matrix	tmpRow1,
					tmpRow2,
					tmpCol,
					tmpMat;

	matrix4f		tmpRow4f [2];

	std::vector <float> lbounds;
	std::vector <float> ubounds;
	std::vector <double> costs;

	GLint			m_max_size;
	size_t			m_alloced_w,  m_alloced_h;
	pixel_buffer*	m_pbo;

	bool			phase1;
	size_t			num_variables;
	size_t			num_artificials;
	size_t			num_nonartificials;
	size_t			num_iterations;
	float			objective_value;
	float			artificial_value;

	static int const max_iterations = 100000000;
	static long const max_seconds = 6 * 60 * 60;

	template <int> struct impl_type {};
	typedef impl_type <simple> simple_implementation;
	typedef impl_type <regular> regular_implementation;
	typedef impl_type <bounded> bounded_implementation;

	bool allocate_memory (size_t cols, size_t rows)
	{
		if (cols > m_max_size || rows > m_max_size)
		{
			printf ("error: problem too large for this GPU\n");
			return false;
		}
		
		m_alloced_w = std::min (cols + rows + rows, static_cast <size_t> (m_max_size));
		m_alloced_h = std::min (rows, static_cast <size_t> (m_max_size));

		Ao.resize (cols, m_alloced_h);
		b.resize (1, m_alloced_h);
		c.resize (m_alloced_w, 5);

		c.bounds.height = 1;

		x.share (c);	 x.bounds.yoffset = 1;
		ub.share (c);	ub.bounds.yoffset = 2;
		lb.share (c);	lb.bounds.yoffset = 3;

		g.resize (m_alloced_w, 1);

		pi.resize (m_alloced_w, 1);
		d.resize (1, m_alloced_h);

		Binv.resize (m_alloced_h, m_alloced_h);
		xB.resize   (1, m_alloced_h);
		cB.resize   (m_alloced_h, 2);

		cB.bounds.height = 1;
		chi.share (cB);
		chi.bounds.yoffset = 1;

		tmpRow1.resize (m_alloced_w, 1);
		tmpRow2.resize (m_alloced_w, 1);
		tmpCol.resize (1, m_alloced_h);
		tmpMat.resize  (m_alloced_h, m_alloced_h);

		tmpRow4f [0].resize (m_alloced_w, 1);
		tmpRow4f [1].resize (m_alloced_w, 1);

		//	NOTE: We later use the PBO to upload 2 full rows (c/x and ux/lx) of floats
		//		  in one go. Thus, if cols*rows would not suffice, we allocate more as
		//		  a special case.
		m_pbo->resize (std::max (cols * rows, m_alloced_w * 4));

		return true;
	}
/*
	feasability_e simplex (simple_implementation, pricing_e pricing)
	{
		if (DEBUG_VERBOSE > 1)
		{
			dbg_print_matrix ("c: ", c);
			dbg_print_matrix ("A: ", A);
			dbg_print_matrix ("b: ", b);
			dbg_print_matrix ("x: ", x);
			dbg_print_matrix ("ux: ", ux);
			dbg_print_matrix ("lx: ", lx);
			dbg_print_matrix ("B^-1: ", Binv);
			dbg_print_matrix ("xB: ", xB);
			dbg_print_matrix ("chi: ", chi);
			dbg_print_matrix ("cB: ", cB);
		}

		time_t const start_time = time (0);
		solve_as = no_upper_bound;

		feasability_e ret = aborted;
		for (int i = max_iterations; i > 0; -- i)
		{
			if (i % 100 == 0 && time (0) - start_time  > max_seconds)
				break;

			//	**	find entering variable
			//	As there should be no artificials, we do not have to
			//	take special care of them
			TIME("multiply", env::pengine->multiply (cB, Binv, pi));
			TIME("multiply", env::pengine->multiply (pi, A, tmpRow1));

			if (DEBUG_VERBOSE > 0)
				dbg_print ("\n");
			if (DEBUG_VERBOSE > 1)
			{
				dbg_print ("========== ENTER [ROUND %i] ==========\n\n", max_iterations - i);
				dbg_print_matrix ("pi = cB * B^-1: ", pi);
				dbg_print_matrix ("res = (cB * B^-1) * A: ", tmpRow1);
				dbg_print_matrix ("c (is subtracted from res):", c);
			}

			TIME("find_entering", env::pengine->find_entering_reduce (tmpRow1, c, tmpRow4f));
			float const	fenter = env::pengine->get_last_result (0);
			int const	ienter = static_cast <int> (fenter);

			if (DEBUG_VERBOSE > 0)
				dbg_print ("enter: %f\n\n", fenter);

			if (ienter < 0)
			{
				num_iterations = max_iterations - i;

				TIME("get_solution", env::pengine->get_solution (cB, xB, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = optimal;
				break;
			}

			//	**	find leaving variable
			TIME("multiply_col", env::pengine->multiply (Binv, A, fenter, d));
			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("Binv * A_enter: ", d);

			TIME("find_leaving", env::pengine->find_leaving_reduce (xB, d, tmpRow4f));
			float const	fleave = env::pengine->get_last_result (0);
			int const	ileave = static_cast <int> (fleave);
			if (DEBUG_VERBOSE > 0)
				dbg_print ("leave: %f\n\n", fleave);
			if (ileave < 0)
			{
				num_iterations = max_iterations - i;

				TIME("get_solution", env::pengine->get_solution (cB, xB, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = unbounded;
				break;
			}

			if (DEBUG_VERBOSE > 1)
			{
				dbg_print ("\n========== B^-1 ==========\n\n");
				dbg_print_matrix ("B^-1: ", Binv);
				dbg_print_matrix ("d: ", d);
				dbg_print ("leave: %i\n\n", ileave);
			}

			//	**	update B^-1
			TIME("pivot_multiply", env::pengine->pivot_multiply (d, fleave, Binv, tmpMat));
			tmpMat.swap (Binv);

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("B^-1: ", Binv);

			//	**	update xB
			TIME("multiply", env::pengine->multiply (Binv, b, xB));

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("xB: ", xB);

			//	**	update cB
			TIME("set_element", chi.set_element (fenter, ileave, 0));
			TIME("lookup", cB.lookup (c, chi));

			if (DEBUG_VERBOSE > 1)
			{
				dbg_print_matrix ("cB: ", cB);
				dbg_print_matrix ("chi: ", chi);
			}
		}
		return ret;
	}

	feasability_e simplex (regular_implementation, pricing_e pricing)
	{
		if (DEBUG_VERBOSE > 1)
		{
			dbg_print_matrix ("c: ", c);
			dbg_print_matrix ("A: ", A);
			dbg_print_matrix ("b: ", b);
			dbg_print_matrix ("x: ", x);
			dbg_print_matrix ("ux: ", ux);
			dbg_print_matrix ("lx: ", lx);
			dbg_print_matrix ("B^-1: ", Binv);
			dbg_print_matrix ("xB: ", xB);
			dbg_print_matrix ("chi: ", chi);
			dbg_print_matrix ("cB: ", cB);
		}

		time_t const start_time = time (0);
		solve_as = no_upper_bound;

		feasability_e ret = aborted;
		for (int i = max_iterations; i > 0; -- i)
		{
			if (i % 100 == 0 && time (0) - start_time  > max_seconds)
				break;

			//	**	find entering variable
			TIME("multiply", env::pengine->multiply (cB, Binv, pi));
			TIME("multiply", env::pengine->multiply (pi, A, tmpRow1, num_nonartificials));

			if (DEBUG_VERBOSE > 0)
				dbg_print ("\n");
			if (DEBUG_VERBOSE > 1)
			{
				dbg_print ("========== ENTER [ROUND %i] ==========\n\n", max_iterations - i);
				dbg_print_matrix ("pi = cB * B^-1: ", pi);
				dbg_print_matrix ("res = (cB * B^-1) * A: ", tmpRow1);
				dbg_print_matrix ("c (is subtracted from res):", c);
			}

			//	tmpRow1 may be shorter than c, because all artificials were left
			//	out (they must never *enter* but always leave the basis). The
			//	implementation of find_entering_reduce takes care of this.
			TIME("find_entering", env::pengine->find_entering_reduce (tmpRow1, c, tmpRow4f));
			float const	fenter = env::pengine->get_last_result (0);
			int const	ienter = static_cast <int> (fenter);
//			float const	xenter = env::pengine->get_last_result (2);

			if (DEBUG_VERBOSE > 0)
//				dbg_print ("enter: %f; xenter: %f; dbg: %f\n\n", fenter, xenter, env::pengine->get_last_result (2));
				dbg_print ("enter: %f\n\n", fenter);

			if (ienter < 0)
			{
				num_iterations = max_iterations - i;

				if (DEBUG_VERBOSE > 0)
				{
					dbg_print_matrix ("chi: ", chi);
					dbg_print_matrix ("xB: ", xB);
				}

				TIME("get_solution", env::pengine->get_solution (cB, xB, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = optimal;
				break;
			}

			//	**	find leaving variable
			TIME("multiply_col", env::pengine->multiply (Binv, A, fenter, d));
			if (DEBUG_VERBOSE > 1)
			{
				dbg_print_matrix ("xB: ", xB);
				dbg_print_matrix ("Binv * A_enter: ", d);
			}

			TIME("find_leaving", env::pengine->find_leaving_reduce (xB, d, tmpRow4f));
			float const	fleave = env::pengine->get_last_result (0);
			int const	ileave = static_cast <int> (fleave);
			float const	tvalue = env::pengine->get_last_result (1);
			if (DEBUG_VERBOSE > 0)
				dbg_print ("leave: %f; t: %f\n\n", fleave, tvalue);
			if (ileave < 0)
			{
				num_iterations = max_iterations - i;

				TIME("get_solution", env::pengine->get_solution (cB, xB, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = unbounded;
				break;
			}

			if (DEBUG_VERBOSE > 1)
			{
				dbg_print ("\n========== B^-1 ==========\n\n");
				dbg_print_matrix ("B^-1: ", Binv);
				dbg_print_matrix ("d: ", d);
				dbg_print ("leave: %i\n\n", ileave);
			}

			//	**	update B^-1
			TIME("pivot_multiply", env::pengine->pivot_multiply (d, fleave, Binv, tmpMat));
			tmpMat.swap (Binv);

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("B^-1: ", Binv);

			//	**	update xB
			TIME("weighted_subtract", env::pengine->weighted_subtract (xB, tvalue, d, tmpCol));
			xB.swap (tmpCol);
 
												  xenter +   tvalue
			TIME("set_element", xB.set_element (tvalue, 0, ileave));
#if SANITY_CHECK > 0
			{
				int leave_idx = chi.get_element (ileave, 0);
				TIME("set_element", fx.set_element (0, leave_idx, 0));
				TIME("set_element", fx.set_element (1, ienter, 0));
				TIME("set_element", x.set_element (tvalue, ienter, 0));
				if (DEBUG_VERBOSE > 1)
					dbg_print_matrix ("x: ", x);
			}
#endif

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("xB: ", xB);

			//	**	update cB
			TIME("set_element", chi.set_element (fenter, ileave, 0));
			TIME("lookup", cB.lookup (c, chi));

			if (DEBUG_VERBOSE > 1)
			{
				dbg_print_matrix ("cB: ", cB);
				dbg_print_matrix ("chi: ", chi);
			}
		}
		return ret;
	}
*/
	feasability_e simplex (bounded_implementation, pricing_e pricing)
	{
		time_t const start_time = time (0);
		solve_as = upper_bound_1;

		int const cols = num_variables;
		int const rows = num_nonartificials - num_variables;

		feasability_e	ret = aborted;
		bool			update_basis = true;
		for (int i = max_iterations - 1; i >= 0; -- i)
		{
//printf("it: %i\n", max_iterations - i);
			if (i % 100 == 0)
			{
				if (progress_cb)
					progress_cb ();
				if (time (0) - start_time  > max_seconds)
					break;
			}

			//	**	find entering variable
			if (update_basis)
				TIME("multiply", env::pengine->multiply (cB, Binv, pi));
			TIME("multiply_n", env::pengine->multiply (pi, Ao, tmpRow1, cols));
			TIME("copy", env::pengine->copy (tmpRow1, cols, 0, pi, 0, 0, rows, 1));
			tmpRow1.bounds.width = cols + rows;

			if (DEBUG_VERBOSE > 0)
				dbg_print ("\n");
			if (DEBUG_VERBOSE > 1)
			{
				tmpRow1.bounds.width = num_nonartificials;
				dbg_print ("========== ENTER [ROUND %i] ==========\n\n", max_iterations - i);
				dbg_print_matrix ("pi = cB * B^-1: ", pi);
				dbg_print_matrix ("res = (cB * B^-1) * A: ", tmpRow1);
				dbg_print_matrix ("c (is subtracted from res):", c);
				dbg_print_matrix ("x (@lb -> smallest, else -> largest): ", x);
				dbg_print_matrix ("g: ", g);
			}
/*
#ifdef _DEBUG
			matrix tmp;
			matrix gamma;
			tmp.resize (num_nonartificials, A.bounds.height);
			gamma.resize (num_nonartificials, 1);
			env::pengine->multiply (Binv, A, tmp, num_nonartificials);
			env::pengine->norm2 (tmp, gamma);

			if (DEBUG_VERBOSE > 1)
			{
				dbg_print_matrix ("current A: ", tmp);
				dbg_print_matrix ("full gamma: ", gamma);
				dbg_print_matrix ("gamma: ", g);
			}

			env::pengine->subtract (gamma, g, tmp);
			std::vector <float> gdiff (tmp.bounds.width);
			tmp.get_elements (0, 0, tmp.bounds.width, 1, &gdiff.front ());
			for (std::vector <float>::iterator i = gdiff.begin (); i != gdiff.end (); ++ i)
				if (abs (*i) > 0.001) __asm int 3
#endif
/*
			static matrix tmp;
			tmp.resize (A.bounds.width, A.bounds.height);
			env::pengine->multiply (Binv, A, tmp);
			env::pengine->norm2 (tmp, g);
*/
			//	tmpRow1 may be shorter than c, because all artificials were left
			//	out (they must never *enter* but always leave the basis). The
			//	implementation of find_entering_reduce takes care of this.
			if (pricing == steepest_edge)
			{
				TIME("find_entering", env::pengine->find_entering_bounded_reduce (tmpRow1, c, g, tmpRow4f));
			}
			else
			{
				TIME("find_entering", env::pengine->find_entering_bounded_reduce (tmpRow1, c, tmpRow4f));
			}
			float const	fenter = env::pengine->get_last_result (0);
			int const	ienter = static_cast <int> (fenter);
			float const	xenter = env::pengine->get_last_result (2);

//			if (ienter >= (int) lbounds.size ())
//				__asm int 3;

//dbg_log(" enter: ", ienter);
			if (DEBUG_VERBOSE > 0)
				dbg_print ("enter: %i (value=%f, x_enter=%f)\n\n", ienter, env::pengine->get_last_result (1), xenter);

			if (ienter < 0)
			{
				num_iterations += max_iterations - i;

				TIME("get_solution", env::pengine->get_solution_bounded (chi, xB, x, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = optimal;
				break;
			}


			//	**	find leaving variable
			if (fenter < cols)
			{
				TIME("multiply_c", env::pengine->multiply (Binv, Ao, fenter, d));
			}
			else
			{
				TIME("copy", env::pengine->copy (d, 0, 0, Binv, fenter - cols, 0, 1, rows));
			}
			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("d = Binv * A_enter: ", d);

			float const maxDiff = abs (xenter - lbounds [ienter]) < 0.0001 ? ubounds [ienter] - xenter
																		   : lbounds [ienter] - xenter;

			TIME("find_leaving", env::pengine->find_leaving_bounded_reduce (xB, d, chi, x, fenter, tmpRow4f, maxDiff));
			float const	fleave = env::pengine->get_last_result (0);
			int const	ileave = static_cast <int> (fleave);
			float const	tvalue = env::pengine->get_last_result (1);
			float const	dvalue = env::pengine->get_last_result (2);
			float const	dnorm = env::pengine->get_last_result (3);
						update_basis = dnorm >= 0;
//dbg_log(" leave: ", ileave);
			if (DEBUG_VERBOSE > 0)
				dbg_print ("leave: %i (t=%f, d_l=%f, |d|=%f, update=%i)\n\n", ileave, tvalue, dvalue, dnorm, update_basis);
			if (ileave < 0)
			{
				num_iterations += max_iterations - i;

				TIME("get_solution", env::pengine->get_solution_bounded (chi, xB, x, num_nonartificials));
				objective_value = env::pengine->get_last_result (0);
				artificial_value = env::pengine->get_last_result (1);

				ret = unbounded;
				break;
			}
			int const iglobleave = (int) chi.get_element (ileave, 0);

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("xB: ", xB);

			//	**	update xB
			TIME("weighted_subtract", env::pengine->weighted_subtract (xB, tvalue, d, tmpCol));
			xB.swap (tmpCol);
			if (update_basis)
			{
				TIME("set_element", xB.set_element (xenter + tvalue, 0, ileave));
			}

			if (DEBUG_VERBOSE > 1)
				dbg_print_matrix ("xB: ", xB);

			if (!update_basis)
			{
				//	**	update xN

				float const newx = tvalue > 0 ? ubounds [ienter] : lbounds [ienter];
				TIME("set_element", x.set_element (newx, ienter, 0));
			}
			else
			{
				//	**	update xN

				//	TODO: Vllt mal ueberlegen, ob man die 2 set_element und ein
				//		  get_element in einem shader verpacken sollte. Bei cB
				//		  machen wir das ja auch so. Oder eben andersrum, bei cB
				//		  das ganze so machen wie hier. So oder so sollten wir
				//		  konsistent die schnellere Variante waehlen.

				float const newx = maxDiff * dvalue < 0 ? ubounds [iglobleave] : lbounds [iglobleave];
				TIME("set_element", x.set_element (NaN, ienter, 0));
				TIME("set_element", x.set_element (newx, iglobleave, 0));


				//	**	update gamma
				bool const full_update_g = i % 1000 == 0;
				if (pricing == steepest_edge && !full_update_g)
				{
					//	v = B^-T * d
					TIME("multiply_t1", env::pengine->multiply_transposed1 (Binv, d, tmpCol));
					if (DEBUG_VERBOSE > 1)
						dbg_print_matrix ("v: ", tmpCol);
				}


				//	**	update B^-1
				if (DEBUG_VERBOSE > 1)
				{
					dbg_print ("\n========== B^-1 ==========\n\n");
					dbg_print_matrix ("B^-1: ", Binv);
					dbg_print_matrix ("d: ", d);
					dbg_print ("leave: %i\n\n", ileave);
				}

				TIME("pivot_multiply", env::pengine->pivot_multiply (d, fleave, Binv, tmpMat));
				tmpMat.swap (Binv);

				if (DEBUG_VERBOSE > 1)
					dbg_print_matrix ("new B^-1: ", Binv);


				//	**	update gamma (cont.)
				if (pricing == steepest_edge)
				{
					if (full_update_g)
					{
						int offx = 0;
						int ccols = cols;
						while (ccols > 0)
						{
							env::pengine->multiply (Binv, Ao, tmpMat, offx, std::min (ccols, Binv.bounds.width));
							env::pengine->norm2 (tmpMat, g, offx);

							offx += Binv.bounds.width;
							ccols -= Binv.bounds.width;
						}
						env::pengine->norm2 (Binv, g, cols);
					}
					else
					{
						//	gamma_e = 1 + d^T * d
						//	we reuse the dnorm value from above:
						//		|d| = d^T * d
						float const g_e = 1 + dnorm;

						//	alpha = (B^-T)_leave * A
						TIME("multiply_r", env::pengine->multiply (Binv, fleave, Ao, tmpRow2, num_variables));
						TIME("copy", env::pengine->copy (tmpRow2, cols, 0, Binv, 0, fleave, rows, 1));
						tmpRow2.bounds.width = cols + rows;
						if (DEBUG_VERBOSE > 1)
							dbg_print_matrix ("alpha: ", tmpRow2);

						//	gamma = (..)
						TIME("update_g", env::pengine->update_g (g, tmpRow2, tmpCol, Ao, g_e, tmpRow1));
						TIME("update_g2", env::pengine->update_g2 (g, tmpRow2, tmpCol, cols, g_e, tmpRow1));
						g.swap (tmpRow1);

						if (iglobleave < num_nonartificials)
						{
							TIME("set_element", g.set_element (dnorm / (dvalue * dvalue), iglobleave, 0));
						}
					}

					if (DEBUG_VERBOSE > 1)
						dbg_print_matrix ("g: ", g);
				}
			}

			if (update_basis)
			{
				//	**	update cB
				TIME("set_element", chi.set_element (fenter, ileave, 0));

				float c_enter = 0.0;
				if (!phase1 && ienter < cols)
					c_enter = static_cast <float> (costs [ienter]);

				TIME("set_element", cB.set_element (c_enter, ileave, 0));
			}
			if (DEBUG_VERBOSE > 1)
			{
				dbg_print_matrix ("cB: ", cB);
				dbg_print_matrix ("chi: ", chi);
			}
		}
		return ret;
	}



public:


	revised_simplex_impl ()
		:	solve_as (not_set),
			m_pbo (0),
			num_variables (0),
			num_nonartificials (0),
			num_iterations (0),
			objective_value (0.0f),
			artificial_value (0.0f),
			progress_cb (0)
	{
		env::engine::instance ();

		glGetIntegerv (GL_MAX_TEXTURE_SIZE, &m_max_size);
		
		//	allocating 16MB PBO will use ~50MB main memory according to Dominik Gddeke:
		//	http://www.mathematik.uni-dortmund.de/~goeddeke/gpgpu/tutorial3.html#remarks
		m_pbo = new pixel_buffer (2048 * 2048);	//	2048*2048 floats -> 16MB

		//	pre-allocate the textures
		allocate_memory (512, 256);
	}

	~revised_simplex_impl ()
	{
	}

	void precache ()
	{
		timer::start ("precache");
		timer::block_new_timers ();
		precache_run = true;

		double fA [] = {   0.000000f,   0.000000f, -0.56459242f, 0.000000f,
					     -16.341440f,   0.000000f, 18.930632f,  11.624500f,
					     -18.384960f, -17.265541f,  0.000000f, -17.014069f,
					       0.000000f,   0.000000f,  0.63905758f, 8.8229008f };
		double fb [] = { 0.011597034f,  6.4815211f, 2.7024140f,  2.5528734f };
		double fc [] = { 9.5675526f,   28.026978f,  5.3587451f,  25.727104f };
/*
		set_constraints (std::vector <double> (fA, fA + 16),
						 std::vector <double> (fb, fb + 4),
						 std::vector <double> (fc, fc + 4)); 
		solve_optimal <simple> (steepest_edge);

		set_constraints (std::vector <double> (fA, fA + 16),
						 std::vector <double> (fb, fb + 4),
						 std::vector <double> (fc, fc + 4)); 
		is_feasible <regular> (steepest_edge);
*/
		set_constraints (std::vector <double> (fA, fA + 16),
						 std::vector <double> (fb, fb + 4),
						 std::vector <double> (fc, fc + 4)); 
		is_feasible <bounded> (steepest_edge);

		precache_run = false;
		timer::unblock_new_timers ();
		timer::stop ("precache");
	}

	bool set_constraints (std::vector <double> const& vA,
						  std::vector <double> const& vb,
						  std::vector <double> const& vc,
						  double const* fub = 0, double const* flb = 0)
	{
		size_t const rows = vb.size ();
		size_t const cols = vc.size ();

		num_variables = cols;
		num_nonartificials = cols + rows;

		//	allocate a bit more than we might actually need
		if (!allocate_memory (cols, rows))
			return false;

		float* tmp;

		//	First build the matrices and vectors for Ax <= b,
		//	i.e., no surplus and no artificial variables added.
		m_pbo->activate ();
		{
			//	build A-matrix
			{
				assert (vA.size () == rows * cols);

				tmp = m_pbo->get_buffer ();
				std::copy (vA.begin (), vA.end (), tmp);
				m_pbo->release_buffer ();
				Ao.set_elements (*m_pbo, 0, 0, cols, rows);
			}

			//	build b-vector
			{
				tmp = m_pbo->get_buffer ();
				std::copy (vb.begin (), vb.end (), tmp);
				m_pbo->release_buffer ();
				b.set_elements (*m_pbo, 0, 0, 1, rows);
			}

			//	build first part of c- and x-vector
			{
				tmp = m_pbo->get_buffer ();

				//	set the objective function
				std::fill (tmp, tmp + cols, 0.0f);
				costs.assign (vc.begin (), vc.end ());

				//	find initial solution for regular x (note, xB is
				//	initialized farther below):
				//	  if c_i > 0 and u_i < +inf, set to upper bound
				//	  if c_i < 0 and l_i > -inf, set to lower bound
				//	  if l_i > -inf, set to lower bound
				//	  else set to 0
				//	for testing:
				//	  if l_i > -inf, set to lower bound
				//	  if u_i < +inf, set to upper bound
				//	  else set to 0
				for (size_t i = cols; i < 2 * cols; ++ i)
					if (flb && !isinf (flb [i - cols]))
						tmp [i] = static_cast <float> (flb [i - cols]);
					else if (fub && !isinf (fub [i - cols]))
						tmp [i] = static_cast <float> (fub [i - cols]);
					else
						tmp [i] = 0;

					m_pbo->release_buffer ();
				c.texture->data.fill_rect (0, 0, cols, 2, GL_RED, *m_pbo);
			}
		}
		m_pbo->deactivate ();



		//	Now that we built A and x (without slacks etc), we want to determine the value
		//	for the slack and artificial variables. Since x=0 does not hold anymore, we must
		//	first calculate tmp_i = sum a_ij * x_j for each row i so we can do:
		//		xs_i = b_i - tmp_i
		//	This will assign all slack variables. If tmp_i > b_i, we need an artificial
		//	variable and set the slack to zero instead.
		assert (Ao.bounds.width == num_variables);
		x.bounds.width = num_variables;
		env::pengine->multiply_transposed2 (Ao, x, d);
		env::pengine->subtract (b, d, tmpCol);

		//	build B^(-1)-matrix (must be done after vector 'd' above is constructed!)
		Binv.set_signed_identity (0, 0, tmpCol);

		std::vector <float> ax (rows);
		env::pengine->get_elements (tmpCol.texture->data, 0, 0, 1, rows, &ax.front ());



		//	determine number of artificial variables
		std::vector <int>	arts; arts.reserve (rows);
		for (size_t i = 0; i < rows ; ++ i)
		{
			if (ax [i] < 0)
				arts.push_back (i);
		}



		//	Idea: Usually, we would have to calculate B^-1*A and get the
		//		  norms of the column-vectors. But due to the way we
		//		  initialized B^-1 (signed identity), the multiplication
		//		  would only change the signs of A. But for the Euclidian
		//		  norm, the signs are of no relevance! Thus, we can just
		//		  calculate the norms of the column-vectors of A directly.
		//		  Additionally, we only do the calculation on the original
		//		  part of A. The norms of the slack and artificial part are
		//		  always 1!
		env::pengine->norm2 (Ao, g);
		g.set_elements (1, num_variables, 0, rows, 1);



		Ao.bounds.width = cols;
		c.bounds.width =
		x.bounds.width =
		ub.bounds.width =
		lb.bounds.width = cols + rows + arts.size ();
		g.bounds.width = cols + rows;

		num_artificials = arts.size ();



		int const full_row = cols + rows + num_artificials;

		size_t const c_begin  = 0;					//	c and x are already set partly. Thus,
		size_t const x_begin  = full_row - cols;	//	tmp is used to set the last part only
		size_t const ux_begin = full_row * 2;
		size_t const lx_begin = full_row * 3;

		size_t const c_end  = x_begin;								//	see above, since we do not need the first
		size_t const x_end  = x_begin + rows + num_artificials;	//	'cols' entries anymore, everything is shorter
		size_t const ux_end = lx_begin;
		size_t const lx_end = full_row * 4;

		if (full_row > m_max_size)
		{
			printf ("error: problem too large for this GPU\n");
			return false;
		}



		//	Now assemble the rest of A, x and c and also store the bounds in textures
		m_pbo->activate ();
		{
			//	build surplus/artificial part of c and x
			{
				tmp = m_pbo->get_buffer ();

				//	set the objective function
				std::fill (tmp + c_begin, tmp + rows,   0.0f);	//	values for surplus
				std::fill (tmp + rows,    tmp + c_end, -1.0f);	//	values for artificials
				
				
				//	artificials and slacks:
				//	  tmp = b_i - sum a_ij*x_j
				//	  if tmp >= 0, xs_i = tmp
				//	  if tmp < 0, xa_i = tmp
				//	see above (were ax is calculated) for more detail
				std::fill (tmp + x_begin, tmp + x_end, 0.0f);
				
				int ca = 0;
				for (size_t i = x_begin; i < x_begin + rows; ++ i)
				{
					tmp [i] = (ax [i - x_begin] < 0) ? 0.0f : NaN;
				}
				
				std::fill (tmp + x_begin + rows, tmp + x_end, NaN);
				m_pbo->release_buffer ();
				c.set_elements (*m_pbo, cols, 0, rows + num_artificials, 2);
			}

			//	build ux/lx-vectors
			{
				tmp = m_pbo->get_buffer ();

				// if no upper bounds specified, no upper bounds will be used
				if (!fub)
					std::fill (tmp, tmp + full_row, inf);
				else
				{
					for (size_t i = 0; i < cols; ++ i)
						tmp [i] = static_cast <float> (fub [i]);
					std::fill (tmp + cols, tmp + full_row, inf);
				}
				ubounds.assign (tmp, tmp + full_row);


				// if no lower bounds specified, lbound = 0
				if (!flb)
					std::fill (tmp + full_row, tmp + full_row * 2, 0.0f);
				else
				{
					for (size_t i = 0; i < cols; ++ i)
						tmp [full_row + i] = static_cast <float> (flb [i]);
					std::fill (tmp + full_row + cols, tmp + full_row * 2, 0.0f);
				}
				lbounds.assign (tmp + full_row, tmp + full_row * 2);

				m_pbo->release_buffer ();
				ub.set_elements (*m_pbo, 0, 0, full_row, 2);
			}

			//	build xB-vector
			{
				tmp = m_pbo->get_buffer ();
				for (int i = rows - 1; i >= 0; -- i)	//	xB contains all artificials
					tmp [i] = static_cast <float> (ax [i] < 0 ? -ax [i] : ax [i]);
				m_pbo->release_buffer ();
				xB.set_elements (*m_pbo, 0, 0, 1, rows);
			}

			//	build chi-vector
			{
				tmp = m_pbo->get_buffer ();
				for (int i = rows - 1; i >= 0; -- i)
					tmp [i] = i + cols + 0.5f;
				for (int i = num_artificials - 1; i >= 0; -- i)	//	add artificials
					tmp [arts [i]] = cols + rows + i + 0.5f;
				m_pbo->release_buffer ();
				chi.set_elements (*m_pbo, 0, 0, rows, 1);
			}
		}
		m_pbo->deactivate ();
		
		//	build cB-vector
		cB.lookup (c, chi);

/*
		{
			float fa [] = { 1,2,3,4,3,2,1,0,1,2,3,4,3,2,1,0,1,2,3,4,3,2,1,0 };
			float fb [] = { 1,4,3,2,1,0,1,3,3,4,5,4,3,2,1,0,1,2,3,4,3,2,1,0,1,2,3,4 };
			matrix a, b, c;

			a.resize (5, 3);
			a.set_elements (fa, 0, 0, 5, 3);

			b.resize (4, 5);
			b.set_elements (fb, 0, 0, 4, 5);

			c.resize (10, 10);
			env::pengine->multiply (a, 1.5, b, c);

			dbg_print_matrix ("a: ", a);
			dbg_print_matrix ("b: ", b);
			dbg_print_matrix ("c: ", c);
		}
*/
		if (DEBUG_VERBOSE > 1)
		{
			dbg_print_matrix ("A: ", Ao);
			dbg_print_matrix ("b: ", b);
			dbg_print_matrix ("c: ", c);
			dbg_print_matrix ("ub: ", ub);
			dbg_print_matrix ("lb: ", lb);
			dbg_print_matrix ("x: ", x);
			dbg_print_matrix ("g: ", g);
			dbg_print_matrix ("chi: ", chi);
			dbg_print_matrix ("xB: ", xB);
			dbg_print_matrix ("cB: ", cB);
			dbg_print_matrix ("Binv: ", Binv);
		}
		glFinish ();
	}

	template <implementation_e I> feasability_e	is_feasible (pricing_e pricing)
	{
		num_iterations = 0;

		objective_value =
		artificial_value = 0.0f;

		if (num_artificials == 0)
			return feasible;

		c.set_elements (0.0f, 0, 0, num_variables, 1);

		phase1 = true;
		feasability_e ret = simplex (impl_type <I> (), pricing);
		
		if (ret == aborted)
			return aborted;

		return fabs (artificial_value) < 0.0001 ? feasible : infeasible;
	}

	template <implementation_e I> feasability_e	solve_optimal (pricing_e pricing)
	{
		num_iterations = 0;

		objective_value =
		artificial_value = 0.0f;

		feasability_e f = optimal;
		if (num_artificials > 0)
		{
			c.set_elements (0.0f, 0, 0, num_variables, 1);

			phase1 = true;
			f = simplex (impl_type <I> (), pricing);
		}

		if (f == optimal || f == unbounded)
		{
			if (fabs (artificial_value) < 0.0001)
			{
				m_pbo->activate ();
				float* tmp = m_pbo->get_buffer ();
				std::copy (costs.begin (), costs.end (), tmp);
				m_pbo->release_buffer ();
				c.set_elements (*m_pbo, 0, 0, num_variables, 1);
				m_pbo->deactivate ();

				cB.lookup (c, chi);

				phase1 = false;
				return simplex (impl_type <I> (), pricing);
			}
			else
				return infeasible;
		}
		else
			return aborted;
	}

	int get_result_vector (float* fx, int sz) const
	{
		int const max_sz = std::min (sz, x.bounds.width);

		if (solve_as == upper_bound_1)
			x.get_elements (0, 0, max_sz, 1, fx);
		else
			std::fill (fx, fx + sz, 0.0f);

		float* f_chi = new float [chi.bounds.width];
		float* f_xB = new float [xB.bounds.height];

		chi.get_elements (0, 0, chi.bounds.width, 1, f_chi);
		xB.get_elements (0, 0, 1, xB.bounds.height, f_xB);

		for (int i = chi.bounds.width - 1; i >= 0; -- i)
		{
			if (f_chi [i] < max_sz)
				fx [(int) f_chi [i]] = f_xB [i];
		}
/*
		if (SANITY_CHECK)
		{
			std::vector <float> f_x (x.bounds.width);

			x.get_elements (0, 0, x.bounds.width, x.bounds.height, &f_x.front ());
			for (int i = chi.bounds.width - 1; i >= 0; -- i)
				f_x [(int) f_chi [i]] = f_xB [i];

			matrix x_;
			x_.resize (1, x.bounds.width);
			x_.set_elements (&f_x.front (), 0, 0, 1, x.bounds.width);

			env::pengine->multiply (A, x_, d);
			env::pengine->subtract (d, b, tmpCol);
			tmpCol.get_elements (0, 0, 1, tmpCol.bounds.height, &f_x.front ());

			for (int i = b.bounds.height - 1; i >= 0; -- i)
			{
				if (fabs (f_x [i]) > 0.001)
					printf ("warning: element %i should be 0.0, but is %f\n", i + 1, f_x [i]);
			}
		}
*/
		delete [] f_chi;
		delete [] f_xB;

		return x.bounds.width;
	}

	float get_objective_value () const
	{
		return objective_value;
	}
	int get_num_iterations () const
	{
		return num_iterations;
	}

	void (*progress_cb) ();
};


/*
revised_simplex_impl <cpu::environment>* revised_simplex_cpu::impl = 0;

revised_simplex_cpu::revised_simplex_cpu () : pricing (steepest_edge)
{
	if (!impl)
	{
		static revised_simplex_impl <cpu::environment> cpu_impl;
		impl = &cpu_impl;
	}
}

void revised_simplex_cpu::precache () { impl->precache (); }

void revised_simplex_cpu::set_constraints (int rows, int cols, double const fa [], double const fb [], double const fc [],
															   double const* fu, double const* fl, bounds_e const* bb)
{ impl->set_constraints (rows, cols, fa, fb, fc, fu, fl, bb); }

feasability_e revised_simplex_cpu::is_feasible_simple () { return impl->is_feasible <simple> (pricing); }
feasability_e revised_simplex_cpu::is_feasible_regular () { return impl->is_feasible <regular> (pricing); }
feasability_e revised_simplex_cpu::is_feasible_bounded () { return impl->is_feasible <bounded> (pricing); }

feasability_e revised_simplex_cpu::solve_optimal_simple () { return impl->solve_optimal <simple> (pricing); }
feasability_e revised_simplex_cpu::solve_optimal_regular () { return impl->solve_optimal <regular> (pricing); }
feasability_e revised_simplex_cpu::solve_optimal_bounded () { return impl->solve_optimal <bounded> (pricing); }

int revised_simplex_cpu::get_result_vector (float* fx, int sz) const { return impl->get_result_vector (fx, sz); }
float revised_simplex_cpu::get_objective_value () const { return impl->get_objective_value (); }
int revised_simplex_cpu::get_num_iterations () const { return impl->get_num_iterations (); }
*/


revised_simplex_impl <gpu::environment>* revised_simplex_gpu::impl = 0;

revised_simplex_gpu::revised_simplex_gpu () : pricing (steepest_edge), progress_cb (0)
{
	if (!impl)
	{
		static revised_simplex_impl <gpu::environment> gpu_impl;
		impl = &gpu_impl;
	}
}

void revised_simplex_gpu::precache () { impl->precache (); }

bool revised_simplex_gpu::set_constraints (std::vector <double> const& vA, std::vector <double> const& vb, std::vector <double> const& vc, double const* fub, double const* flb)
{ return impl->set_constraints (vA, vb, vc, fub, flb); }

//feasability_e revised_simplex_gpu::is_feasible_simple () { impl->progress_cb = progress_cb; return impl->is_feasible <simple> (pricing); }
//feasability_e revised_simplex_gpu::is_feasible_regular () { impl->progress_cb = progress_cb; return impl->is_feasible <regular> (pricing); }
feasability_e revised_simplex_gpu::is_feasible_bounded () { impl->progress_cb = progress_cb; return impl->is_feasible <bounded> (pricing); }

//feasability_e revised_simplex_gpu::solve_optimal_simple () { impl->progress_cb = progress_cb; return impl->solve_optimal <simple> (pricing); }
//feasability_e revised_simplex_gpu::solve_optimal_regular () { impl->progress_cb = progress_cb; return impl->solve_optimal <regular> (pricing); }
feasability_e revised_simplex_gpu::solve_optimal_bounded () { impl->progress_cb = progress_cb; return impl->solve_optimal <bounded> (pricing); }

int revised_simplex_gpu::get_result_vector (float* fx, int sz) const { return impl->get_result_vector (fx, sz); }
float revised_simplex_gpu::get_objective_value () const { return impl->get_objective_value (); }
int revised_simplex_gpu::get_num_iterations () const { return impl->get_num_iterations (); }
