/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011

//	Author:	Simon Krueger
//	Date:	April 8, 2011
//	Modifications: 'load_mps' correctly handles equation-rows (before:
//					only inequality-rows were handled)

#include "arg_parser.h"
#include "lp_problem.h"
#include "lp_result.h"
#include "solvers.h"
#include "platform.h"
#include "timer.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <iterator>
#include <stdexcept>
#include <limits>
#include <cmath>
#include <cstdlib>

#ifdef USE_GLPK
#include "../glpk/include/glpk.h"
#endif

static float const inf = std::numeric_limits <float>::infinity ();
static float const fx_eps = 0;

bool read_matrix (std::istream& f, std::map < std::string, std::vector <double> >& m)
{
	float tmp;
	char expect = 'a';
	char rd_char = 0;

	size_t semis = 0;
	size_t cols = 0;
	std::vector <double>* v = 0;

	f >> rd_char;
	if (f.eof ())
		return false;

	do
	{
		switch (rd_char)
		{
		case '{':
			while (f.get () != '}')
				;
			break;
		case 'A':
		case 'b':
		case 'c':
		case 'l':
		case 'u':
		case 'f':
		{
			char tmp [2] = { rd_char, 0 };
			v = &m [tmp];

			if (expect != 'a')
			{
				xprint (1) ("error: literal expected\n");
				return false;
			}
			if (!v->empty ())
			{
				xprint (1) ("error: matrix already defined\n");
				return false;
			}

			expect = '=';
			break;
		}
		case '=':
			if (expect != '=')
			{
				xprint (1) ("error: '%c' expected\n", expect);
				return false;
			}

			expect = '[';
			break;
		case ';':
			if (expect != ']')
			{
				xprint (1) ("error: '%c' expected\n", expect);
				return false;
			}

			++ semis;
			if (semis == 1)
				cols = v->size ();

			if (v->size () / semis != cols)
				xprint (1) ("warning: unequal row lengths\n");
			break;
		case '[':
			if (expect != '[')
			{
				xprint (1) ("error: '%c' expected\n", expect);
				return false;
			}

			expect = ']';
			break;
		case ']':
			if (expect != ']')
			{
				xprint (1) ("error: '%c' expected\n", expect);
				return false;
			}
			return true;
		default:
			if (expect != ']')
			{
				xprint (1) ("error: float expected, %i found\n", rd_char);
				return false;
			}

			f.unget ();
			f >> tmp;
			v->push_back (tmp);
			break;
		}

		f >> rd_char;
	}
	while (!f.eof ());

	return false;
}

bool write_matrix (FILE* f, char const* name, std::map < std::string, std::vector <double> > const& m)
{
	fprintf (f, "%s = [ ", name);

	std::vector <double> const& mat = m.find (name)->second;
	for (std::vector <double>::const_iterator i = mat.begin (); i != mat.end (); ++ i)
		fprintf (f, "%f ", *i);
	fprintf (f, "]\n\n");

	return true;
}


bool lp_problem_t::save_txt (char const* fname)
{
	FILE* f = fopen (fname, "w");

	if (!write_matrix (f, "A", data)) return false;
	if (!write_matrix (f, "b", data)) return false;
	if (!write_matrix (f, "c", data)) return false;
	if (!write_matrix (f, "u", data)) return false;
	if (!write_matrix (f, "l", data)) return false;

	std::vector <double> const& u = data.find ("u")->second;
	std::vector <double> const& l = data.find ("l")->second;

	fprintf (f, "f = [ ");
	for (size_t i = 0; i != u.size (); ++ i)
	{
		int b = no_bounds;
		if (u [i] < std::numeric_limits <double>::infinity ())
			b = upper_bound;
		if (l [i] > -std::numeric_limits <double>::infinity ())
		{
			if (b == upper_bound)
				b = double_bound;
			else
				b = lower_bound;
		}
		fprintf (f, "%i ", b);
	}
	fprintf (f, "]\n");

	fclose (f);
	return true;
}

bool lp_problem_t::load_txt (char const* fname)
{
	std::ifstream f (fname);

	while (read_matrix (f, data))
		;

	if (data.find ("A") == data.end () ||
		data.find ("b") == data.end () ||
		data.find ("c") == data.end ())
		return false;

	//	clean the matrix from the identity stuff etc
	{
		size_t cols = data ["c"].size ();
		size_t rows = data ["b"].size ();

		if (data ["A"].size () != cols * rows)
			return false;

		bool has_id = false;
		if (cols > rows)
		{
			std::vector <double>& A = data ["A"];

			has_id = true;
			//	check if last 'rows X rows submatrix' in A is Id
			for (size_t y = 0; y < rows; ++ y)
			for (size_t x = 0; x < rows; ++ x)
			{
				size_t const idx = y * (cols) + x + (cols - rows);
				if ((A [idx] == 1.0) != (x == y))
				{
					has_id = false;
					break;
				}
			}
		}

		if (has_id)
		{
			std::vector <double>& A = data ["A"];
			std::vector <double>& c = data ["c"];

			c.erase (c.begin () + (cols - rows), c.end ());

			for (int i = rows - 1; i >= 0 ; -- i)
				A.erase (A.begin () + cols * i + cols - rows, A.begin () + cols * i + cols);
		}

		char const* vecs [] = { "u", "l", "f" };
		for (size_t i = 0; i < sizeof (vecs) / sizeof (vecs [0]); ++ i)
			if (data.find (vecs [i]) != data.end ())
			{
				std::vector <double>& v = data [vecs [i]];

				if (v.size () > cols - rows)
				{
					if (has_id) v.erase (v.begin () + (cols - rows), v.end ());
				}
				else
					v.insert (v.end (), (cols - rows) - v.size (), 0);
			}
	}

	return true;
}

bool lp_problem_t::load_mps (char const* fname)
{
#ifndef USE_GLPK
	return false;
#else
	bool		ret = true;
	glp_prob*	lp = glp_create_prob ();

	std::vector <double>&	fa (data ["A"]);
	std::vector <double>&	fb (data ["b"]);
	std::vector <double>&	fc (data ["c"]);
	std::vector <double>&	fu (data ["u"]);
	std::vector <double>&	fl (data ["l"]);
	std::vector <double>&	ff (data ["f"]);

	//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
	//	load the MPS
	//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
	if (glp_read_mps (lp, GLP_MPS_FILE, 0, fname) == 0)
	{
		//int type = glp_get_obj_dir (lp);	//	GLP_MIN, GLP_MAX
		int rows = glp_get_num_rows (lp);
		int cols = glp_get_num_cols (lp);

		std::vector <int>		indices (cols + 1);
		std::vector <double>	values  (cols + 1);

		//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
		//	Modification by Simon Krueger, University of Illinois
		//	Count the number of rows we will create when importing
		//	since some rows (namely rows that are an equation) will
		//	be imported as more than one (i.e., two) inequality-rows.
		int final_rows = 0;
		for (int i = 1; i <= rows; ++i)
		{
			int const row_type = glp_get_row_type (lp, i);	//	GLP_FR GLP_LO GLP_UP GLP_DB GLP_FX
			if (row_type == GLP_FR)
				continue;  // not counting this type
			else if (row_type == GLP_FX)
				final_rows += 2;
			else if (row_type == GLP_LO || row_type == GLP_UP)
				final_rows += 1;
			else
				assert(false);  // disallowed row
		}
		
		fa.resize (final_rows * cols);
		fc.resize (cols);
		fu.resize (cols);
		fl.resize (cols);
		ff.resize (cols);
		fb.resize (final_rows);
		//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

		char const* nobj = glp_get_obj_name (lp);

		//	NOTE: We only allow GLP_LO and GLP_UP as possible row types
		//		and filter our the objective-row separately
		for (int i = 1, crow = 0; i <= rows && ret; ++ i)
		{
			int const	row_type = glp_get_row_type (lp, i);	//	GLP_FR GLP_LO GLP_UP GLP_DB GLP_FX
			double		sign = 0.0;

			char const* nrow = glp_get_row_name (lp, i);

			switch (row_type)
			{
				case GLP_LO: sign = -1.0; fb [crow] = -glp_get_row_lb (lp, i); break;
				case GLP_UP: sign =  1.0; fb [crow] =  glp_get_row_ub (lp, i); break;
				case GLP_FR:
					if (nobj && nrow && strcmp (nobj, nrow) == 0)
						continue;	// for-loop
				//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
				//	Modification by Simon Krueger, University of Illinois
				//	Handle fixed rows, i.e. equations, by inserting them as
				//	two inequality-rows.
				case GLP_FX:
				{
					/* GLP_FX can be represented by adding both a lower and upper bound row
					 */
					// continue;
					int const n = glp_get_mat_row(lp, i, &indices [0], &values [0]);
					
					/* If the row is an equality constraint (i.e., the corresponding
					 * auxiliary variable is of fixed type), only the parameter lb is used
					 * while the parameter ub is ignored. */
					const double row_bound = glp_get_row_lb(lp, i);
					
					fb [crow] = -row_bound + fx_eps;
					for (int j = 1; j <= n; ++ j)
					{
						if (values[j] == 0)
							continue;
						fa[crow * cols + indices[j] - 1] = -values[j];
					}
					++ crow;

					fb [crow] = row_bound + fx_eps;
					for (int j = 1; j <= n; ++ j)
						fa[crow * cols + indices[j] - 1] = values[j];

					++ crow;
					continue;
					//	'continue' skips the 'fill A' lines after this switch, see below
				}
				//	* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
				default:	ret = false; break;
			}

			if (ret)
			{
				//	Fill the current row of A with values. The sign is flipped
				//	automatically for the whole row in order to only effectively
				//	produce 'less-or-equal-than'-rows.
				int const n = glp_get_mat_row (lp, i, &indices [0], &values [0]);
				for (int j = 1; j <= n; ++ j)
					fa [crow * cols + indices [j] - 1] = sign * values [j];

				++ crow;
			}
		}

		for (int i = 1; i <= cols && ret; ++ i)
		{
			int col_type = glp_get_col_type (lp, i);	//	GLP_FR GLP_LO GLP_UP GLP_DB GLP_FX
			
			switch (col_type)
			{
				case GLP_FR: ff [i - 1] = no_bounds; break;
				case GLP_LO: ff [i - 1] = lower_bound;  fl [i - 1] = glp_get_col_lb (lp, i); break;
				case GLP_UP: ff [i - 1] = upper_bound;  fu [i - 1] = glp_get_col_ub (lp, i); break;
				case GLP_DB: ff [i - 1] = double_bound; fl [i - 1] = glp_get_col_lb (lp, i);
				                                        fu [i - 1] = glp_get_col_ub (lp, i); break;
				default: ret = false; break;
			}

			//fc [i - 1] = glp_get_obj_coef (lp, i); // maximize
			fc [i - 1] = -glp_get_obj_coef (lp, i); // minimize
		}
		
		printf ("A: %i rows, %i columns\n", final_rows, cols);
	}

	glp_delete_prob (lp);
	return ret;
#endif
}

void lp_problem_t::set_random (unsigned seed, int rows, int cols, bool negative_b)
{
	float const MAX_ROW_UB = 10.0;
	float const MAX_OBJ_COEFF = 30.0;
	float const MIN_MATRIX_COEFF = -20.0;	//	-20
	float const MAX_MATRIX_COEFF = 20.0;
	float const PERCENT_NEGATIVE_B = 50.0; // 50.0
	float const PERCENT_ZERO_A = 97.0;

	srand (seed);

	std::vector <double>&	fa (data ["A"]);
	std::vector <double>&	fb (data ["b"]);
	std::vector <double>&	fc (data ["c"]);
	std::vector <double>	fs (rows);
	std::vector <double>	fz (rows * cols);

	fa.resize (rows * cols);
	fb.resize (rows);
	fc.resize (cols);

	xprint (4) ("Generating random problem with:\n"
				"\tseed = %u\n"
				"\t%i rows\n"
				"\t%i columns\n"
				"\tmax row bound = %f\n"
				"\tmax objective function coefficient = %f\n"
				"\tmatrix coefficients in [%f,%f]\n"
				"\t%0.1f%% negative b\n"
				"\t%0.1f%% zeros in A\n",
				seed, rows, cols, MAX_ROW_UB, MAX_OBJ_COEFF, MIN_MATRIX_COEFF, MAX_MATRIX_COEFF, negative_b ? PERCENT_NEGATIVE_B : 0.0, PERCENT_ZERO_A);

	if (SAT_ONLY)
	{
		xprint (4) ("Using zero objective function.\n");
	}
	else
	{
		xprint (4) ("Using the specified objective function.\n");
	}

	for (int i = 0; i < rows; ++ i)
	{
		fb [i] = rand () / static_cast <double> (RAND_MAX) * MAX_ROW_UB;
		fs [i] = true;
	}

	if (negative_b)
	{
		for (int i = 0; i < rows; ++ i)
			fs [i] = rand () / static_cast <double> (RAND_MAX);
	}

	for (int i = 0; i < rows; ++ i)
		fc [i] = rand () / static_cast <double> (RAND_MAX) * MAX_OBJ_COEFF;

	for (int i = 0; i < rows; ++ i)
		for (int j = 0; j < cols; ++ j)
			fz [i * cols + j] = rand () / static_cast <double> (RAND_MAX);

	double const COEFF_RANGE = MAX_MATRIX_COEFF - MIN_MATRIX_COEFF;
	for (int i = 0; i < rows; ++ i)
	{
		for (int j = 0; j < cols; ++ j)
		{
			int const	idx = i * cols + j;
			bool const	zero = fz [idx] < (PERCENT_ZERO_A / 100.0);
			if (zero)
				fa [idx] = 0.0f;
			else
				fa [idx] = MIN_MATRIX_COEFF + rand () / static_cast <double> (RAND_MAX) * COEFF_RANGE;
		}

		if (fs [i] < (PERCENT_NEGATIVE_B / 100.0))
		{
			for (int j = 0; j < cols; ++ j)
			{
				if (fa [i * cols + j] != 0.0f)
					fa [i * cols + j] = -fa [i * cols + j];
			}
			fb [i] = -fb [i];
		}
	}
}

void lp_problem_t::set_random_bnd (unsigned seed, int rows, int cols, bool negative_b)
{
	float const MIN_COL_UB = 1.0;
	float const MAX_COL_UB = 1.0;
	float const MIN_COL_LB = 0.0;
	float const MAX_COL_LB = 0.0;
	float const PERCENT_UB = 100.0;
	float const PERCENT_LB = 100.0;

	std::vector <double>&	fu (data ["u"]);
	std::vector <double>&	fl (data ["l"]);
	std::vector <double>	ff;

	fu.resize (cols);
	fl.resize (cols);
	ff.resize (cols, no_bounds);

	set_random (seed, rows, cols, negative_b);

	if (RANDOM_BOUNDS)
	{
		xprint (4) ("Using %f%% upper bounds in [%f,%f]\n"
					"Using %f%% lower bounds in [%f,%f]\n",
					PERCENT_UB, MIN_COL_UB, MAX_COL_UB, PERCENT_LB, MIN_COL_LB, MAX_COL_LB);

		for (int i = 0; i < cols; ++ i)
		{
			if (rand () / static_cast <double> (RAND_MAX) <= (PERCENT_UB / 100.0))
				ff [i] = upper_bound;
		}

		for (int i = 0; i < cols; ++ i)
		{
			if (rand () / static_cast <double> (RAND_MAX) <= (PERCENT_LB / 100.0))
				ff [i] = (ff [i] == upper_bound) ? double_bound : lower_bound;
		}

		double const UB_RANGE = MAX_COL_UB - MIN_COL_UB;
		for (int i = 0; i < cols; ++ i)
			fu [i] = MIN_COL_UB + rand () / static_cast <double> (RAND_MAX) * UB_RANGE;

		double const LB_RANGE = MAX_COL_UB - MIN_COL_UB;
		for (int i = 0; i < cols; ++ i)
			fl [i] = MIN_COL_LB + rand () / static_cast <double> (RAND_MAX) * LB_RANGE;

		for (int i = 0; i < cols; ++ i)
		{
			switch ((int) ff [i])
			{
				case no_bounds: fu [i] = +inf; fl [i] = -inf; break;
				case lower_bound: fu [i] = +inf; break;
				case upper_bound: fl [i] = -inf; break;
			}
		}
	}
	else
	{
		xprint (4) ("Using upper bounds of 1.0\n"
					"Using lower bounds of 0.0\n");

		for (int i = 0; i < cols; ++ i)
		{
			fu [i] = 1.0;
			fl [i] = 0.0;
		}
	}
}

void lp_problem_t::set_random_old (unsigned seed, int rows, int cols, bool negative_b)
{	//	uses the random problem generator from an earlier version that is not quite correct
	float const MAX_ROW_UB = 10.0;
	float const MAX_OBJ_COEFF = 30.0;
	float const MIN_MATRIX_COEFF = -20.0;
	float const MAX_MATRIX_COEFF = 20.0;
	float const PERCENT_NEGATIVE_B = 50.0;
	float const PERCENT_ZERO_A = 50.0;

	srand (seed);

	std::vector <double>&	fa (data ["A"]);
	std::vector <double>&	fb (data ["b"]);
	std::vector <double>&	fc (data ["c"]);
	std::vector <double>	fs (rows);


	fa.resize (rows * cols);
	fb.resize (rows);
	fc.resize (cols);

	xprint (4) ( "Generating random problem with:\n"
			"\tseed = %u\n"
			"\t%i rows\n"
			"\t%i columns\n"
			"\tmax row bound = %f\n"
			"\tmax objective function coefficient = %f\n"
				"\tmatrix coefficients in [%f,%f]\n"
				"\t%0.1f%% negative b\n"
				"\t%0.1f%% zeros in A\n",
				seed, rows, cols, MAX_ROW_UB, MAX_OBJ_COEFF, MIN_MATRIX_COEFF, MAX_MATRIX_COEFF, negative_b ? PERCENT_NEGATIVE_B : 0.0, PERCENT_ZERO_A);

	if (SAT_ONLY)
	{
		xprint (4) ("Using zero objective function.\n");
	}
	else
	{
		xprint (4) ("Using the specified objective function.\n");
	}

	for (int i = 0; i < rows; ++ i)
	{
		fb [i] = rand () / static_cast <double> (RAND_MAX) * MAX_ROW_UB;
		if (ALLOW_NEG_ROW_UB)
			fs [i] = rand () > RAND_MAX / 2;
		else
		{
			fs [i] = true;
			rand ();
		}
	}

	for (int i = 0; i < rows; ++ i)
		fc [i] = rand () / static_cast <double> (RAND_MAX) * MAX_OBJ_COEFF;


	double const COEFF_RANGE = MAX_MATRIX_COEFF - MIN_MATRIX_COEFF;
	for (int i = 0; i < rows; ++ i)
	{
		for (int j = 0; j < cols; ++ j)
		{
			int const	idx = i * cols + j;
			bool const zero = rand () <= RAND_MAX / 2;
			if (zero)
				fa [idx] = 0.0f;
			else
				fa [idx] = MIN_MATRIX_COEFF + rand () / static_cast <double> (RAND_MAX) * COEFF_RANGE;
		}

		if (!fs [i])
		{
			for (int j = 0; j < cols; ++ j)
				fa [i * cols + j] = -fa [i * cols + j];
			fb [i] = -fb [i];
		}
	}
}

bool lp_problem_t::solve (solver_t const& solver, char const* prefix, lp_result_t& res) const
{
	if (solver.skip)
		return true;

	char* tmp = new char [strlen (prefix) + strlen (solver.name) + 4];
	strcpy (tmp, prefix);
	strcat (tmp, " ");
	strcat (tmp, solver.name);
	
	bool ret = false;
	try 
	{ 
		timer::prefix (tmp);
		res.solver = &solver;
		solver.solve (*this, res);

		timer::insert_measurement ("iterations", static_cast <double> (res.it_cnt));
		ret = true;
	}
	catch (...)
	{
	}
	
	delete [] tmp;
	return ret;
}


