/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011


#pragma once

#include <vector>
#include <cmath>
#include <cassert>


#include "arg_parser.h"



enum feasability_e
{
	undefined = -2,
	aborted = -1,
	infeasible,
	feasible,
	unbounded,
	optimal
};

struct solver_t;

struct lp_result_t
{
	solver_t const*		solver;
	feasability_e		status;
	int					it_cnt;
	float				obj_val;
	std::vector <float> x;

	lp_result_t ()
		:	status (undefined),
			it_cnt (0),
			obj_val (0.0)
	{
	}

	bool operator== (lp_result_t const& rhs) const
	{
		bool ret = true;
		
		if (status != rhs.status)
		{
			xprint (4) ("  status differs: %i != %i\n", status, rhs.status);
			ret = false;
		}

		if ((status == feasible || status == optimal) && ret)
		{
			if (!eqdigits <4> (obj_val, rhs.obj_val))
			{
				xprint (4) ("  objective value differs: %f != %f\n", obj_val, rhs.obj_val);
				ret = false;
			}

			if (status == optimal && ret)
			{
				for (size_t j = 0; j < std::min (x.size (), rhs.x.size ()); j++)
				{
					if (!eqdigits <4> (x [j], rhs.x [j]))
					{
						xprint (4) ("  x%03lu differs: %f != %f\n", j + 1, x [j], rhs.x [j]);
						ret = false;
					}
				}
			}
		}

		return ret;
	}

	bool operator!= (lp_result_t const& rhs) const
	{
		return !(*this == rhs);
	}

	void reset ()
	{
		status = undefined;
		it_cnt = 0;
		obj_val = 0.0;
		x.clear ();
	}

	void print (bool info, bool result_vector) const
	{
		if (info)
		{
			char const* result = "unknown status";
			switch (status)
			{
			case optimal:
				result = "optimal solution found";
				break;
			case feasible:
				result = "feasible solution found";
				break;
			case infeasible:
				result = "problem has no feasible solution";
				break;
			case unbounded:
				result = "problem is unbounded";
				break;
			case aborted:
				result = "solution undefined (max iterations)";
				break;
			case undefined:
				result = "solution undefined";
				break;
			}

			xprint (4) ("Result: %s\n", result);
			xprint (4) ("Optimal solution = %f\n", obj_val);
			xprint (4) ("Number of iterations: %i\n", it_cnt);
		}

		if (result_vector)
		{
			for (size_t j = 0; j < x.size (); j++)
				xprint (4) ("x%03lu = %f\n", j + 1, x [j]);
			xprint (4) ("\n");
		}
	}

private:

	template <int significant_digit>
	static bool eqdigits (float f1, float f2)
	{
		if (std::abs (f1 - f2) < 0.00001)
			return true;
		else
		{
			float const	m = pow (10.0f, significant_digit - (int) ceil (log10 (std::abs (f1))));
			return (int) (f1 * m) == (int) (f2 * m);
		}
	}
};
