/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February17, 2010

#include "arg_parser.h"
#include "platform.h"
#include "lp_problem.h"
#include "lp_result.h"
#include "solvers.h"
#include "logger.h"
#include "timer.h"

#include <numeric>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <ctime>
#include <cstdlib>


void progress ()
{
	static char const* progress [] = { "-", "\\", "|", "/" };
	static int cur_char = 0;

	xprint (2) ("\b");
	xprint (2) ("%c", *progress [cur_char]);
	cur_char = (cur_char + 1) % (sizeof (progress) / sizeof (progress [0]));
}

template <size_t C>
bool solve_all (solver_t const (&slv) [C], char const* prefix, lp_problem_t const& prob, lp_result_t (&res) [C])
{
	bool r = true;

	xprint (2) (" ");
	for (size_t i = 0; i < C; ++ i)
	{
		res [i].reset ();
		if (!prob.solve (slv [i], prefix, res [i]))
		{
			FILE* f = fopen (PROCESS_FILE, "a+");
			fprintf (f, "error solving problem %s, %s, %s, %s with %s\n",
						prefix,
						RUN_BOUNDED ? "bounded" : "unbounded",
						ALLOW_NEG_ROW_UB ? "negative b's" : "positive b's",
						SAT_ONLY ? "feasible" : "optimal",
						slv [i].name);
			fclose (f);

			r = false;
			break;
		}

		progress ();
	}
	xprint (2) ("\b");

	return r;
}


int main (int argc, char const* argv [])
{
	printf ("gpu_simplex  Copyright (C) 2011  Jakob Bieling <bieling at cs.uni-bonn.de>\n"
			"This program is free software: you can redistribute it and/or modify\n"
			"it under the terms of the GNU General Public License as published by\n"
			"the Free Software Foundation, either version 3 of the License, or\n"
			"(at your option) any later version.\n"
			"\n"
			"This program is distributed in the hope that it will be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
			"GNU General Public License for more details.\n"
			"\n"
			"You should have received a copy of the GNU General Public License\n"
			"along with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n");
	
	if (!set_realtime_priority ())
		printf ("error: unable to change to a higher priority\n");

	switch (read_args (argc - 1, argv + 1))
	{
	case -1:
		printf ("error: unknown argument\n");
	case 1:
		print_help ();
		return 0;
	}


	solver_t solver [] =
	{
		{ !USE_METHOD [imp_gpu],    "gpu",		solve_gpu },
		{ !USE_METHOD [imp_glpk],   "glpk",		solve_glpk },
	};

	lp_problem_t		vs;
	lp_result_t			res [sizeof (solver) / sizeof (solver [0])] = {};
	FILE*				f;



#if defined(_DEBUG) || !defined(NDEBUG)
	SKIP_WARMUP = true;
#endif

 	if (PROBLEM_FILE)
	{
		f = fopen (PROBLEM_FILE, "r");
		if (f) fclose (f);

		if (!f)
		{
			xprint (1) ("error: unable to open file\n");
			return 1;
		}

		if (stricmp (PROBLEM_FILE + strlen (PROBLEM_FILE) - 3, "mps") == 0)
		{
			if (!vs.load_mps (PROBLEM_FILE))
			{
				xprint (1)  ("error: unable to read MPS file\n");
				return 1;
			}
		}
		else if (stricmp (PROBLEM_FILE + strlen (PROBLEM_FILE) - 6, "mps.gz") == 0)
		{
			xprint (1)  ("error: compressed MPS cannot be read\n");
			return 1;
		}
		else
		{
			if (!vs.load_txt (PROBLEM_FILE))
			{
				xprint (1)  ("error: unable to read file\n");
				return 1;
			}
		}

		solve_all (solver, PROBLEM_FILE, vs, res);
		log_benchmark (PROCESS_FILE, 0, 0, res);

		f = fopen (TIMINGS_FILE, "a+");
		timer::print_all_times (f);
		fclose (f);
		
		f = fopen (STATISTICS_FILE, "a+");
		timer::print_statistics (f);
		fclose (f);
	}
	else
	{
		xprint (2)  ("Benchmarking:\n"
					"    size = [%i,%i], step %i\n"
					"    seed = [%i,%i]\n"
					"    %s, %s, %s\n\n",
					PROBLEM_SIZE [0], PROBLEM_SIZE [1], PROBLEM_SIZE [2],
					PROBLEM_SEED [0], PROBLEM_SEED [1],
					RUN_BOUNDED ? "bounded" : "unbounded",
					ALLOW_NEG_ROW_UB ? "negative b's" : "positive b's",
					SAT_ONLY ? "feasible" : "optimal");

		if (!SKIP_WARMUP)
		{
			xprint (2) ("Warm-up");
			{
				for (int seed = 0; seed < 30; ++ seed)
				{
					vs.set_random_bnd (0, 4, 4, ALLOW_NEG_ROW_UB);
					solve_all (solver, "seed", vs, res);
				}

				Sleep (100);
				timer::clear_timers ();
			}
			xprint (2) (" done\n\n");
		}

		bool quit = false;
		char tmp [32];
		for (int s = PROBLEM_SIZE [0]; PROBLEM_SIZE [2] > 0 ? s <= PROBLEM_SIZE [1] : s >= PROBLEM_SIZE [1]; s += PROBLEM_SIZE [2])
		{
			sprintf (tmp, "%03i", s);

			for (int seed = PROBLEM_SEED [0]; seed <= PROBLEM_SEED [1]; ++ seed)
			{
				timer::current_run (seed);

				vs.set_random_bnd (seed, s, s, ALLOW_NEG_ROW_UB);
				quit = !solve_all (solver, tmp, vs, res);
				log_benchmark (PROCESS_FILE, s, seed, res);

//				if (USE_METHOD [imp_gpu])
//				{
//					glFinish ();
//					if (RUN_CALCS)
//						Sleep (60000);
//				}

				f = fopen ("~timings.tmp", "w");
				timer::print_all_times (f);
				fclose (f);
			}

			f = fopen (TIMINGS_FILE, "a+");
			timer::print_all_times (f);
			fclose (f);
			
			f = fopen (STATISTICS_FILE, "a+");
			timer::print_statistics (f);
			fclose (f);
			
			timer::clear_timers ();
			remove ("~timings.tmp");
/*
			if (_kbhit ())
			{
				//beep ();
				system ("pause");
			}
*/
			if (quit)
				break;
		}
	}
}
