/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011


#include "arg_parser.h"
#include "lp_problem.h"
#include "lp_result.h"
#include "solvers.h"
#include "platform.h"
#include "timer.h"

#include <vector>
#include <stdexcept>

#ifdef USE_GLPK
#include "../glpk/include/glpk.h"
#endif

void solve_glpk (lp_problem_t const& prob, lp_result_t& res)
{
#ifndef USE_GLPK
	not_implemented (res.solver->name);
#else
	glp_prob*		lp = glp_create_prob ();
	glp_smcp		cp;

	std::vector <double> const& A = prob ["A"];
	std::vector <double> const& b = prob ["b"];
	std::vector <double> const& c = prob ["c"];
	
	size_t const	rows = b.size ();
	size_t const	cols = c.size ();
	
	timer::start ("with_upload");
	{
		xprint (4) ("GLPK Loading constraints ... ");
		
		timer::start ("upload");
		{
			char tmp [16];
			
			glp_set_prob_name (lp, "Random");
			glp_set_obj_dir (lp, GLP_MAX);
			
			glp_init_smcp (&cp);
			//			cp.it_lim = 1;
			//			cp.out_frq = 1;
			cp.msg_lev = GLP_MSG_ERR;	// GLP_MSG_OFF GLP_MSG_ERR GLP_MSG_ON GLP_MSG_ALL
			cp.meth = GLP_PRIMAL;		// GLP_PRIMAL GLP_DUAL GLP_DUALP
			cp.presolve = GLP_OFF;		// GLP_MSG_ON GLP_MSG_OFF
			
			if (HEURISTIC == 0)
			{
				cp.pricing = GLP_PT_STD;
				cp.r_test = GLP_RT_STD;
			}
			
			glp_add_rows (lp, rows);
			for (size_t i = 1; i <= rows; ++ i)
			{
				glp_set_row_name (lp, i, itoa (i, tmp, 10));
				glp_set_row_bnds (lp, i, GLP_UP, 0.0, b [i - 1]);	//	<= b_i
																	//				glp_set_row_bnds (lp, i, GLP_LO, b [i - 1].at (i - 1), 0.0);	//	>= b_i
			}
			
			glp_add_cols (lp, cols);
			tmp [0] = 'x';
			for (size_t j = 1; j <= cols; ++ j)
			{
				int const i = j - 1;
				
				itoa (j, tmp + 1, 10);
				glp_set_col_name (lp, j, tmp);
				
				if (RUN_BOUNDED && prob.has ("u") && prob.has ("l"))
				{
					std::vector <double> const& u = prob ["u"];
					std::vector <double> const& l = prob ["l"];

					if (!isinf (l [i]) && !isinf (u [i]))
						glp_set_col_bnds (lp, j, GLP_DB, l [i], u [i]);
					else if (!isinf (l [i]))
						glp_set_col_bnds (lp, j, GLP_LO, l [i], 0.0);
					else if (!isinf (u [i]))
						glp_set_col_bnds (lp, j, GLP_UP, 0.0, u [i]);
					else
						glp_set_col_bnds (lp, j, GLP_FR, 0.0, 0.0);
				}
				else
					glp_set_col_bnds (lp, j, GLP_LO, 0.0, 0.0);
				
				if (!SAT_ONLY)
					glp_set_obj_coef (lp, j, c [i]);
				else
					glp_set_obj_coef (lp, j, 0.0);
			}
			
			timer::start ("upload_conversion");
			int* ia = new int [1 + cols * rows];
			int* ja = new int [1 + cols * rows];
			double* ar = new double [1 + cols * rows];
			
			int ne = 1;
			for (size_t i = 0; i < rows; ++ i)
				for (size_t j = 0; j < cols; ++ j)
				{
					//				ia [1 + i * cols + j] = i + 1;
					//				ja [1 + i * cols + j] = j + 1;
					
					if (double const a = A [i * cols + j])
					{
						ia [ne] = i + 1;
						ja [ne] = j + 1;
						ar [ne] = a;
						++ ne;
					}
				}
			timer::stop ("upload_conversion");
			
			glp_load_matrix (lp, ne - 1, ia, ja, ar);
			
			delete [] ia;
			delete [] ja;
			delete [] ar;
		}
		timer::stop ("upload");
		xprint (4) ("done\n");
		
		xprint (4)  ("GLPK Solving ... ");
		if (RUN_CALCS)
		{
			timer::start ("*simplex");
			glp_term_out (GLP_OFF);
			
			try
			{
				glp_simplex (lp, &cp);
			}
			catch (...)
			{
			}
			
			timer::stop ("*simplex");
			/*
			 std::vector <double> xB (4);
			 std::vector <double> x (4);
			 xB [0] = glp_get_row_prim (lp, 1);
			 xB [1] = glp_get_row_prim (lp, 2);
			 xB [2] = glp_get_row_prim (lp, 3);
			 xB [3] = glp_get_row_prim (lp, 4);
			 x [0] = glp_get_col_prim (lp, 1);
			 x [1] = glp_get_col_prim (lp, 2);
			 x [2] = glp_get_col_prim (lp, 3);
			 x [3] = glp_get_col_prim (lp, 4);*/
		}
		xprint (4) ("done\n");
	}
	timer::stop ("with_upload");
	
	//	GLPK solution
	if (RUN_CALCS)
	{
		switch (glp_get_status (lp))
		{
			case GLP_OPT:
				if (SAT_ONLY)
					res.status = feasible;
				else
					res.status = optimal;
				break;
			case GLP_FEAS:
				res.status = feasible;
				break;
			case GLP_INFEAS:
				res.status = undefined;
				break;
			case GLP_NOFEAS:
				res.status = infeasible;
				break;
			case GLP_UNBND:
				res.status = unbounded;
				break;
			case GLP_UNDEF:
				res.status = aborted;
				break;
		}
		
		res.obj_val = static_cast <float> (glp_get_obj_val (lp));
		//	TODO: get the iteration count properly?
		res.it_cnt = *reinterpret_cast<int*> (reinterpret_cast <unsigned char*> (lp) + 168);
		if (res.status == feasible || res.status == optimal)
		{
			res.x.resize (cols);
			for (size_t j = 1; j <= cols; j++)
				res.x [j - 1] = static_cast <float> (glp_get_col_prim (lp, j));
		}
	}
	xprint (4) ("\n");
	
	glp_delete_prob (lp);
}



void cpp_throw ()
{
	throw std::runtime_error ("fatal: GLPK aborted.");
}

extern "C"
{
	
	void glpk_abort ()
	{
		cpp_throw ();
	}
	
	void glpk_callback (glp_prob* lp, struct csa* c)
	{
#if defined(_DEBUG) && 0
		std::vector <double> x (c->m + c->n);
		std::vector <double> xB (c->m);
		
		for (size_t j = 1; j <= x.size (); j++)
		{
			if (c->stat [j] == GLP_NL)
				x [j - 1] = c->lb [j];
			else if (c->stat [j] == GLP_NU)
				x [j - 1] = c->ub [j];
		}
		for (int j = 1; j <= c->m; j++)
		{
			xB [j - 1] = c->bbar [j];
			x [c->head [j] - 1] = xB [j - 1];
		}
		
		
		//	print augmented matrix
		std::stringstream ss;
		ss.flags (std::ios::right);
		ss.width (8);
		ss.precision (4); 
		ss << std::fixed << std::setprecision (4);
		
		for (int r = 1; r <= c->m; ++ r)
		{
			int row_start = c->N_ptr [r];
			int row_end = c->N_ptr [r + 1];
			
			int i = 1;
			while (i != c->m + 1)
			{
				ss.width (8);
				if (i == c->N_ind [row_start])
				{
					ss << c->N_val [row_start] << " ";
					++ row_start;
				}
				else
				{
					ss << 0.0 << " ";
				};
				++ i;
			}
			ss << std::endl;
		}
		std::string const& str (ss.str ());
		fwrite (str.c_str (), 1, str.size (), stdout);
		
		xB [0] = glp_get_row_prim (lp, 1);
		xB [1] = glp_get_row_prim (lp, 2);
		xB [2] = glp_get_row_prim (lp, 3);
		xB [3] = glp_get_row_prim (lp, 4);
		
		return;
#endif
	}
#endif
}
