/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	May 4, 2011

#include <iostream>
#include <string>
#include <map>

#include "arg_parser.h"
#include "lp_problem.h"
#include "lp_result.h"
#include "solvers.h"
#include "platform.h"
#include "timer.h"

#ifdef USE_GPU
#include "engine_gpu/revised_simplex.h"
#endif

void solve_gpu (lp_problem_t const& prob, lp_result_t& res)
{
#ifndef USE_GPU
	not_implemented (res.solver->name);
#else
	static revised_simplex_gpu rs;
	feasability_e	ret;

	if (HEURISTIC == 0)
		rs.pricing = dantzig;
	else
		rs.pricing = steepest_edge;

	timer::start ("with_upload");
	{
		xprint (4) ("GPU Loading constraints ... ");
		timer::start ("upload");
		{
			if (RUN_BOUNDED && prob.has ("u") && prob.has ("l"))
				rs.set_constraints (prob ["A"], prob ["b"], prob ["c"],
									&prob ["u"].front (), &prob ["l"].front ());
			else
				rs.set_constraints (prob ["A"], prob ["b"], prob ["c"]);
		}
		timer::stop ("upload");
		xprint (4) ("done\n");

		xprint (4) ("GPU Solving ... ");

		void progress ();
		rs.progress_cb = progress;
		if (RUN_CALCS)
		{
			if (SAT_ONLY)
			{
				timer::start ("*simplex");
				ret = rs.is_feasible_bounded ();
				timer::stop ("*simplex");
			}
			else
			{
				timer::start ("*simplex");
				ret = rs.solve_optimal_bounded ();
				timer::stop ("*simplex");
			}
		}
		xprint (4) ("done\n");
	}
	timer::stop ("with_upload");

	//	GPU solution
	if (RUN_CALCS)
	{
		res.status = ret;
		res.obj_val = rs.get_objective_value ();
		res.it_cnt = rs.get_num_iterations ();

		if (res.status == feasible || res.status == optimal)
		{
			res.x.resize (prob ["c"].size ());
			rs.get_result_vector (&res.x.front (), res.x.size ());
		}
	}
#endif
}
