/**	gpu_simplex, a simplex solver using the GPU
 *	Copyright (C) 2011  Jakob Bieling, University of Bonn <bieling@cs.uni-bonn.de>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//	Author:	Jakob Bieling
//	Date:	February 03, 2009


#include "timer.h"

#ifdef _WIN32
	#include <windows.h>
	#include <GL/gl.h>
	#ifdef min
	#undef min
	#endif

	#ifdef max
	#undef max
	#endif
#elif defined (__APPLE__)
	#include <mach/mach.h>
	#include <mach/mach_time.h>
	#include <OpenGL/gl.h>
#else
	#include <ctime>
	#include <GL/gl.h>
#endif

#include <climits>
#include <utility>
#include <vector>
#include <string>
#include <list>
#include <map>
#include <cmath>


namespace timer
{

	struct functimer
	{
		friend void current_run (int r);

	private:
		static double			freq;	// frequency of high-resolution timer
		static long long		mark;
		static int				current_run_id;

		std::vector <int>		run_ids;
		std::vector <long long>	previous;

		unsigned long			runs;
		long long				accum;

		long long				begin,
								end;

		static inline void query_timer (long long& res)
		{
#ifdef _WIN32
			QueryPerformanceCounter ((LARGE_INTEGER*) &res);
#elif defined (__APPLE__)
			res = mach_absolute_time ();
#else
			timespec ts;
			clock_gettime (CLOCK_REALTIME, &ts);
			res = ts.tv_sec * 1000 * 1000 * 1000 + ts.tv_nsec;
#endif
		}

	public:
		static inline double get_frequency ()
		{
			if (freq < 0)
			{
#ifdef _WIN32
				long long llfreq;
				QueryPerformanceFrequency ((LARGE_INTEGER *) &llfreq);
				freq = static_cast <double> (llfreq);
#elif defined (__APPLE__)
				mach_timebase_info_data_t info;
				mach_timebase_info (&info);
				freq = 1e9 * static_cast <double> (info.denom) / static_cast <double> (info.numer);
#else
				//	Linux gives us the values in nanoseconds already,
				//	so just convert that to seconds
				freq = 1e9;
#endif
			}
			
			return freq;
		}
		
		functimer ()
			:	runs (0), accum (0), begin (0), end (0)
		{
		}

		void start ()
		{
			query_timer (begin);
		}

		void stop_at_mark ()
		{
			end = mark;

			if (!begin)
				printf ("error: timer stopped without being started\n");

			insert_measurement (end - begin);
		}
		
		void insert_measurement (double value)
		{
			run_ids.push_back (current_run_id);
			previous.push_back (value);
			
			accum += value;
		}

		void reset ()
		{
			run_ids.clear ();
			previous.clear ();

			runs = 0;
			accum = 0;

			begin =
			end = 0;

			previous.clear ();
		}

		size_t get_runs () const			{ return run_ids.size (); }
		double get_total_time () const		{ return static_cast <double> (accum) / get_frequency (); }

		static int get_current_run_id ()
		{
			return current_run_id;
		}

		int get_run_id (size_t i) const
		{
			return run_ids [i];
		}

		double get_run_time (size_t i) const
		{
			return static_cast <double> (previous [i]) / get_frequency ();
		}

		static void set_mark ()
		{
#ifdef USE_GPU
			glFinish ();
#endif
			query_timer (mark);
		}
	};

	double		functimer::freq = -1.0;
	long long	functimer::mark;
	int			functimer::current_run_id;

	typedef std::list < std::pair <std::string, functimer> >	history_t;
	typedef std::list < std::pair <std::string, history_t> >	histories_t;
	history_t& get_default_history ();

	long		block_new = 0;
	histories_t	histories;
	history_t*	history = &get_default_history ();

	history_t* find_history (char const* n)
	{
		for (histories_t::iterator i = histories.begin (); i != histories.end (); ++ i)
			if (i->first.compare (n) == 0)
				return &i->second;
		return 0;
	}

	history_t& get_default_history ()
	{
		history_t* h = find_history ("");
		if (h) return *h;

		histories.push_back (std::make_pair <std::string, history_t> ("", history_t ()));
		return histories.back ().second;
	}

	functimer* find (char const* n)
	{
		for (history_t::iterator i = history->begin (); i != history->end (); ++ i)
			if (i->first.compare (n) == 0)
				return &i->second;

		return 0;
	}

	void start (char const* n)
	{
		if (block_new)
			return;

		if (functimer* ft = find (n))
			ft->start ();
		else
		{
			history->push_back (std::make_pair <std::string, functimer> (n, functimer ()));
			history->back ().second.start ();
		}
	}

	void stop (char const* n)
	{
		if (block_new)
			return;

		functimer::set_mark ();

		if (functimer* ft = find (n))
			ft->stop_at_mark ();
		else
			printf ("error: trying to stop non-existant timer %s\n", n);
	}

	void insert_measurement (char const* n, double value)
	{
		if (block_new)
			return;

		functimer* ft = find (n);
		if (!ft)
		{
			history->push_back (std::make_pair <std::string, functimer> (n, functimer ()));
			ft = &history->back ().second;
		}

		ft->insert_measurement (value * functimer::get_frequency ());
	}

	void block_new_timers ()
	{
		++ block_new;
	}

	void unblock_new_timers ()
	{
		-- block_new;
	}

	void clear_timers ()
	{
		histories.clear ();
		history = &get_default_history ();
	}

	void current_run (int r)
	{
		functimer::current_run_id = r;
	}

	double get_timer (char const* function)
	{
		//history_t* h = find_history (prefix);
		//if (!h) return 0.0;

		for (history_t::iterator i = history->begin (); i != history->end (); ++ i)
		{
			if (i->first.compare (function) == 0)
				return i->second.get_run_time (functimer::get_current_run_id ());
		}
		return 0.0;
	}

	char const* get_prefix ()
	{
		for (histories_t::iterator i = histories.begin (); i != histories.end (); ++ i)
		{
			if (&i->second == history)
				return i->first.c_str ();
		}
		return "";
	}

	void prefix (char const* p)
	{
		history = find_history (p);
		if (!history)
		{
			histories.push_back (std::make_pair <std::string, history_t> (p, history_t ()));
			history = &histories.back ().second;
		}
	}

	void print_statistics (FILE* f)
	{
		fprintf (f, "runs\ttotal\tavg\t5%%\tprefix\tfunction\n");

		for (histories_t::iterator h = histories.begin (); h != histories.end (); ++ h)
		{
			for (history_t::iterator i = h->second.begin (); i != h->second.end (); ++ i)
			{
				char const* prefix = h->first.c_str ();
				char const* tname = i->first.c_str ();
				functimer&	ft = i->second;

				size_t		N = ft.get_runs ();
				double		avg =ft.get_total_time () / static_cast <double> (N);
				double		var = 0.0;

				for (size_t j = 0; j < N; ++ j)
				{
					double const tmp = ft.get_run_time (j) - avg;
					var += tmp * tmp;
				}
				
				fprintf (f, " %lu\t%.6e\t%.6e\t%.6e\t%s\t%s\n",
							N,
							ft.get_total_time (),
							ft.get_total_time () / N,
							1.96 * std::sqrt (var / (N - 1.0)) / std::sqrt ((double) N),
							prefix,
							tname);
			}
		}
	}

	void print_all_times (FILE* f)
	{
		fprintf (f, "run\ttime\tprefix\tfunction\n");
		
		for (histories_t::iterator h = histories.begin (); h != histories.end (); ++ h)
		{
			for (history_t::iterator i = h->second.begin (); i != h->second.end (); ++ i)
			{
				char const* prefix = h->first.c_str ();
				char const* tname = i->first.c_str ();
				functimer&	ft = i->second;
				size_t		runs = ft.get_runs ();

				for (size_t j = 0; j < runs; ++ j)
				{
					fprintf (f, " %lu\t%.6e\t%s\t%s\n",
								ft.get_run_id (j),
								ft.get_run_time (j),
								prefix,
								tname);
				}
			}
		}
	}
};
